using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using CDBurnerXP.Burning;

namespace CDBurnerXP.CommandLine
{
    public enum Command
    {
        ListDrives,
        Eject,
        Load,
        BurnData,
        BurnAudio,
        Erase,
        BurnIso,
        WaitForDisc
    }

    public struct FileFolderArgument
    {
        public string SourcePath;
        public string TargetPath;

        public bool IsFile
        {
            get
            {
                try
                {
                    return File.Exists(SourcePath);
                }
                catch (Exception)
                {
                    return false;
                }
            }
        }

        public bool IsDir
        {
            get
            {
                try
                {
                    return Directory.Exists(SourcePath);
                }
                catch (Exception)
                {
                    return false;
                }
            }
        }

        public FileFolderArgument(string source, string target)
        {
            SourcePath = source.Trim();
            TargetPath = target.Trim();
        }
    }

    class CommandLineOptions
    {
        private Command m_Command = Command.ListDrives;
        private short m_DeviceIndex = 0;
        private List<string> m_Flags = new List<string>();
        private FileAndFolderList m_FilesAndFolders = new FileAndFolderList();
        private Dictionary<string, string> m_Arguments = new Dictionary<string, string>();

        public class FileAndFolderList : List<FileFolderArgument>
        {
            public bool HasAtLeastOneFile()
            {
                foreach (FileFolderArgument arg in this)
                {
                    if (arg.IsFile) return true;
                }
                return false;
            }

            public FileFolderArgument GetFirstFile()
            {
                foreach (FileFolderArgument arg in this)
                {
                    if (arg.IsFile) return arg;
                }
                return new FileFolderArgument();
            }
        }

        #region Properties

        public Dictionary<string, string> Arguments
        {
            get
            {
                return m_Arguments;
            }
        }

        public FileAndFolderList FilesAndFolders
        {
            get
            {
                return m_FilesAndFolders;
            }
        }

        public Command Command
        {
            get
            {
                return m_Command;
            }
            set
            {
                m_Command = value;
            }
        }

        public short DeviceIndex
        {
            get
            {
                return m_DeviceIndex;
            }
        }

        public List<string> Flags
        {
            get
            {
                return m_Flags;
            }
        }

        #endregion

        public CommandLineOptions(string[] args)
        {
            if (args.Length == 0)
            {
                throw new ArgumentNullException("args");
            }

            // Check which argument fits
            string clean = args[0].Replace("-", "").ToLower();
            foreach (object value in Enum.GetValues(typeof(Command)))
            {
                if (value.ToString().ToLower() == clean)
                {
                    m_Command = (Command)value;
                    break;
                }
            }

            if (args.Length < 2) return;

            // Read flags and key-value pairs
            for (int i = 1; i < args.Length; i++)
            {
                if (!args[i].Contains(":"))
                {
                    m_Flags.Add(args[i]);
                    continue;
                }

                string key = args[i].Substring(0, args[i].IndexOf(':'));
                string baseKey = key;
                if (baseKey.Contains("["))
                {
                    baseKey = key.Substring(0, key.IndexOf('['));
                }
                string value = args[i].Substring(key.Length+1, args[i].Length - key.Length - 1);

                switch (baseKey)
                {
                    case "-device":
                        // If letter, find the right number
                        string driveLetter = value.ToUpper();
                        for (short k = 0; k < BurningEngine.StarBurnEngine.Drives.Count; k++)
                        {
                            if (BurningEngine.StarBurnEngine.Drives[k].DriveInfo.Letter == driveLetter)
                            {
                                m_DeviceIndex = k;
                                continue;
                            }
                        }
                        try
                        {
                            m_DeviceIndex = Convert.ToInt16(value);
                        }
                        catch (FormatException)
                        {
                            throw new ArgumentNullException("device");
                        }
                        break;
                    case "-file":
                    case "-folder":
                        string sourcePath = value;
                        string targetPath = "";
                        if (key.Contains("["))
                        {
                            targetPath = key.TrimEnd(']');
                            int pos = key.IndexOf('[');
                            targetPath = targetPath.Substring(pos + 1, targetPath.Length - pos - 1);
                        }
                        m_FilesAndFolders.Add(new FileFolderArgument(sourcePath, targetPath));
                        break;
                    default:
                        m_Arguments.Add(key, value);
                        break;
                }
            }
        }
    }
}
