using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.Windows.Forms;
using System.IO;
using CDBurnerXP.IO;
using CDBurnerXP.Controls;
using NMSDVDXLib;
using RocketDivision.StarBurnX;

namespace CDBurnerXP.Burning
{
    public class DataCompilation : Compilation
    {
        private DirectoryNode m_DirNode = null;
        private string m_Name = string.Empty;
        private FileStream m_CurrentStream = null;
        private int m_FileAndFolderCount = -1;
        private int m_FilesAndFoldersSavedCount = -1;
        private IDataImageX m_DataImage = null;
        private IDataFolder m_DataBurner = null;

        #region Properties

        public IDataImageX DataImage
        {
            get {
                return m_DataImage;
            }
            set {
                m_DataImage = value;
            }
        }

        private eVolumeTypeX VolumeType
        {
            get
            {
                return (m_DataImage == null) ? eVolumeTypeX.vtISO9660_Joliet_UDF : m_DataImage.VolumeType;
            }
        }

        public string Name
        {
            get
            {
                return m_Name;
            }
            set
            {
                m_Name = value;
                if (m_DataBurner != null)
                {
                    m_DataBurner.Name = value;
                }
            }
        }

        internal DirectoryNode Layout
        {
            get
            {
                return m_DirNode;
            }
        }

        #endregion

        public DataCompilation()
        {
            m_DataImage = BurningEngine.NmsEngine.CreateDataImage();
            m_DataImage.VolumeType = eVolumeTypeX.vtISO9660_Joliet;
            m_DataImage.UseLongJolietNames = true;
            m_DataImage.GetVolumeInfo().ApplicationID = "CDBurnerXP " + Application.ProductVersion;
        }

        public DataCompilation(IDataFolder dataBurner)
        {
            m_DataBurner = dataBurner;
        }

        /// <summary>
        /// Builds a new compilation from a given directory node.
        /// </summary>
        internal DataCompilation(DirectoryNode dirNode)
        {
            m_DirNode = dirNode;
        }

        public override void Save(string filename)
        {
            m_LastProgress = -1;
            m_FilesAndFoldersSavedCount = 0;
            m_FileAndFolderCount = m_DirNode.GetFileAndFolderCount();

            XmlDocument xmlDoc = new XmlDocument();

            xmlDoc.AppendChild(xmlDoc.CreateXmlDeclaration("1.0", "utf-8", "yes"));
            xmlDoc.AppendChild(xmlDoc.CreateDocumentType("layout", "http://www.cdburnerxp.se/help/data.dtd", null, null));

            xmlDoc.AppendChild(xmlDoc.CreateProcessingInstruction("xml-stylesheet", "type='text/xsl' href='http://www.cdburnerxp.se/help/compilation.xsl'"));

            string comment = "data compilation created by CDBurnerXP " + Application.ProductVersion + " (http://cdburnerxp.se)";
            xmlDoc.AppendChild(xmlDoc.CreateComment(comment));

            XmlElement header = xmlDoc.CreateElement("layout");

            header.SetAttribute("type", "Data");
            header.SetAttribute("version", Application.ProductVersion);
            header.SetAttribute("date", Conversion.ToXmlDate(DateTime.Now));

            XmlElement compilation = xmlDoc.CreateElement("compilation");
            compilation.SetAttribute("name", Name);

            XmlElement rootDir = xmlDoc.CreateElement("dir");
            rootDir.SetAttribute("name", Name);
            rootDir.SetAttribute("path", "\\");
            rootDir.SetAttribute("realpath", "");
            BuildXmlRecursive(m_DirNode, rootDir);

            xmlDoc.AppendChild(header);

            // Additional options, boot etc.
            if (m_DataImage != null)
            {
                IBootOptionsX bootOptions = m_DataImage.GetBootOptions();

                XmlElement xmlOptions = xmlDoc.CreateElement("options");
                AddElementString(xmlOptions, "BootFilePath", bootOptions.BootFilePath);
                AddElementString(xmlOptions, "BootPlatform", ((int)bootOptions.BootPlatform).ToString());
                AddElementString(xmlOptions, "LoadSegmentWord", bootOptions.LoadSegmentWord.ToString());
                AddElementString(xmlOptions, "MediaTypeEmulation", ((int)bootOptions.MediaTypeEmulation).ToString());
                AddElementString(xmlOptions, "PartitionType", bootOptions.PartitionType.ToString());
                AddElementString(xmlOptions, "SectorCount", bootOptions.SectorCount.ToString());
                AddElementString(xmlOptions, "UseVersionDelimiter", (Convert.ToInt16(m_DataImage.UseVersionDelimiter)).ToString());
                AddElementString(xmlOptions, "UseLongJolietNames", (Convert.ToInt16(m_DataImage.UseLongJolietNames)).ToString());
                AddElementString(xmlOptions, "VolumeType", ((int)m_DataImage.VolumeType).ToString());
                AddElementString(xmlOptions, "BootableEnabled", (Convert.ToInt16(m_DataImage.GetBootableMode())).ToString());
                AddElementString(xmlOptions, "ISOLevel", ((int)m_DataImage.ISOLevel).ToString());
                header.AppendChild(xmlOptions);
                
                // Save ISO labels
                XmlElement xmlVolumeInfo = xmlDoc.CreateElement("VolumeInfo");
                AddElementString(xmlVolumeInfo, "AbstractFile", m_DataImage.GetVolumeInfo().AbstractFile);
                AddElementString(xmlVolumeInfo, "ApplicationID", m_DataImage.GetVolumeInfo().ApplicationID);
                AddElementString(xmlVolumeInfo, "BiblioFile", m_DataImage.GetVolumeInfo().BiblioFile);
                AddElementString(xmlVolumeInfo, "CopyrightFile", m_DataImage.GetVolumeInfo().CopyrightFile);
                AddElementString(xmlVolumeInfo, "DataPreparerID", m_DataImage.GetVolumeInfo().DataPreparerID);
                AddElementString(xmlVolumeInfo, "PublisherID", m_DataImage.GetVolumeInfo().PublisherID);
                AddElementString(xmlVolumeInfo, "SystemID", m_DataImage.GetVolumeInfo().SystemID);
                AddElementString(xmlVolumeInfo, "VolumeSetID", m_DataImage.GetVolumeInfo().VolumeSetID);

                AddElementString(xmlVolumeInfo, "CreationDateTime", Conversion.ToXmlDate(m_DataImage.GetVolumeInfo().CreationDateTime));
                AddElementString(xmlVolumeInfo, "EffectiveDateTime", Conversion.ToXmlDate(m_DataImage.GetVolumeInfo().EffectiveDateTime));
                AddElementString(xmlVolumeInfo, "ExpirationDateTime", Conversion.ToXmlDate(m_DataImage.GetVolumeInfo().ExpirationDateTime));
                AddElementString(xmlVolumeInfo, "ModificationDateTime", Conversion.ToXmlDate(m_DataImage.GetVolumeInfo().ModificationDateTime));
                
                header.AppendChild(xmlVolumeInfo);
            }

            header.AppendChild(compilation);
            compilation.AppendChild(rootDir);

            if (m_Cancel) return;

            xmlDoc.Save(filename);
        }

        private void AddElementString(XmlElement root, string elementName, string content)
        {
            XmlElement tmp = root.OwnerDocument.CreateElement(elementName);
            tmp.InnerText = content;
            root.AppendChild(tmp);
        }

        private void BuildXmlRecursive(DirectoryNode currentDir, XmlElement currentElement)
        {
            foreach (FileNode fileNode in currentDir.Files.Values)
            {
                if (m_Cancel) return;

                // Save all files except those which are imported from disc
                if (fileNode.IsImported) continue;

                XmlElement file = currentElement.OwnerDocument.CreateElement("file");
                file.SetAttribute("name", fileNode.Name);
                file.SetAttribute("path", fileNode.FileSystemPath);
                file.SetAttribute("hidden", fileNode.Hidden ? "1" : "0");
                file.SetAttribute("date", Conversion.ToXmlDate(fileNode.ModificationDate));

                currentElement.AppendChild(file);
                OnSaveProgressChanged(m_FilesAndFoldersSavedCount++, m_FileAndFolderCount);
            }

            foreach (DirectoryNode dirNode in currentDir.Directories.Values)
            {
                if (m_Cancel) return;

                // Imported directories could possibly contain non-imported
                // files, so save them as well.
                XmlElement dir = currentElement.OwnerDocument.CreateElement("dir");
                dir.SetAttribute("name", dirNode.Name);
                dir.SetAttribute("path", dirNode.FullVirtualPath);
                dir.SetAttribute("realpath", dirNode.FileSystemPath);
                dir.SetAttribute("hidden", dirNode.Hidden ? "1" : "0");
                dir.SetAttribute("date", Conversion.ToXmlDate(dirNode.ModificationDate));

                currentElement.AppendChild(dir);
                OnSaveProgressChanged(m_FilesAndFoldersSavedCount++, m_FileAndFolderCount);

                BuildXmlRecursive(dirNode, dir);
            }
        }

        public override ItemAddError[] Load(string filename)
        {
            m_LastProgress = -1;

            List<ItemAddError> errors = new List<ItemAddError>();

            XmlReaderSettings settings = new XmlReaderSettings();
            settings.ProhibitDtd = false;
            settings.ValidationType = ValidationType.None;

            using (m_CurrentStream = new FileStream(filename, FileMode.Open, FileAccess.Read))
            {
                XmlReader reader = XmlReader.Create(m_CurrentStream, settings);

                while (reader.Read())
                {
                    if (reader.NodeType == XmlNodeType.Element)
                    {
                        if (reader.Name == "compilation") break; // go on with files

                        if (reader.Name == "ROOT")
                        {
                            // 3.5 compatibility
                            m_Name = reader.GetAttribute("CompilationName") as string;
                            break;
                        }

                        if (m_DataImage == null) continue;

                        IBootOptionsX bootOptions = m_DataImage.GetBootOptions();

                        try
                        {
                            switch (reader.Name)
                            {
                                case "BootFilePath":
                                    bootOptions.BootFilePath = reader.ReadElementContentAsString();
                                    break;
                                case "BootPlatform":
                                    bootOptions.BootPlatform = (eBootPlatformX)reader.ReadElementContentAsInt();
                                    break;
                                case "LoadSegmentWord":
                                    bootOptions.LoadSegmentWord = (short)reader.ReadElementContentAsInt();
                                    break;
                                case "MediaTypeEmulation":
                                    bootOptions.MediaTypeEmulation = (eMediaTypeEmulationX)reader.ReadElementContentAsInt();
                                    break;
                                case "PartitionType":
                                    bootOptions.PartitionType = reader.ReadElementContentAsInt();
                                    break;
                                case "SectorCount":
                                    bootOptions.SectorCount = (short)reader.ReadElementContentAsInt();
                                    break;
                                case "UseLongJolietNames":
                                    m_DataImage.UseLongJolietNames = reader.ReadElementContentAsBoolean();
                                    break;
                                case "UseVersionDelimiter":
                                    m_DataImage.UseVersionDelimiter = reader.ReadElementContentAsBoolean();
                                    break;
                                case "VolumeType":
                                    m_DataImage.VolumeType = (eVolumeTypeX)reader.ReadElementContentAsInt();
                                    break;
                                case "ISOLevel":
                                    m_DataImage.ISOLevel = (eISOLevelX)reader.ReadElementContentAsInt();
                                    break;
                                case "BootableEnabled":
                                    m_DataImage.SetBootableMode(reader.ReadElementContentAsBoolean());
                                    break;
                                case "AbstractFile":
                                    m_DataImage.GetVolumeInfo().AbstractFile = reader.ReadElementContentAsString();
                                    break;
                                case "ApplicationID":
                                    m_DataImage.GetVolumeInfo().ApplicationID = reader.ReadElementContentAsString();
                                    break;
                                case "BiblioFile":
                                    m_DataImage.GetVolumeInfo().BiblioFile = reader.ReadElementContentAsString();
                                    break;
                                case "CopyrightFile":
                                    m_DataImage.GetVolumeInfo().CopyrightFile = reader.ReadElementContentAsString();
                                    break;
                                case "DataPreparerID":
                                    m_DataImage.GetVolumeInfo().DataPreparerID = reader.ReadElementContentAsString();
                                    break;
                                case "PublisherID":
                                    m_DataImage.GetVolumeInfo().PublisherID = reader.ReadElementContentAsString();
                                    break;
                                case "SystemID":
                                    m_DataImage.GetVolumeInfo().SystemID = reader.ReadElementContentAsString();
                                    break;
                                case "VolumeSetID":
                                    m_DataImage.GetVolumeInfo().VolumeSetID = reader.ReadElementContentAsString();
                                    break;
                                case "CreationDateTime":
                                    m_DataImage.GetVolumeInfo().CreationDateTime = System.Convert.ToDateTime(reader.ReadElementContentAsString());
                                    break;
                                case "EffectiveDateTime":
                                    m_DataImage.GetVolumeInfo().EffectiveDateTime = System.Convert.ToDateTime(reader.ReadElementContentAsString());
                                    break;
                                case "ExpirationDateTime":
                                    m_DataImage.GetVolumeInfo().ExpirationDateTime = System.Convert.ToDateTime(reader.ReadElementContentAsString());
                                    break;
                                case "ModificationDateTime":
                                    m_DataImage.GetVolumeInfo().ModificationDateTime = System.Convert.ToDateTime(reader.ReadElementContentAsString());
                                    break;
                            }
                        }
                        catch (Exception) { /* if an option is invalid, ignore it */ }
                    }
                }

                // Set name, be fault tolertant
                string name = reader.GetAttribute("name");
                if (!string.IsNullOrEmpty(name))
                {
                    Name = name;
                }

                m_DirNode = new DirectoryNode(null, null, Name);

                if (reader.Name == "ROOT")
                {
                    // 3.5 compatibility
                    reader.ReadToDescendant("DIR");
                    // Start main reading process
                    errors.AddRange(LoadXmlRecursive35(reader, m_DirNode));
                }
                else
                {
                    reader.ReadToDescendant("dir");
                    // Start main reading process
                    errors.AddRange(LoadXmlRecursive(reader, m_DirNode));
                }

                m_Errors = errors.ToArray();

                if (m_Cancel) m_DirNode = null;

                return m_Errors;
            }
        }

        private ItemAddError[] LoadXmlRecursive35(XmlReader reader, DirectoryNode dirNode)
        {
            List<ItemAddError> errors = new List<ItemAddError>();

            while (reader.Read())
            {
                if (m_Cancel) return errors.ToArray();

                if (reader.NodeType != XmlNodeType.Element) continue;

                OnLoadProgressChanged(m_CurrentStream.Position, m_CurrentStream.Length);

                if (reader.Name == "DIR")
                {
                    DirectoryNode newNode = new DirectoryNode(null, dirNode, reader.GetAttribute("Name"));
                    newNode.FileSystemPath = reader.GetAttribute("RealPath");

                    dirNode.Directories.Add(newNode.Name.ToLower(), newNode);
                    // Update Data image
                    if (m_DataImage != null)
                    {
                        m_DataImage.GetFileLayout().AddItem("", newNode.FullVirtualPath);
                    }
                    if (m_DataBurner != null)
                    {
                        EnsureStarBurnFolderExists(newNode.FullVirtualPath, m_DataBurner);
                    }

                    XmlReader dive = reader.ReadSubtree();
                    if (dive.Read())
                    {
                        errors.AddRange(LoadXmlRecursive35(dive, newNode));
                    }
                }
                else if (reader.Name == "FILE")
                {
                    string filePath = reader.GetAttribute("Path");
                    try
                    {
                        FileNode newNode = new FileNode(filePath, dirNode, VolumeType);
                        newNode.Name = reader.GetAttribute("Name");

                        // Update Data image
                        if (m_DataImage != null)
                        {
                            if (!m_DataImage.GetFileLayout().AddItem(filePath, newNode.FullVirtualPath))
                            {
                                throw new ApplicationException("Adding the file failed for an unknown reason.");
                            }

                            dirNode.Files.Add(newNode.Name.ToLower(), newNode);
                        }
                        if (m_DataBurner != null)
                        {
                            // Create and find parent folder
                            string parentFolder = Path.GetDirectoryName(newNode.FullVirtualPath);
                            EnsureStarBurnFolderExists(parentFolder, m_DataBurner);
                            DataFolder folder = m_DataBurner.GetFileObject(parentFolder + "\\") as DataFolder;
                            if (parentFolder.TrimStart('\\') == string.Empty)
                            {
                                m_DataBurner.AddFileEx(filePath, newNode.Name);
                            }
                            else
                            {
                                folder.AddFileEx(filePath, newNode.Name);
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                        ItemAddError error = new ItemAddError(filePath, ex);
                        errors.Add(error);
                    }
                }
            }

            return errors.ToArray();
        }

        /// <summary>
        /// Ensures that a given folder exists. Directories are created recursively.
        /// </summary>
        /// <param name="p">Path to the folder that should exist.</param>
        private void EnsureStarBurnFolderExists(string path, IDataFolder dataBurner)
        {
            path = path.TrimStart('\\');
            if (string.IsNullOrEmpty(path))
            {
                // Root always exists
                return;
            }

            string parentFolder = Path.GetDirectoryName(path);
            if (!string.IsNullOrEmpty(parentFolder))
            {
                // If not root folder, make sure that parent folder exists
                EnsureStarBurnFolderExists(parentFolder, dataBurner);

                // If directory does not exist yet, create.
                // Always search for directories by appending "\" to the path.
                if (dataBurner.GetFileObject(path + "\\") == null)
                {
                    DataFolder parentDir = dataBurner.GetFileObject(parentFolder + "\\") as DataFolder;
                    // Create directory in parent directory, only specify the directory name
                    parentDir.CreateFolder(Path.GetFileName(path));
                }
            }
            else
            {
                // Create directory in the root directory if not present yet
                if (dataBurner.GetFileObject(path + "\\") == null)
                {
                    dataBurner.CreateFolder(path);
                }
            }
        }

        private ItemAddError[] LoadXmlRecursive(XmlReader reader, DirectoryNode dirNode)
        {
            List<ItemAddError> errors = new List<ItemAddError>();

            while (reader.Read())
            {
                if (m_Cancel) return errors.ToArray();

                if (reader.NodeType != XmlNodeType.Element) continue;

                OnLoadProgressChanged(m_CurrentStream.Position, m_CurrentStream.Length);

                if (reader.Name == "dir")
                {
                    DirectoryNode newNode = new DirectoryNode(null, dirNode, reader.GetAttribute("name"));
                    newNode.FileSystemPath = reader.GetAttribute("realpath");
                    newNode.Hidden = Conversion.ToBoolean(reader.GetAttribute("hidden"));
                    newNode.ModificationDate = Convert.ToDateTime(reader.GetAttribute("date"));

                    dirNode.Directories.Add(newNode.Name.ToLower(), newNode);
                    // Update Data image
                    if (m_DataImage != null)
                    {
                        if (m_DataImage.GetFileLayout().AddItem("", newNode.FullVirtualPath))
                        {
                            IDataItemX item = m_DataImage.GetFileLayout().GetLastItemAdded();
                            if (newNode.Hidden)
                            {
                                item.SetIsHidden(newNode.Hidden);
                            }
                            item.SetFileTime(newNode.ModificationDate);
                        }
                    }
                    if (m_DataBurner != null)
                    {
                        EnsureStarBurnFolderExists(newNode.FullVirtualPath, m_DataBurner);
                    }

                    XmlReader dive = reader.ReadSubtree();
                    if (dive.Read())
                    {
                        errors.AddRange(LoadXmlRecursive(dive, newNode));
                    }
                }
                else if (reader.Name == "file")
                {
                    string filePath = reader.GetAttribute("path");
                    try
                    {
                        FileNode newNode = new FileNode(filePath, dirNode, VolumeType);
                        newNode.Name = reader.GetAttribute("name");
                        newNode.Hidden = Conversion.ToBoolean(reader.GetAttribute("hidden"));
                        newNode.ModificationDate = Convert.ToDateTime(reader.GetAttribute("date"));

                        dirNode.Files.Add(newNode.Name.ToLower(), newNode);
                        // Update Data image
                        if (m_DataImage != null)
                        {
                            if (m_DataImage.GetFileLayout().AddItem(filePath, newNode.FullVirtualPath))
                            {
                                IDataItemX item = m_DataImage.GetFileLayout().GetLastItemAdded();
                                if (newNode.Hidden)
                                {
                                    item.SetIsHidden(newNode.Hidden);
                                }
                                item.SetFileTime(newNode.ModificationDate);
                            }
                        }
                        if (m_DataBurner != null)
                        {
                            // Create and find parent folder
                            string parentFolder = Path.GetDirectoryName(newNode.FullVirtualPath);
                            EnsureStarBurnFolderExists(parentFolder, m_DataBurner);
                            DataFolder folder = m_DataBurner.GetFileObject(parentFolder + "\\") as DataFolder;
                            if (parentFolder.TrimStart('\\') == string.Empty)
                            {
                                m_DataBurner.AddFileEx(filePath, newNode.Name);
                            }
                            else
                            {
                                folder.AddFileEx(filePath, newNode.Name);
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                        ItemAddError error = new ItemAddError(filePath, ex);
                        errors.Add(error);
                    }
                }
            }

            return errors.ToArray();
        }

        /// <summary>
        /// Replaces variables within a string for a data disc label.
        /// </summary>
        public static string ReplaceDiscLabelVariables(string labeltext)
        {
            try
            {
                if (string.IsNullOrEmpty(labeltext)) return DateTime.Now.ToShortDateString();

                labeltext = labeltext.Replace("$user", Environment.UserName);
                labeltext = labeltext.Replace("$hour", DateTime.Now.Hour.ToString("00"));
                labeltext = labeltext.Replace("$minute", DateTime.Now.Minute.ToString("00"));
                labeltext = labeltext.Replace("$year", DateTime.Now.Year.ToString());
                labeltext = labeltext.Replace("$yr", DateTime.Now.ToString("yy"));
                labeltext = labeltext.Replace("$month", DateTime.Now.Month.ToString("00"));
                labeltext = labeltext.Replace("$day", DateTime.Now.Day.ToString("00"));
                labeltext = labeltext.Replace("$time", DateTime.Now.ToShortTimeString());
                labeltext = labeltext.Replace("$date", DateTime.Now.ToShortDateString());
            }
            catch (FormatException)
            {
                // can apparently occur at ToShortDateString(), .NET bug
            }
            catch (ArgumentOutOfRangeException)
            {
                // can apparently occur at ToShortDateString(), .NET bug
            }
            return PathEx.ReplaceInvalidFileNameChars(labeltext);
        }

    }
}
