using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Runtime.InteropServices;
using CDBurnerXP.IO;
using RocketDivision.StarBurnX;
using System.Threading;

namespace CDBurnerXP.Burning
{
    /// <summary>
    /// Watches a drive of your choice for inserted
    /// or removed media.
    /// </summary>
    public class MediaNotifier
    {
        private static char m_WatchingDriveLetter = '\0';
        private static bool m_IsEnabled = false;
        private static DEV_BROADCAST_VOLUME lastVolumeInfo;
        private static IntPtr lastEventType;
        public delegate void MediaInsertedEventHandler(string VolumeLabel);
        public delegate void MediaRemovedEventHandler(string VolumeLabel);
        public delegate void MediaInfoLoadedEventHandler(DiscInfo discInfo);

        public static event MediaInsertedEventHandler MediaInserted;
        public static event MediaRemovedEventHandler MediaRemoved;

        public static event MediaInsertedEventHandler AnyMediaInserted;
        public static event MediaRemovedEventHandler AnyMediaRemoved;

        /// <summary>
        /// When a new disc is inserted, the new information is loaded in background
        /// and after that this event fires with the new information.
        /// </summary>
        public static event MediaInfoLoadedEventHandler AnyMediaInfoLoaded;

        #region Properties

        /// <summary>
        /// Gets whether the Notifier is currently
        /// fireing events.
        /// </summary>
        public static bool Enabled
        {
            get
            {
                return m_IsEnabled;
            }
            set
            {
                m_IsEnabled = value;
            }
        }

        /// <summary>
        /// Sets the drive which is supposed to be watched.
        /// </summary>
        public static string DriveLetter
        {
            set
            {
                if (value == null)
                {
                    m_WatchingDriveLetter = '\0';
                }
                else
                {
                    m_WatchingDriveLetter = value[0];
                }
            }
        }

        #endregion

        #region WINAPI

        /// <summary>
        /// Serves as a standard header for information related to a device event reported through the WM_DEVICECHANGE message.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        struct DEV_BROADCAST_HDR
        {
            /// <summary>
            /// The size of this structure, in bytes. 
            /// </summary>
            public uint dbch_Size;

            /// <summary>
            /// The device type, which determines the event-specific information that follows the first three members.
            /// This member can be one of the following values.
            /// </summary>
            public uint dbch_DeviceType;

            /// <summary>
            /// Reserved; do not use.
            /// </summary>
            public uint dbch_Reserved;
        }

        /// <summary>
        /// Contains information about a logical volume.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
        struct DEV_BROADCAST_VOLUME
        {
            /// <summary>
            /// The size of this structure, in bytes.
            /// </summary>
            public int dbcv_Size;

            /// <summary>
            /// Set to DBT_DEVTYP_VOLUME.
            /// </summary>
            public int dbcv_Devicetype;

            /// <summary>
            /// Reserved; do not use.
            /// </summary>
            public int dbcv_Reserved;

            /// <summary>
            /// The logical unit mask identifying one or more logical units. Each bit in the mask corresponds to one logical drive. Bit 0 represents drive A, bit 1 represents drive B, and so on.
            /// </summary>
            public int dbcv_Unitmask;

            /// <summary>
            /// This parameter can be one of the following values.
            /// DBTF_MEDIA = 0x0001, DBTF_NET = 0x0002
            /// </summary>
            public int dbcv_Flags;
        };

        /// <summary>
        /// The device type, which determines the event-specific information that follows the first three members. 
        /// </summary>
        private enum DBT_DEVTYP
        {
            /// <summary>
            /// Class of devices. This structure is a DEV_BROADCAST_DEVICEINTERFACE structure.
            /// </summary>
            DEVICEINTERFACE = 0x00000005,

            /// <summary>
            /// File system handle. This structure is a DEV_BROADCAST_HANDLE structure.
            /// </summary>
            HANDLE = 0x00000006,

            /// <summary>
            /// OEM- or IHV-defined device type. This structure is a DEV_BROADCAST_OEM structure.
            /// </summary>
            OEM = 0x00000000,

            /// <summary>
            /// Port device (serial or parallel). This structure is a DEV_BROADCAST_PORT structure.
            /// </summary>
            PORT = 0x00000003,

            /// <summary>
            /// Logical volume. This structure is a DEV_BROADCAST_VOLUME structure.
            /// </summary>
            VOLUME = 0x00000002
        }

        /// <summary>
        /// http://msdn.microsoft.com/en-us/library/aa363480(VS.85).aspx
        /// </summary>
        private enum DBT_EventType
        {
            /// <summary>
            /// A request to change the current configuration (dock or undock) has been canceled.
            /// </summary>
            CONFIGCHANGECANCELED = 0x0019,

            /// <summary>
            /// The current configuration has changed, due to a dock or undock.
            /// </summary>
            CONFIGCHANGED = 0x0018,

            /// <summary>
            /// A custom event has occurred.
            /// </summary>
            CUSTOMEVENT = 0x8006,

            /// <summary>
            /// A device or piece of media has been inserted and is now available.
            /// </summary>
            DEVICEARRIVAL = 0x8000,

            /// <summary>
            /// Permission is requested to remove a device or piece of media. Any application can deny this request and cancel the removal.
            /// </summary>
            DEVICEQUERYREMOVE = 0x8001,

            /// <summary>
            /// A request to remove a device or piece of media has been canceled.
            /// </summary>
            DEVICEQUERYREMOVEFAILED = 0x8002,

            /// <summary>
            /// A device or piece of media has been removed.
            /// </summary>
            DEVICEREMOVECOMPLETE = 0x8004,

            /// <summary>
            /// A device or piece of media is about to be removed. Cannot be denied.
            /// </summary>
            DEVICEREMOVEPENDING = 0x8003,

            /// <summary>
            /// A device-specific event has occurred.
            /// </summary>
            DEVICETYPESPECIFIC = 0x8005,

            /// <summary>
            /// A device has been added to or removed from the system.
            /// </summary>
            DEVNODES_CHANGED = 0x0007,

            /// <summary>
            /// Permission is requested to change the current configuration (dock or undock).
            /// </summary>
            QUERYCHANGECONFIG = 0x0017,

            /// <summary>
            /// The meaning of this message is user-defined.
            /// </summary>
            USERDEFINED = 0xFFFF
        }

        #endregion

        /// <summary>
        /// Processes a WM_DEVICECHANGE message and fires the events accordingly.
        /// </summary>
        public static void ProcessDeviceChangedEvent(IntPtr wParam, IntPtr lParam)
        {
            // Do not fire events
            if (!m_IsEnabled)
            {
                return;
            }

            // Only watch removal and insertion
            if (wParam.ToInt32() != (int)DBT_EventType.DEVICEARRIVAL && wParam.ToInt32() != (int)DBT_EventType.DEVICEREMOVECOMPLETE)
            {
                return;
            }

            DEV_BROADCAST_HDR structure = (DEV_BROADCAST_HDR)Marshal.PtrToStructure(lParam, typeof(DEV_BROADCAST_HDR));
            // Check device type. We need "VOLUME".
            if (structure.dbch_DeviceType != (uint)DBT_DEVTYP.VOLUME)
            {
                return;
            }

            // Now create a special volume strucure
            DEV_BROADCAST_VOLUME volumeInfo = (DEV_BROADCAST_VOLUME)Marshal.PtrToStructure(lParam, typeof(DEV_BROADCAST_VOLUME));
            if (volumeInfo.dbcv_Flags == 2)
            {
                // Network device, not in our interest
                return;
            }

            // Ignore if the same event is fired twice in a row
            if (lastEventType == wParam && lastVolumeInfo.dbcv_Devicetype == volumeInfo.dbcv_Devicetype && lastVolumeInfo.dbcv_Flags == volumeInfo.dbcv_Flags && lastVolumeInfo.dbcv_Unitmask == volumeInfo.dbcv_Unitmask)
            {
                return;
            }

            lastVolumeInfo = volumeInfo;
            lastEventType = wParam;

            // Fire "any" events
            for (char i = 'a'; i <= 'z'; i++)
            {
                if (PathEx.IsDriveInUnitmask(volumeInfo.dbcv_Unitmask, i))
                {
                    if (wParam.ToInt32() == (int)DBT_EventType.DEVICEARRIVAL)
                    {
                        OnAnyMediaInserted(i);
                    }
                    else if (wParam.ToInt32() == (int)DBT_EventType.DEVICEREMOVECOMPLETE)
                    {
                        if (AnyMediaRemoved != null)
                        {
                            AnyMediaRemoved(i + ":");
                        }
                    }
                }
            }

            // Fire events for the watched drive, if necessary
            if (PathEx.IsDriveInUnitmask(volumeInfo.dbcv_Unitmask, m_WatchingDriveLetter))
            {
                if (wParam.ToInt32() == (int)DBT_EventType.DEVICEARRIVAL)
                {
                    if (MediaInserted != null)
                    {
                        MediaInserted(m_WatchingDriveLetter + ":");
                    }
                }
                else if (wParam.ToInt32() == (int)DBT_EventType.DEVICEREMOVECOMPLETE)
                {
                    if (MediaRemoved != null)
                    {
                        MediaRemoved(m_WatchingDriveLetter + ":");
                    }
                }
            }
        }

        /// <summary>
        /// Invoke reloading the drive information, even if not necessary.
        /// </summary>
        public static void InvokeMediaInfoRefresh(char driveLetter)
        {
            OnAnyMediaInserted(driveLetter);
        }

        private static void OnAnyMediaInserted(char driveLetter)
        {
            if (AnyMediaInserted != null)
            {
                AnyMediaInserted(driveLetter + ":");
            }

            Thread loadInfoThread = new Thread(new ParameterizedThreadStart(OnAnyMediaInfoLoaded));
            loadInfoThread.Start(driveLetter);
        }

        private static void OnAnyMediaInfoLoaded(object letter)
        {
            try
            {
                string driveLetter = Convert.ToString(letter);

                Drive drive = BurningEngine.GetStarBurnDriveByLetter(driveLetter);
                if (drive == null) return;

                drive.DiscInfo.Refresh();

                if (AnyMediaInfoLoaded != null)
                {
                    AnyMediaInfoLoaded(drive.DiscInfo);
                }
            }
            catch (Exception)
            {
                // We don't want this to cause problems!
            }
        }
    }
}
