﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using CDBurnerXP.Burning;
using CDBurnerXP.Forms;
using CDBurnerXP.Windows.Forms;
using CDBurnerXP.IO;
using CDBurnerXP.IO.Cue;
using System.Windows.Forms;
using System.ComponentModel;
using System.Drawing;
using Microsoft.Win32;
using CDBurnerXP.Collections;
using System.Threading;

namespace CDBurnerXP.Controls
{
    public class AudioLayoutManager : LayoutManager
    {
        private ManagedAudioListView m_ListView = null;
        private static Dictionary<UIText, string> m_Translation = new Dictionary<UIText, string>();
        private string m_LastProcessedPath = string.Empty;
        private List<AudioTrack> m_AudioFiles = new List<AudioTrack>();
        private List<AudioTrack> m_TempAudioFiles = new List<AudioTrack>();
        private Form m_WindowOwner = null;
        private List<string> m_CdaTracks = new List<string>();
        private RipCdaTracksDelegate m_CdaRipper = null;
        private string m_CDTextArtist = string.Empty;
        private string m_CDTextTitle = string.Empty;
        private ReplaceOption m_ReplaceOption = ReplaceOption.None;

        public delegate string[] RipCdaTracksDelegate(string[] tracks);

        private class UserCancelledException : ApplicationException { };

        #region Enums

        private enum ReplaceOption
        {
            None,
            AddAll,
            SkipDuplicate
        }

        public enum UIText
        {
            TrackNo,
            Title,
            Length,
            Type,
            Bitrate,
            Artist,
            Album,
            KBits,
            TrackNoByTag,
            SortWarning,
            TrackListHasBeenSorted,
            NeverTellAgain,
            FullPath,
            FileName
        }

        #endregion

        #region Properties

        /// <summary>
        /// Gets or sets the CD-Text Artist information.
        /// </summary>
        [Browsable(false)]
        public string CDTextArtist
        {
            get { return m_CDTextArtist; }
            set { m_CDTextArtist = value; }
        }

        /// <summary>
        /// Gets or sets the CD-Text Title information.
        /// </summary>
        [Browsable(false)]
        public string CDTextTitle
        {
            get { return m_CDTextTitle; }
            set { m_CDTextTitle = value; }
        }

        /// <summary>
        /// Gets whether or not the user has entered any CD-Text information.
        /// </summary>
        public bool HasUserCDTextInformation
        {
            get
            {
                if (!string.IsNullOrEmpty(m_CDTextArtist) || !string.IsNullOrEmpty(m_CDTextTitle))
                {
                    return true;
                }

                foreach (AudioTrack track in m_AudioFiles)
                {
                    if (!string.IsNullOrEmpty(track.CDTextTitle) || !string.IsNullOrEmpty(track.CDTextArtist))
                    {
                        return true;
                    }
                }

                return false;
            }
        }

        public RipCdaTracksDelegate CdaRipper
        {
            get { return m_CdaRipper; }
            set { m_CdaRipper = value; }
        }

        public override string LayoutName
        {
            get
            {
                return string.IsNullOrEmpty(m_SavePath) ? string.Empty : Path.GetFileNameWithoutExtension(m_SavePath);
            }
            set
            {
                throw new NotImplementedException();
            }
        }

        public override ISynchronizeInvoke SynchronizedObject
        {
            get { return BaseForm.GetParentForm(m_ListView); }
        }

        public static Dictionary<UIText, string> Translation
        {
            get {
                if (m_Translation.Count == 0)
                {
                    m_Translation[UIText.Bitrate] = "Bitrate";
                    m_Translation[UIText.Title] = "Title";
                    m_Translation[UIText.Length] = "Length";
                    m_Translation[UIText.TrackNo] = "Track #";
                    m_Translation[UIText.Type] = "Type";
                    m_Translation[UIText.Artist] = "Artist";
                    m_Translation[UIText.Album] = "Album";
                    m_Translation[UIText.KBits] = "KBit/s";
                    m_Translation[UIText.TrackNoByTag] = "Track # by tags";
                    m_Translation[UIText.SortWarning] = "When sorting tracks, keep in mind that this does not only affect the view, but also the order of the tracks when burned to disc.";
                    m_Translation[UIText.TrackListHasBeenSorted] = "The track list has been sorted.";
                    m_Translation[UIText.NeverTellAgain] = "&Never tell again";
                    m_Translation[UIText.FullPath] = "Full path";
                    m_Translation[UIText.FileName] = "File name";
                }
                return m_Translation;
            }
        }

        public ManagedAudioListView ListView
        {
            get { return m_ListView; }
            set {
                m_ListView = value;
                m_ListView.LayoutManager = this;
                m_ListView.ColumnClick += new ColumnClickEventHandler(ReadItemOrderFromList);
            }
        }

        /// <summary>
        /// Returns an array of audio tracks, in the
        /// right order.
        /// </summary>
        public AudioTrack[] AudioTracks
        {
            get
            {
                return m_AudioFiles.ToArray();
            }
        }

        /// <summary>
        /// Returns an array of file names, which should
        /// be burned to disc in the desired order.
        /// </summary>
        public string[] Tracks
        {
            get
            {
                List<string> tracks = new List<string>();

                foreach (AudioTrack track in m_AudioFiles)
                {
                    tracks.Add(track.FileName);
                }

                return tracks.ToArray();
            }
        }

        /// <summary>
        /// Returns an array containing all track
        /// lengths in milliseconds in the right order.
        /// </summary>
        public int[] TrackTimes
        {
            get
            {
                List<int> times = new List<int>();

                foreach (AudioTrack track in m_AudioFiles)
                {
                    times.Add((int)track.Length.TotalMilliseconds);
                }

                return times.ToArray();
            }
        }

        #endregion

        public AudioLayoutManager()
        {
        }

        /// <summary>
        /// Adds audio files from the specified path.
        /// </summary>
        /// <param name="sourceReal">The path which is to be searched recursively</param>
        /// <param name="pos">Position at which the tracks should be added. -1 is at the end.</param>
        /// <returns>A collection of all errors which happen during the process.</returns>
        public ItemAddError[] BeginAddPaths(string[] sourceReal, int pos)
        {
            return BeginAddPaths(sourceReal, pos, BaseForm.GetParentForm(ListView));
        }

        public ItemAddError[] BeginAddPaths(string[] source, int pos, Form ownerForm)
        {
            m_CancelAddingFiles = false;
            m_WindowOwner = ownerForm;
            m_ReplaceOption = ReplaceOption.None;

            using (ProgressDialog dialog = new ProgressDialog())
            {
                dialog.OnDoWork = delegate()
                {
                    m_TempAudioFiles.Clear();
                    ItemAddError[] errors = AddPaths(source);

                    // Fix possibly incorrect position
                    if (pos > m_AudioFiles.Count) pos = m_AudioFiles.Count;

                    // Insert tracks in the right order
                    if (pos == -1)
                    {
                        m_AudioFiles.AddRange(m_TempAudioFiles);
                    }
                    else
                    {
                        foreach (AudioTrack track in m_TempAudioFiles)
                        {
                            m_AudioFiles.Insert(pos++, track);
                        }
                    }
                    return errors;
                };
                dialog.OnCancel = delegate()
                {
                    this.CancelAddingFiles();
                };
                dialog.OnUpdateStatus = delegate()
                {
                    return m_LastProcessedPath;
                };

                // Show the necessary dialogs
                Form owner = ownerForm == null ? BaseForm.GetParentForm(m_ListView) : ownerForm;
                dialog.ShowDialog(owner);
                FileAddErrorsDialog.Show(owner, dialog.Result as ItemAddError[]);

                RefreshList();
                OnCompilationChanged(true, true);

                return dialog.Result as ItemAddError[];
            }
        }

        /// <summary>
        /// Moves the given tracks to the given position in the list.
        /// </summary>
        public void MoveTracks(AudioTrack[] tracks, int newPos)
        {
            if (newPos > m_AudioFiles.Count) return;

            // Remove tracks from list
            foreach (AudioTrack track in tracks)
            {
                // We might have to update the position
                if (newPos > m_AudioFiles.IndexOf(track))
                {
                    newPos--;
                }
                m_AudioFiles.Remove(track);
            }

            if (newPos == -1)
            {
                m_AudioFiles.AddRange(tracks);
            }
            else
            {
                m_AudioFiles.InsertRange(newPos, tracks);
            }

            RefreshList();
            m_ListView.SelectObjects(tracks);
            OnCompilationChanged(false, true);
        }

        /// <summary>
        /// Randomizes the complete track list.
        /// </summary>
        public void RandomizeTracks()
        {
            this.m_AudioFiles.Shuffle();
            RefreshList();
            OnCompilationChanged(false, true);
        }

        public void LoadAudioCompilation(string filename)
        {
            AudioCompilation comp = new AudioCompilation();

            using (ProgressDialog dialog = new ProgressDialog("Loading Compilation", "Loading your compilation, please wait..."))
            {
                comp.LoadProgressChanged += delegate(object sender, CompilationProgressEventArgs e)
                {
                    dialog.ReportProgress(e.Value);
                };
                dialog.OnDoWork = delegate()
                {
                    comp.Load(filename);
                    if (!dialog.Cancelled)
                    {
                        m_AudioFiles = comp.AudioTracks;
                        m_CDTextArtist = comp.Artist;
                        m_CDTextTitle = comp.Title;
                        RefreshList();
                        m_UnsavedChanges = false;
                        m_SavePath = null;
                    }
                    return null;
                };
                dialog.OnCancel = delegate()
                {
                    comp.Cancel();
                };
                dialog.ShowDialog(BaseForm.GetParentForm(ListView));

                if (dialog.Error != null) throw dialog.Error;

                if (!dialog.Cancelled)
                {
                    FileAddErrorsDialog.Show(BaseForm.GetParentForm(ListView), comp.Errors);
                    m_SavePath = filename;
                    OnCompilationChanged(true, false);
                }
            }
        }

        public void SaveAudioCompilation(string filename)
        {
            bool isNativeCompilation = Playlists.IsCUE(filename) || Playlists.IsM3U(filename);

            using (ProgressDialog dialog = new ProgressDialog("Saving Compilation", "Saving your compilation, please wait..."))
            {
                AudioCompilation comp = new AudioCompilation(m_AudioFiles.ToArray());
                comp.Artist = m_CDTextArtist;
                comp.Title = m_CDTextTitle;

                comp.SaveProgressChanged += delegate(object sender, CompilationProgressEventArgs e)
                {
                    dialog.ReportProgress(e.Value);
                };
                dialog.OnDoWork = delegate()
                {
                    if (Playlists.IsCUE(filename))
                    {
                        comp.SaveAsCue(filename);
                    }
                    else if (Playlists.IsM3U(filename))
                    {
                        comp.SaveAsM3u(filename);
                    }
                    else
                    {
                        comp.Save(filename);
                    }
                    return null;
                };
                dialog.ShowDialog(BaseForm.GetParentForm(ListView));

                if (dialog.Error != null) throw dialog.Error;
            }
            
            // Do not mark as saved when using a non "audio compilation", since it's not a "lossless" method
            if (!isNativeCompilation)
            {
                // "Changed" only in the sense of "we are no longer unsaved"!
                OnCompilationChanged(false, false);
                m_SavePath = filename;
            }
        }

        private ItemAddError[] AddPaths(string[] paths)
        {
            List<ItemAddError> errors = new List<ItemAddError>();
            m_CdaTracks = new List<string>();

            foreach (string path in paths)
            {
                if (m_CancelAddingFiles) break;

                m_LastProcessedPath = path;

                try
                {
                    if (Directory.Exists(path))
                    {
                        errors.AddRange(AddPaths(Directory.GetFiles(path)));
                        errors.AddRange(AddPaths(Directory.GetDirectories(path)));
                    }
                    else
                    {
                        errors.AddRange(AddFile(path, default(Track), null));
                    }
                }
                catch (UserCancelledException)
                {
                    return new ItemAddError[0];
                }
                catch (Exception ex)
                {
                    errors.Add(new ItemAddError(path, ex));
                }
            }

            if (m_CdaTracks.Count > 0 && m_CdaRipper != null && m_WindowOwner != null)
            {
                m_WindowOwner.Invoke((MethodInvoker)delegate() {
                    errors.AddRange(AddPaths(m_CdaRipper(m_CdaTracks.ToArray())));
                });
            }

            return errors.ToArray();
        }

        /// <summary>
        /// Adds a track to the audio compilation.
        /// </summary>
        /// <param name="filename">File name of the audio file to add</param>
        /// <param name="cueSheetTrack">If the audio file has been added from a cue sheet, the associated cue sheet information</param>
        /// <param name="sourceFile">Source of the specified audio file (for example a cue sheet or playlist)</param>
        /// <exception cref="UserCancelledException">Thrown, if user clicks cancel in the replace dialog</exception>
        /// <returns>Array of errors that occured while adding files</returns>
        private ItemAddError[] AddFile(string filename, Track cueSheetTrack, string sourceFile)
        {
            try
            {
                // Filter files that do not match the criteria
                if (!Shell32.FileMatchesFilter(filename, ApplicationSettings.AudioFilePattern, ApplicationSettings.AudioFilePatternInclude))
                {
                    return new ItemAddError[0];
                }

                if (Playlists.IsAudioFile(filename))
                {
                    // Check if track has already been added
                    Dictionary<string, bool> allTracks = new Dictionary<string, bool>(StringComparer.OrdinalIgnoreCase);
                    foreach (AudioTrack track in this.AudioTracks)
                    {
                        allTracks[track.FileName] = true;
                    }
                    foreach (AudioTrack track in this.m_TempAudioFiles)
                    {
                        allTracks[track.FileName] = true;
                    }

                    if (allTracks.ContainsKey(filename) && m_ReplaceOption == ReplaceOption.None)
                    {
                        string msg = BaseForm.FormatTranslation("The audio compilation already contains the track \"{0}\".", Path.GetFileName(filename));
                        using (Icon icon = IconReader.GetFileIcon(filename, IconReader.IconSize.Large, false))
                        {
                            string extInfo = BaseForm.FormatTranslation("File name: {0}", filename);
                            if (!string.IsNullOrEmpty(sourceFile))
                            {
                                extInfo += Environment.NewLine + BaseForm.FormatTranslation("Source: {0}", sourceFile);
                            }
                            TaskDialogResult res = null;
                            InvokeIfNecessary((MethodInvoker)delegate()
                            {
                                res = TaskDialog.ShowTaskDialogBox(m_WindowOwner, Application.ProductName, msg, "", extInfo, "", BaseForm.RM.GetString("Do this for &all following conflicts"), "", BaseForm.RM.GetString("A&dd track anyway") + "|" + BaseForm.RM.GetString("Do&n't add track"), eTaskDialogButtons.Cancel, eSysIcons.Information, eSysIcons.Information, icon);
                            });

                            // Cancel whole process
                            if (res.DialogResult == DialogResult.Cancel)
                            {
                                throw new UserCancelledException();
                            }
                            // Option: Do not add cue sheet at all
                            if (res.SelectedCommand == 1)
                            {
                                if (res.VerificationChecked) m_ReplaceOption = ReplaceOption.SkipDuplicate;
                                return new ItemAddError[0];
                            }
                            else if (res.VerificationChecked)
                            {
                                // Always add all tracks
                                m_ReplaceOption = ReplaceOption.AddAll;
                            }
                        }
                    }

                    // Auto-skip duplicate tracks
                    if (allTracks.ContainsKey(filename) && m_ReplaceOption == ReplaceOption.SkipDuplicate) return new ItemAddError[0];

                    AudioTrack newTrack = new AudioTrack(filename);

                    // If track has been added from Cue sheet, add title and artist information
                    if (!string.IsNullOrEmpty(cueSheetTrack.Title))
                    {
                        newTrack.CDTextTitle = cueSheetTrack.Title;
                    }
                    if (!string.IsNullOrEmpty(cueSheetTrack.Performer))
                    {
                        newTrack.CDTextArtist = cueSheetTrack.Performer;
                    }

                    m_TempAudioFiles.Add(newTrack);
                }
                else if (Playlists.IsSupportedPlaylist(filename))
                {
                    string[] files = Playlists.GetFilesFromPlaylist(filename);
                    Array.ForEach(files, delegate(string x) { AddFile(x, default(Track), filename); });
                }
                else if (Playlists.IsCUE(filename))
                {
                    CueSheet sheet = new CueSheet(filename);
                    SplitCueDialog dialog = new SplitCueDialog(sheet, PathEx.GetTempRipPath());
                    InvokeIfNecessary((MethodInvoker)delegate() { dialog.ShowDialog(m_WindowOwner); });

                    List<ItemAddError> errors = new List<ItemAddError>();
                    for (int i = 0; i < dialog.ResultFiles.Length; i++)
                    {
                        errors.AddRange(this.AddFile(dialog.ResultFiles[i], sheet.Tracks[i], filename));
                    }

                    return errors.ToArray();
                }
                else if (Playlists.IsCda(filename))
                {
                    // Rip later, all at once
                    m_CdaTracks.Add(filename);
                }
            }
            catch (UserCancelledException)
            {
                throw;
            }
            catch (Exception ex)
            {
                return new ItemAddError[] { new ItemAddError(filename, ex) };
            }

            return new ItemAddError[0];
        }

        private void InvokeIfNecessary(MethodInvoker methodInvoker)
        {
            if (m_WindowOwner != null)
            {
                m_WindowOwner.Invoke(methodInvoker);
            }
            else
            {
                methodInvoker();
            }
        }

        internal void RefreshList()
        {
            m_ListView.LastSortOrder = SortOrder.None;
            m_ListView.SetObjects(m_AudioFiles);
            // If sorting is in place, read sorted files
            ReadItemOrderFromList();
        }

        public override void Clear()
        {
            m_AudioFiles.Clear();
            m_SavePath = null;
            // CD-Text
            m_CDTextArtist = string.Empty;
            m_CDTextTitle = string.Empty;

            OnCompilationChanged(true, false);
            RefreshList();

            Thread thread = new Thread(delegate()
            {
                PathEx.TryDeleteDirectory(PathEx.GetTempRipPath());
            });
            thread.IsBackground = true;
            thread.Start();
        }

        public override void Clear(string newLayoutName)
        {
            // Does not support setting the name
            Clear();
        }

        /// <summary>
        /// Returns the total length of all tracks.
        /// </summary>
        public TimeSpan GetTotalLength(bool includePauses)
        {
            TimeSpan length = TimeSpan.Zero;
            foreach (AudioTrack track in m_AudioFiles)
            {
                length += track.Length;
            }

            if (includePauses && Tracks.Length > 0)
            {
                // Add track-at-once gaps
                length += TimeSpan.FromMilliseconds(2000 * (Tracks.Length - 1));
            }
      
            return length;
        }

        public void RemoveTracks(IEnumerable<AudioTrack> tracks)
        {
            foreach (AudioTrack track in tracks)
            {
                m_AudioFiles.Remove(track);
            }
            OnCompilationChanged(true, true);
            ListView.SetObjects(m_AudioFiles);
        }

        /// <summary>
        /// Re-reads the tracks from the list view.
        /// Can be done after sorting, so use the new order.
        /// </summary>
        private void ReadItemOrderFromList(object sender, ColumnClickEventArgs e)
        {
            TaskDialog.ShowOptionalMessageBox(BaseForm.GetParentForm(m_ListView), "AudioTracksSorted", Application.ProductName, Translation[UIText.TrackListHasBeenSorted], Translation[UIText.SortWarning], eTaskDialogButtons.OK, eSysIcons.Warning, 0);

            ReadItemOrderFromList();

            OnCompilationChanged(false, true);
        }

        /// <summary>
        /// Re-reads the tracks from the list view.
        /// Can be done after sorting, so use the new order.
        /// </summary>
        private void ReadItemOrderFromList()
        {
            m_AudioFiles.Clear();
            for (int i = 0; i < ListView.GetItemCount(); i++)
            {
                AudioTrack track = ListView.GetModelObject(i) as AudioTrack;
                m_AudioFiles.Add(track);
            }
        }
    }
}
