﻿using System;
using System.Collections.Generic;
using System.Text;
using CDBurnerXP.IO;
using BassLib = Un4seen.Bass.Bass;
using BassTags = Un4seen.Bass.AddOn.Tags;
using System.IO;
using System.Runtime.InteropServices;

namespace CDBurnerXP.Controls
{
    public class AudioTrack
    {
        /// <summary>
        /// The WMIsContentProtected function checks a file for DRM-protected content.
        /// This function is a shortcut so that your application can quickly identify protected files.
        /// </summary>
        [DllImport("WMVCore.dll", CharSet = CharSet.Unicode)]
        private static extern int WMIsContentProtected(string pwszFileName, out bool pfIsProtected);

        #region TrackInvalidException

        public class TrackInvalidException : ApplicationException
        {
            private static string m_Message = string.Empty;

            public static string MessageText
            {
                get { return m_Message; }
                set { m_Message = value; }
            }

            public TrackInvalidException() : base(MessageText)
            {
            }
        }

        #endregion

        #region FileIsDrmProtectedException

        public class FileIsDrmProtectedException : ApplicationException
        {
            private static string m_Message = string.Empty;

            public static string MessageText
            {
                get { return m_Message; }
                set { m_Message = value; }
            }

            public FileIsDrmProtectedException()
                : base(MessageText)
            {
            }
        }

        #endregion
        
        private TimeSpan m_Length = TimeSpan.Zero;
        private string m_FileName = null;
        private BassTags.TAG_INFO m_TagInfo = null;
        private long m_FileSize = 0;
        private string m_CDTextArtist = string.Empty;
        private string m_CDTextTitle = string.Empty;

        #region Properties

        /// <summary>
        /// Gets or sets the CD-Text Artist information.
        /// </summary>
        public string CDTextArtist
        {
            get { return m_CDTextArtist; }
            set { m_CDTextArtist = value; }
        }

        /// <summary>
        /// Gets or sets the CD-Text Title information.
        /// </summary>
        public string CDTextTitle
        {
            get { return m_CDTextTitle; }
            set { m_CDTextTitle = value; }
        }

        public int TrackNo
        {
            get
            {
                int trackNo;
                if (m_TagInfo != null && Int32.TryParse(m_TagInfo.track, out trackNo))
                {
                    return trackNo;
                }
                return 0;
            }
        }

        public string FileType
        {
            get
            {
                return Path.GetExtension(m_FileName).Trim('.').ToUpper();
            }
        }

        public int Bitrate
        {
            get
            {
                return (int)(m_FileSize * 8 / m_Length.TotalSeconds / 1000);
            }
        }

        public string FileName
        {
            get { return m_FileName; }
        }

        public TimeSpan Length
        {
            get { return m_Length; }
        }

        public string Album
        {
            get
            {
                if (m_TagInfo == null)
                {
                    return string.Empty;
                }
                if (!string.IsNullOrEmpty(m_TagInfo.album)) 
                {
                    return m_TagInfo.album;
                }
                return m_TagInfo.genre;
            }
        }

        public string Artist
        {
            get
            {
                if (m_TagInfo == null)
                {
                    return string.Empty;
                }
                // Try hard to find a more or less suitable value
                if (!string.IsNullOrEmpty(m_TagInfo.artist))
                {
                    return m_TagInfo.artist;
                }
                if (!string.IsNullOrEmpty(m_TagInfo.albumartist))
                {
                    return m_TagInfo.albumartist;
                }
                return m_TagInfo.composer;
            }
        }

        public string Title
        {
            get
            {
                if (m_TagInfo != null && !string.IsNullOrEmpty(m_TagInfo.title))
                {
                    return m_TagInfo.title;
                }

                return Path.GetFileNameWithoutExtension(m_FileName);
            }
        }

        #endregion

        public AudioTrack(string filename): this(filename, false)
        {
        }

        public AudioTrack(string filename, bool doNotReadBinStream)
        {
            m_FileName = filename;

            if (!File.Exists(filename)) throw new FileNotFoundException();

            // Get tags
            m_TagInfo = BassTags.BassTags.BASS_TAG_GetFromFile(filename);
            FixTagEncoding();

            // Calculate length
            int audioStream = BassLib.BASS_StreamCreateFile(filename, 0, 0, Un4seen.Bass.BASSFlag.BASS_STREAM_PRESCAN);

            if (audioStream == 0)
            {
                // PRESCAN only required for VBR. It seems to fail in rare occasions though.
                audioStream = BassLib.BASS_StreamCreateFile(filename, 0, 0, Un4seen.Bass.BASSFlag.BASS_STREAM_DECODE);
            }

            if (audioStream == 0 && !(PathEx.IsExtension(filename, "bin") && doNotReadBinStream))
            {
                if (IsFileDrmProtected(filename))
                {
                    throw new FileIsDrmProtectedException();
                }
                else
                {
                    Debug.LogError(new ArgumentException("Track cannot be opened: " + BassLib.BASS_ErrorGetCode().ToString()));
                    throw new TrackInvalidException();
                }
            }
            else if (audioStream != 0)
            {
                double seconds = BassLib.BASS_ChannelBytes2Seconds(audioStream, BassLib.BASS_ChannelGetLength(audioStream));
                m_Length = TimeSpan.FromSeconds(seconds);

                BassLib.BASS_StreamFree(audioStream);
                if (m_Length == TimeSpan.Zero) throw new TrackInvalidException();
            }

            // Get file size
            FileInfo info = new FileInfo(filename);
            m_FileSize = info.Length;
        }

        /// <summary>
        /// Determines whether or not a file is DRM protected.
        /// </summary>
        public static bool IsFileDrmProtected(string filename)
        {
            bool isProtected;
            try
            {
                WMIsContentProtected(filename, out isProtected);
                return isProtected;
            }
            catch (DllNotFoundException)
            {
                // Ignore if check fails
            }
            catch (EntryPointNotFoundException)
            {
                // Ignore if check fails
            }
            return false;
        }

        public override string ToString()
        {
            string desc = FormatAudioTags(m_TagInfo);

            // Fallback if tags could not be read
            return string.IsNullOrEmpty(desc) ? Path.GetFileNameWithoutExtension(m_FileName) : desc;
        }

        /// <summary>
        /// Reduces the string to a size which is allowed for CD-Text.
        /// </summary>
        public static string StripToCDTextSize(string text)
        {
            if (string.IsNullOrEmpty(text)) return string.Empty;

            return text.Substring(0, Math.Min(32, text.Length));
        }

        /// <summary>
        /// Nicely prints a track description based on the tag information.
        /// </summary>
        /// <param name="info">Existing tags</param>
        /// <returns>string.empty if no description could be determined.</returns>
        public static string FormatAudioTags(BassTags.TAG_INFO info)
        {
            string result = string.Empty;
            if (info == null) return result;

            if (!string.IsNullOrEmpty(info.title))
            {
                result = info.title;
            }

            if (!string.IsNullOrEmpty(info.album))
            {
                result = info.album + " - " + result;
            }

            if (!string.IsNullOrEmpty(info.artist))
            {
                result = info.artist + " - " + result;
            }

            return result;
        }

        /// <summary>
        /// Checks if the encoding of the tag data might be written as UTF-8, and converts the text properly.
        /// </summary>
        private void FixTagEncoding()
        {
            if (m_TagInfo == null)
            {
                return;
            }

            if ((m_TagInfo.channelinfo.ctype & Un4seen.Bass.BASSChannelType.BASS_CTYPE_STREAM_WAV) != Un4seen.Bass.BASSChannelType.BASS_CTYPE_STREAM_WAV)
            {
                return;
            }

            // Fix: If UTF-8 encoded data is written into RIFF tags (incorrect),
            // check if the UTF-8 string has a shorter representation of the text.
            // In this case, it is likely that UTF-8 data has been written.
            if (!string.IsNullOrEmpty(m_TagInfo.title))
            {
                byte[] titleData = Encoding.GetEncoding("windows-1252").GetBytes(m_TagInfo.title);
                string utf8Text = Encoding.UTF8.GetString(titleData);
                if (utf8Text.Length < m_TagInfo.title.Length)
                {
                    m_TagInfo.title = utf8Text;
                }
            }

            if (!string.IsNullOrEmpty(m_TagInfo.artist))
            {
                byte[] artistData = Encoding.GetEncoding("windows-1252").GetBytes(m_TagInfo.artist);
                string utf8Text = Encoding.UTF8.GetString(artistData);
                if (utf8Text.Length < m_TagInfo.artist.Length)
                {
                    m_TagInfo.artist = utf8Text;
                }
            }
        }
    }
}
