﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Drawing2D;
using CDBurnerXP.Drawing;

namespace CDBurnerXP.Controls
{
    /// <summary>
    /// A bar which can be used to represent disc space usage (Vista style).
    /// If it is more thn 90% full, the color turns from blue to red.
    /// </summary>
    public partial class DiscSpaceBar : UserControl
    {
        private float maxValue;
        private float currentValue;

        #region Properties

        /// <summary>
        /// Gets or sets the maximum value for the bar.
        /// <example>Total disc space</example>
        /// </summary>
        [DefaultValue(0)]
        public float MaxValue
        {
            get { return this.maxValue; }
            set
            {
                if (this.maxValue != value)
                {
                    this.maxValue = value;
                    Invalidate();
                }
            }
        }

        /// <summary>
        /// Gets or sets the current value of the bar.
        /// <example>Used disc space</example>
        /// </summary>
        [DefaultValue(0)]
        public float CurrentValue
        {
            get { return this.currentValue; }
            set
            {
                if (this.currentValue != value)
                {
                    this.currentValue = value;
                    Invalidate();
                }
            }
        }

        #endregion

        public DiscSpaceBar()
        {
            InitializeComponent();
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            // White background
            using (SolidBrush white = new SolidBrush(Color.FromArgb(251, 251, 251)))
            {
                e.Graphics.FillRectangle(white, 1, 1, Width - 2, Height - 2);
            }

            // Both gradients
            float percentFilled = Math.Min(1, (maxValue == 0) ? 0 : currentValue / maxValue);
            int xPos = Convert.ToInt32(Width * percentFilled);
            bool almostFull = (percentFilled > 0.9);

            Point start = new Point(0, Height / 2);
            Point end = new Point(xPos, Height / 2);

            Color topDarkColor = almostFull ? Color.FromArgb(255, 119, 119) : Color.FromArgb(43, 159, 195);
            Color topLightColor = almostFull ? Color.FromArgb(255, 191, 191) : Color.FromArgb(120, 227, 248);

            Color bottomDarkColor = almostFull ? Color.FromArgb(178, 0, 0) : Color.FromArgb(1, 81, 103);
            Color bottomLightColor = almostFull ? Color.FromArgb(228, 0, 0) : Color.FromArgb(87, 207, 229);

            if (xPos > 0)
            {
                using (LinearGradientBrush gradient = new LinearGradientBrush(start, end, topDarkColor, topLightColor))
                {
                    gradient.SetSigmaBellShape(0.5F, 1.0F);
                    e.Graphics.FillRectangle(gradient, 1, 1, xPos - 2, Height / 2);
                }

                using (LinearGradientBrush gradient = new LinearGradientBrush(start, end, bottomDarkColor, bottomLightColor))
                {
                    gradient.SetSigmaBellShape(0.5F, 1.0F);
                    e.Graphics.FillRectangle(gradient, 1, Height / 2, xPos - 2, Height / 2);
                }
            }

            // Inner border
            using (Pen pen = new Pen(Color.FromArgb(40, 114, 114, 114)))
            {
                e.Graphics.DrawRectangle(pen, 1, 1, Width - 3, Height - 3);
            }

            // Outer border
            using (Pen pen = new Pen(Color.FromArgb(178, 178, 178)))
            {
                GraphicsPath path = RoundedRectangle.Create(0, 0, Width - 1, Height - 1, 2);
                e.Graphics.DrawPath(pen, path);
            }        
        }
    }
}
