using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.ComponentModel;
using System.IO;
using CDBurnerXP.IO;
using CDBurnerXP.Burning;
using CDBurnerXP.Forms;
using CDBurnerXP.Windows.Forms;
using Microsoft.Win32;
using NMSDVDXLib;
using System.Drawing;
using System.Collections;
using CDBurnerXP.Drawing;

namespace CDBurnerXP.Controls
{
    public class FileLayoutManager : LayoutManager
    {
        private RememberReplaceAction m_ReplaceAction = RememberReplaceAction.None;
        private string m_LayoutName = string.Empty;
        private ManagedFileListView m_ListView = null;
        private ManagedDirectoryTreeView m_TreeView = null;
        private DirectoryNode m_Root = null;
        private IDataImageX m_DataImage = null;
        private static ImageList m_FileIcons = new ImageList();
        private static Dictionary<UIText, string> m_Translations = new Dictionary<UIText, string>();
        private bool m_CancelUpdatingCompilation = false;
        private bool m_CancelFileLimitationSearch = false;
        private double m_ImportTotalSize = 0;
        private double m_ImportDoneSize = 0;
        private SpecialFileHandlerDelegate m_SpecialFileHandler;
        private static string m_CurrentProgressString = string.Empty;

        public delegate bool SpecialFileHandlerDelegate(string[] sourcePaths);

        #region UpdateInstruction

        private class UpdateInstruction {
            public string[] SourcePaths;
            public string TargetPath;
            public ActionType Action;

            public enum ActionType
            {
                Add,
                Remove,
                UpdateDate
            }

            public UpdateInstruction(string[] sourcePaths, string targetPath, ActionType action)
            {
                SourcePaths = sourcePaths;
                TargetPath = targetPath;
                Action = action;
            }
        }

        #endregion

        #region Enums

        public enum UIText
        {
            BlankDirectoryOrFile,
            InvalidFileName,
            CouldNotRenameAlreadyExists,
            File,
            Folder,
            Name,
            Size,
            Date,
            Type,
            NewFolder,
            Caching,
            AddFilesHere,
            InvalidLetters
        }

        private enum RememberReplaceAction
        {
            None,
            YesAll,
            NoAll,
            Cancel,
            AllChanged
        }

        public enum ReplaceAction
        {
            Replace,
            Keep,
            Cancel
        }

        #endregion

        #region Properties

        public static string CurrentProgressString
        {
            get
            {
                return m_CurrentProgressString;
            }
            set
            {
                m_CurrentProgressString = value;
            }
        }

        public SpecialFileHandlerDelegate SpecialFileHandler
        {
            set
            {
                m_SpecialFileHandler = value;
            }
            get
            {
                return m_SpecialFileHandler;
            }
        }

        public override ISynchronizeInvoke SynchronizedObject
        {
            get { return BaseForm.GetParentForm(m_ListView, m_TreeView); }
        }

        /// <summary>
        /// Contains icons for all file types.
        /// Special images:
        /// ?Folder
        /// ?FolderOpen
        /// ?File
        /// ?FolderImport
        /// ?FolderOpenImport
        /// </summary>
        public static ImageList FileIcons
        {
            get
            {
                // Init?
                if (m_FileIcons.Images.Count == 0)
                {
                    m_FileIcons.TransparentColor = System.Drawing.Color.Transparent;
                    m_FileIcons.ColorDepth = ColorDepth.Depth32Bit;
                    m_FileIcons.Images.Add("?Folder", CsLib.Properties.Resources.Folder);
                    m_FileIcons.Images.Add("?FolderOpen", CsLib.Properties.Resources.FolderAdd);
                    m_FileIcons.Images.Add("?File", CsLib.Properties.Resources.Document);
                    m_FileIcons.Images.Add("?FileImport", CsLib.Properties.Resources.ImportedFile);
                    m_FileIcons.Images.Add("?Root", CsLib.Properties.Resources.CD);
                    m_FileIcons.Images.Add("?Boot", CsLib.Properties.Resources.BootableDisc);

                    try
                    {
                        m_FileIcons.Images[0] = IconReader.GetFolderIcon(IconReader.IconSize.Small, IconReader.FolderType.Closed).ToBitmap();
                        m_FileIcons.Images[1] = IconReader.GetFolderIcon(IconReader.IconSize.Small, IconReader.FolderType.Open).ToBitmap();
                        // Use the existing icons, and make them red
                        Bitmap importOpen = new Bitmap(m_FileIcons.Images["?FolderOpen"].Width, m_FileIcons.Images["?FolderOpen"].Height, m_FileIcons.Images["?FolderOpen"].PixelFormat);
                        Bitmap importClosed = new Bitmap(m_FileIcons.Images["?Folder"].Width, m_FileIcons.Images["?Folder"].Height, m_FileIcons.Images["?Folder"].PixelFormat);

                        Rectangle rctPicture = new Rectangle(0, 0, importOpen.Width, importOpen.Height);
                        float[][] ptsArray = { new float[] { 1, 0, 0, 0, 0 }, new float[] { 0, 0.5f, 0, 0, 0 }, new float[] { 0, 0, 0.5f, 0, 0 }, new float[] { 0, 0, 0, 1, 0 }, new float[] { 0, 0, 0, 0, 1 } };
                        System.Drawing.Imaging.ColorMatrix clrMatrix = new System.Drawing.Imaging.ColorMatrix(ptsArray);
                        System.Drawing.Imaging.ImageAttributes imgAttrs = new System.Drawing.Imaging.ImageAttributes();
                        imgAttrs.SetColorMatrix(clrMatrix);

                        using (Graphics gr = Graphics.FromImage(importOpen))
                        {
                            gr.DrawImage(m_FileIcons.Images["?FolderOpen"], rctPicture, 0, 0, importOpen.Width, importOpen.Height, GraphicsUnit.Pixel, imgAttrs);
                        }
                        using (Graphics gr = Graphics.FromImage(importClosed))
                        {
                            gr.DrawImage(m_FileIcons.Images["?Folder"], rctPicture, 0, 0, importOpen.Width, importOpen.Height, GraphicsUnit.Pixel, imgAttrs);
                        }

                        imgAttrs.Dispose();
                        m_FileIcons.Images.Add("?FolderImport", importOpen);
                        m_FileIcons.Images.Add("?FolderOpenImport", importClosed);
                    }
                    catch
                    {
                        // if it doesn't work, we simply get the default icons automatically
                    }

                    // Invert icons so that they are drawn correctly for RTL cultures
                    if (BaseForm.IsRightToLeftCulture)
                    {
                        foreach (string key in m_FileIcons.Images.Keys)
                        {
                            Image img = RtlHelper.FlipIfNecessary(m_FileIcons.Images[key], BaseForm.IsRightToLeftCulture);
                            m_FileIcons.Images.RemoveByKey(key);
                            m_FileIcons.Images.Add(key, img);
                        }
                    }
                }
                return m_FileIcons;
            }
        }

        public static Dictionary<UIText, string> Translation
        {
            get
            {
                // Init?
                if (m_Translations.Count == 0)
                {
                    m_Translations[UIText.BlankDirectoryOrFile] = "File or directory names can not be blank.";
                    m_Translations[UIText.InvalidFileName] = "The entry contains invalid characters and cannot be renamed.";
                    m_Translations[UIText.InvalidLetters] = "Do not use these letters: {0}";
                    m_Translations[UIText.CouldNotRenameAlreadyExists] = "The entry could not be renamed, because an entry with the same name already exists.";
                    m_Translations[UIText.File] = "File";
                    m_Translations[UIText.Folder] = "Folder";
                    m_Translations[UIText.Name] = "Name";
                    m_Translations[UIText.Size] = "Size";
                    m_Translations[UIText.Date] = "Date";
                    m_Translations[UIText.Type] = "Type";
                    m_Translations[UIText.NewFolder] = "New folder";
                    m_Translations[UIText.Caching] = "Caching {0}";
                    m_Translations[UIText.AddFilesHere] = "Drag and drop files here or use the \"Add\" button to add files to the compilation.";
                }
                return m_Translations;
            }
        }

        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public IDataImageX DataImage
        {
            get
            {
                return m_DataImage;
            }
            set
            {
                m_DataImage = value;
                if (m_DataImage != null) {
                    m_DataImage.GetVolumeInfo().VolumeID = LayoutName;
                }
            }
        }

        public ManagedFileListView ListView
        {
            get
            {
                return m_ListView;
            }
            set
            {
                if (m_ListView != value)
                {
                    if (m_ListView != null)
                    {
                        m_ListView.Manager = null;
                    }
                    m_ListView = value;
                    m_ListView.Manager = this;
                }
            }
        }

        public bool IsDataImageEmpty
        {
            get
            {
                return m_DataImage.GetFileLayout().GetFileCount() == 0 && m_DataImage.GetFileLayout().GetFolderCount() == 0;
            }
        }

        /// <summary>
        /// The layout name determines the name of the root node
        /// in the TreeView.
        /// </summary>
        public override string LayoutName
        {
            get
            {
                return m_LayoutName;
            }
            set
            {
                if (value == null) {
                    throw new ArgumentNullException("value");
                }

                if (value.Length > 32)
                {
                    value = value.Substring(0, 32);
                }

                if (m_LayoutName != value)
                {
                    m_LayoutName = value;
                    if (m_TreeView != null)
                    {
                        m_TreeView.Invoke((MethodInvoker)delegate()
                        {
                            m_TreeView.Nodes[0].Text = value;
                        });
                    }
                    OnCompilationChanged(false, true);
                }
                // Always set: After loading a new compilation, layout name may 
                // be the same, but volume info needs to be updated.
                if (m_DataImage != null)
                {
                    m_DataImage.GetVolumeInfo().VolumeID = value;
                }
            }
        }

        public ManagedDirectoryTreeView TreeView
        {
            get
            {
                return m_TreeView;
            }
            set
            {
                if (m_TreeView != value)
                {
                    if (m_TreeView != null)
                    {
                        m_TreeView.Manager = null;
                    }
                    m_TreeView = value;
                    if (m_TreeView != null)
                    {
                        m_TreeView.Nodes.Clear();
                        
                        if (DesignMode) return;
                        
                        m_TreeView.Nodes.Add(m_LayoutName);
                        m_TreeView.Nodes[0].Tag = m_Root;
                        m_TreeView.Nodes[0].ImageKey = "?Root";
                        m_TreeView.Nodes[0].SelectedImageKey = "?Root";
                        m_TreeView.Manager = this;
                    }
                }
            }
        }

        internal DirectoryNode SelectedDirectory
        {
            get
            {
                if (TreeView.SelectedNode == null)
                {
                    return m_Root;
                }
                else {
                    return TreeView.SelectedNode.Tag as DirectoryNode;
                }
            }
        }

        #endregion


        /// <summary>
        /// Can be used to fire the event if the DataImage has
        /// been modified directly.
        /// </summary>
        /// <param name="contentsChanged"></param>
        public void SetCompilationChanged(bool contentsChanged) {
            OnCompilationChanged(contentsChanged, true);
        }

        /// <summary>
        /// If the boot settings have changed, use a different icon.
        /// The LayoutManager cannot watch such changes.
        /// </summary>
        public void UpdateRootIcon()
        {
            if (DataImage == null) return;
            string key = DataImage.GetBootableMode() ? "?Boot" : "?Root";
            m_TreeView.Nodes[0].SelectedImageKey = key;
            m_TreeView.Nodes[0].ImageKey = key;
        }

        /// <summary>
        /// Creates a new directory and "asks" for a name.
        /// </summary>
        public void CreateNewDirectory()
        {
            if (TreeView.Focused)
            {
                TreeNode parent = TreeView.GetNode(SelectedDirectory.FullVirtualPath);
                TreeNode newNode = parent.Nodes.Add(m_Translations[UIText.NewFolder]);
                DirectoryNode temp = new DirectoryNode(this, SelectedDirectory, m_Translations[UIText.NewFolder]);
                temp.VirtualDirectory = true;
                newNode.Tag = temp;
                newNode.EnsureVisible();
                newNode.BeginEdit();
            }
            else
            {
                ArrayList list = ListView.Objects.Clone() as ArrayList;
                DirectoryNode temp = new DirectoryNode(this, SelectedDirectory, m_Translations[UIText.NewFolder]);
                temp.VirtualDirectory = true;
                list.Add(temp);
                ListView.SetObjects(list);
                // Find the new index
                for (int i = 0; i < ListView.Items.Count; i++)
                {
                    if (ListView.GetModelObject(i) == temp)
                    {
                        ListView.EnsureVisible(i);
                        ListView.EditSubItem(ListView.GetItem(i), 0);
                        break;
                    }
                }
                if (!ListView.IsCellEditing)
                {
                    // Apparently, initiating the edit did not work.
                    // Make sure, that the folder disappears again.
                    TreeView.RefreshDirContents();
                }
            }
        }

        /// <summary>
        /// Adds any number of paths (files or directories) to the layout
        /// and shows a dialog during the process.
        /// </summary>
        /// <returns>List of errors (non-readable files etc.)</returns>
        public ItemAddError[] BeginAddPaths(string destVirtual, string[] sourceReal, Form ownerForm)
        {
            if (SpecialFileHandler != null && SpecialFileHandler(sourceReal))
            {
                // File wil be handled elsewhere
                return new ItemAddError[0];
            }

            using (ProgressDialog dialog = new ProgressDialog())
            {
                dialog.OnDoWork = delegate()
                {
                    return AddPaths(destVirtual, sourceReal);
                };
                dialog.OnCancel = delegate()
                {
                    this.CancelAddingFiles();
                };
                dialog.OnUpdateStatus = delegate()
                {
                    return CurrentProgressString;
                };

                // Show the necessary dialogs
                Form owner = ownerForm == null ? BaseForm.GetParentForm(m_TreeView, m_ListView) : ownerForm;
                dialog.ShowDialog(owner);
                FileAddErrorsDialog.Show(owner, dialog.Result as ItemAddError[]);

                if (TreeView != null)
                {
                    TreeView.Nodes[0].Expand();
                    TreeView.RefreshDirContents();
                }

                return dialog.Result as ItemAddError[];
            }
        }

        /// <summary>
        /// Adds any number of paths (files or directories) to the layout
        /// and shows a dialog during the process.
        /// </summary>
        /// <returns>List of errors (non-readable files etc.)</returns>
        public ItemAddError[] BeginAddPaths(string destVirtual, string[] sourceReal)
        {
            return BeginAddPaths(destVirtual, sourceReal, null);
        }

        /// <summary>
        /// Adds any number of paths (files or directories) to the layout.
        /// </summary>
        /// <returns>List of errors (non-readable files etc.)</returns>
        private ItemAddError[] AddPaths(string destVirtual, string[] sourceReal)
        {
            CheckPath(destVirtual, "destVirtual");

            StartReplace();
            m_CurrentProgressString = string.Empty;
            DirectoryNode node = GetDirectoryNodeOrCreate(destVirtual);

            // Resolve drive letters
            List<string> sourceList = new List<string>(sourceReal);
            // Filter empty sources
            sourceList.Remove(string.Empty);

            for (int i = 0; i < sourceReal.Length; i++)
            {
                try
                {
                    string dirName = Path.GetFileName(sourceReal[i]);
                    if (string.IsNullOrEmpty(dirName) && sourceReal[i].Contains(Path.VolumeSeparatorChar.ToString()))
                    {
                        sourceList.AddRange(Directory.GetFileSystemEntries(PathEx.QualifyPath(sourceReal[i])));
                        sourceList.Remove(sourceReal[i]);
                    }
                }
                catch (Exception ex)
                {
                    return new ItemAddError[] { new ItemAddError(sourceReal[i], ex) };
                }
            }
            sourceReal = sourceList.ToArray();

            // Check if we need to apply caching
            if (ApplicationSettings.DataCacheFilesAutomatically || ApplicationSettings.DataCacheNetworkFilesAutomatically)
            {
                EventHandler<GenericEventArgs<string>> eventHandler = delegate(object sender, GenericEventArgs<string> e)
                {
                    m_CurrentProgressString = string.Format(Translation[UIText.Caching], Path.GetFileName(e.Value));
                };

                try
                {
                    PathEx.FileCopying += eventHandler;
                    PathEx.CancelCopy = false;

                    for (int i = 0; i < sourceReal.Length; i++)
                    {
                        if (m_CancelAddingFiles) return new ItemAddError[0];

                        if (ApplicationSettings.DataCacheFilesAutomatically && PathEx.IsRemovableSource(sourceReal[i]))
                        {
                            sourceReal[i] = PathEx.CacheLocally(sourceReal[i]);
                        }
                        else if (ApplicationSettings.DataCacheNetworkFilesAutomatically && PathEx.IsNetworkLocation(sourceReal[i]))
                        {
                            sourceReal[i] = PathEx.CacheLocally(sourceReal[i]);
                        }
                    }
                }
                finally
                {
                    PathEx.FileCopying -= eventHandler;
                }
            }

            ItemAddError[] result = node.AddPaths(sourceReal);

            if (m_TreeView != null)
            {
                m_TreeView.UpdateNode(destVirtual);
            }

            OnCompilationChanged();
            return result;
        }


        private void CancelUpdatingCompilation()
        {
            m_CancelUpdatingCompilation = true;
        }

        public override void Clear()
        {
            Clear(null);
        }

        public override void Clear(string newName)
        {
            if (SynchronizedObject != null && SynchronizedObject.InvokeRequired)
            {
                SynchronizedObject.Invoke(new MethodInvoker(Clear), null);
                return;
            }

            if (m_DataImage != null)
            {
                bool useJoliet = m_DataImage.UseLongJolietNames;
                eVolumeTypeX volumeType = m_DataImage.VolumeType;
                eISOLevelX isoLevel = m_DataImage.ISOLevel;
                string imageName = m_DataImage.GetVolumeInfo().ApplicationID;
                m_DataImage.GetFileLayout().ResetLayout();
                // These settings should not be changed in this step
                m_DataImage.UseLongJolietNames = useJoliet;
                m_DataImage.VolumeType = volumeType;
                m_DataImage.ISOLevel = isoLevel;
                m_DataImage.GetVolumeInfo().ApplicationID = imageName;
            }
            if (m_TreeView != null)
            {
                m_TreeView.Nodes[0].Nodes.Clear();
                UpdateRootIcon();
            }
            if (m_ListView != null)
            {
                m_ListView.SetObjects(null);
            }
            m_Root.Directories.Clear();
            m_Root.Files.Clear();

            m_SavePath = null;
            if (newName != null)
            {
                m_SuppressChangedEvent = true;
                try
                {
                    LayoutName = newName;
                }
                finally
                {
                    m_SuppressChangedEvent = false;
                }
            }

            OnCompilationChanged(true, false);

            System.Threading.Thread thread = new System.Threading.Thread(delegate()
            {
                PathEx.TryDeleteDirectory(PathEx.RemoveableMediaCache);
            });
            thread.IsBackground = true;
            thread.Start();
        }

        public bool AddDVDVideoLayout(string path)
        {
            BurningEngine.NmsEngine.OnItemAdded += new _IDVDEngineXEvents_OnItemAddedEventHandler(OnVideoItemAdded);
            try
            {
                return m_DataImage.AddDVDVideoLayout(path);
            }
            finally
            {
                BurningEngine.NmsEngine.OnItemAdded -= new _IDVDEngineXEvents_OnItemAddedEventHandler(OnVideoItemAdded);
                if (TreeView != null)
                {
                    TreeView.RefreshDirContents();
                }
            }
        }

        void OnVideoItemAdded(int lImageID, ref DataItemX objDataItem)
        {
            if (objDataItem.IsFolder())
            {
                DirectoryNode dirNode = GetDirectoryNodeOrCreate(objDataItem.GetImagePath());
            }
            else
            {
                string path = Path.GetDirectoryName(objDataItem.GetImagePath());
                string fileKey = Path.GetFileName(objDataItem.GetImagePath()).ToLower();
                DirectoryNode dirNode = GetDirectoryNodeOrCreate(path);
                FileNode newFile = new FileNode(objDataItem.GetImageName(), dirNode, false);
                newFile.FileSystemPath = objDataItem.GetSourceFilePath();
                dirNode.Files[fileKey] = newFile;
            }
        }

        /// <summary>
        /// Imports the last session from the given device.
        /// </summary>
        public bool ImportFromDisc(DeviceX device)
        {
            return ImportFromDisc(device, 0);
        }

        /// <summary>
        /// Imports the specified session to the data layout.
        /// </summary>
        public bool ImportFromDisc(DeviceX device, short number)
        {
            BurningEngine.NmsEngine.OnItemAdded += new _IDVDEngineXEvents_OnItemAddedEventHandler(OnItemImported);
            _IDVDEngineXEvents_OnItemAddedEventHandler progressDelegate = null;
            m_ImportDoneSize = 0;
            m_ImportTotalSize = 0;

            try
            {
                m_SuppressChangedEvent = true;
                
                // Find out how much data we have to import
                TrackInfoX trackInfo = null;
                DiscInfoX discInfo = device.GetDiscInfo();
                if (device.ReadTrackInfo(ref trackInfo, (number > 0) ? number : BurningEngine.GetLastTrackNumber(discInfo)))
                {
                    m_ImportTotalSize = trackInfo.GetSizeBlocks() * m_DataImage.GetBlockSize();
                }

                // Bring up the progress dialog again
                using (ProgressDialog dialog = new ProgressDialog("Importing Session", "Importing session, please wait..."))
                {
                    dialog.OnDoWork = delegate()
                    {
                        Clear(); // cleaning up the layout before is required
                        return m_DataImage.ImportDiscSession(device, eVolumeTypeX.vtUnknown, number); // "unknown" means any
                    };
                    if (m_ImportTotalSize > 0)
                    {
                        // If progress tracking is possible, subscribe a useful event
                        progressDelegate = delegate(int lImageID, ref DataItemX objDataItem)
                        {
                            m_ImportDoneSize += objDataItem.GetFileSize();
                            dialog.ReportProgress(Convert.ToInt32(m_ImportDoneSize / m_ImportTotalSize * 100.0));
                        };
                        BurningEngine.NmsEngine.OnItemAdded += progressDelegate;
                    }

                    dialog.ShowDialog(BaseForm.GetParentForm(TreeView, ListView));

                    if (TreeView != null) TreeView.Nodes[0].Expand();
                    LayoutName = m_DataImage.GetVolumeInfo().VolumeID;

                    return (bool)dialog.Result;
                }
            }
            finally
            {
                // Clean up events
                m_SuppressChangedEvent = false;
                BurningEngine.NmsEngine.OnItemAdded -= OnItemImported;
                if (progressDelegate != null)
                {
                    BurningEngine.NmsEngine.OnItemAdded -= progressDelegate;
                }
                if (TreeView != null) TreeView.RefreshDirContents();
                OnCompilationChanged(true, true);
            }
        }

        private void OnItemImported(int lImageID, ref DataItemX objDataItem)
        {
            if (objDataItem.IsFolder())
            {
                DirectoryNode dirNode = GetDirectoryNodeOrCreate(MakeDirPath(objDataItem), true);
                dirNode.ModificationDate = objDataItem.GetFileTime();
                dirNode.Hidden = objDataItem.GetIsHidden();
            }
            else
            {
                string path = Path.GetDirectoryName(objDataItem.GetImagePath());
                string fileKey = Path.GetFileName(objDataItem.GetImagePath()).ToLower();
                DirectoryNode dirNode = GetDirectoryNodeOrCreate(path, true);

                // A file may appear twice, if it has been overwritten
                // in an earlier session incorrectly
                if (!dirNode.Files.ContainsKey(fileKey))
                {
                    FileNode newFileNode = new FileNode(objDataItem.GetImageName(), dirNode, true);
                    // Caching required here for sorting
                    newFileNode.ModificationDate = objDataItem.GetFileTime();
                    newFileNode.Size = Convert.ToInt64(objDataItem.GetFileSize());
                    newFileNode.Hidden = objDataItem.GetIsHidden();

                    dirNode.Files.Add(fileKey, newFileNode);
                }
            }
        }

        /// <summary>
        /// GetImagePath() is all uppercase, so we need to combine it with GetImageName()
        /// </summary>
        private string MakeDirPath(DataItemX item)
        {
            return Path.Combine(Path.GetDirectoryName(item.GetImagePath()), item.GetImageName());
        }

        /// <summary>
        /// Moves a given object in the data image file
        /// system (and only there) from one location to the other.
        /// </summary>
        internal void MoveDataImagePath(IFileSystemNode from, DirectoryNode to)
        {
            string imgPath = GetDataImagePath(from.FullVirtualPath);
            IFileLayoutX layout = m_DataImage.GetFileLayout();

            if (from is FileNode)
            {
                if (imgPath != null) layout.RemoveItem(imgPath);
                if (!layout.AddItem(from.FileSystemPath, PathCombine(to.FullVirtualPath, from.Name)))
                {
                    throw new ApplicationException("Moving the file failed for an unknown reason.");
                }
            }
            else if (from is DirectoryNode)
            {
                if (imgPath == null)
                {
                    if (!layout.AddItem("", to.FullVirtualPath))
                    {
                        throw new ApplicationException("Moving the directory failed for an unknown reason.");
                    }
                }
                else
                {
                    CopyDataImageDir(layout.GetItem(imgPath), to.FullVirtualPath); 
                }
                // Last but not least
                m_DataImage.GetFileLayout().RemoveItem(imgPath);
            }
        }

        private void CopyDataImageDir(DataItemX source, string pathDest)
        {
            string newFolderPath = PathCombine(pathDest, source.GetImageName());
            if (!m_DataImage.GetFileLayout().AddItem("", newFolderPath))
            {
                throw new ApplicationException("Move operation failed for an unknown reason.");
            }

            for (int i = 0; i < source.GetSubItemCount(); ++i)
            {
                DataItemX current = source.GetSubItem(i);
                if (current.IsFolder())
                {
                    CopyDataImageDir(current, newFolderPath);
                }
                else
                {
                    if (!m_DataImage.GetFileLayout().AddItem(current.GetSourceFilePath(), PathCombine(newFolderPath, current.GetImageName())))
                    {
                        throw new ApplicationException("Adding the file failed for an unknown reason.");
                    }
                }
            }
        }

        public static string PathCombine(string a, string b)
        {
            return a.TrimEnd(Path.DirectorySeparatorChar) + Path.DirectorySeparatorChar + b;
        }

        public bool Rename(string existingPath, string newName)
        {
            return Rename(existingPath, newName, null);
        }

        /// <summary>
        /// Renames the given file or directory.
        /// </summary>
        /// <returns>true if rename was successful, false otherwise</returns>
        internal bool Rename(string existingPath, string newName, DirectoryNode existingNode)
        {
            IFileSystemNode node = (existingNode == null) ? GetNode(existingPath) : existingNode;
            if (node == null)
            {
                return false;
            }

            // No whitespace allowed at the end or beginning of file names
            newName = newName.Trim();
            string newNameKey = newName.ToLower();

            // Other sanity checks...
            if (PathEx.IsInvalidFileName(newName))
            {
                StringBuilder invalidChares = new StringBuilder();
                foreach (char c in Path.GetInvalidFileNameChars()) { if (char.IsSymbol(c) || char.IsPunctuation(c)) { invalidChares.Append(c); invalidChares.Append(" "); } }
                string content = Translation[UIText.InvalidLetters].Replace("{0}", invalidChares.ToString());
                ShowEditNodeErrorMessage(Translation[UIText.InvalidFileName], content, node);
                return false;
            }

            if (node == m_Root)
            {
                m_TreeView.Nodes[0].Text = newName;
                LayoutName = newName;
                if (m_DataImage != null)
                {
                    m_DataImage.GetVolumeInfo().VolumeID = newName;
                }
                return true;
            }

            if (newName.Length == 0)
            {
                ShowEditNodeErrorMessage(Translation[UIText.BlankDirectoryOrFile], null, node);
                return false;
            }

            DirectoryNode dirNode = node as DirectoryNode;

            if (dirNode != null)
            {
                TreeNode treeNode = m_TreeView.GetNode(node.FullVirtualPath);

                if (newNameKey == node.Name.ToLower())
                {
                    // If the same name, only different casing, less work
                    if (treeNode != null) treeNode.Text = newName;
                    node.Name = newName;
                    return true;
                }

                if (node.Parent != null)
                {
                    if (node.Parent.Directories.ContainsKey(newNameKey) || node.Parent.Files.ContainsKey(newNameKey))
                    {
                        ShowEditNodeErrorMessage(Translation[UIText.CouldNotRenameAlreadyExists], null, node);
                        return false;
                    }
                }

                // Only do that if a real and not virtual node has been renamed
                if (existingNode == null)
                {
                    treeNode.Text = newName;

                    if (m_DataImage != null)
                    {
                        // Bug in NMS: Renaming does not work in UDF images
                        // Do a move-operation instead.
                        if (m_DataImage.VolumeType == eVolumeTypeX.vtISO9660_Joliet_UDF || m_DataImage.VolumeType == eVolumeTypeX.vtISO9660_UDF || m_DataImage.VolumeType == eVolumeTypeX.vtUDF)
                        {
                            DirectoryNode targetNode = GetDirectoryNodeOrCreate(PathCombine(node.Parent.FullVirtualPath, newName));
                            foreach (DirectoryNode containedNode in dirNode.Directories.Values)
                            {
                                DirectoryNode newFolder = GetDirectoryNodeOrCreate(PathCombine(targetNode.FullVirtualPath, containedNode.Name));
                                newFolder.AddDirectory(containedNode);
                            }
                            foreach (FileNode containedNode in dirNode.Files.Values)
                            {
                                MoveDataImagePath(containedNode, targetNode);
                                targetNode.Files[containedNode.Name.ToLower()] = containedNode;
                            }

                            this.Delete(node.FullVirtualPath);

                            treeNode.Name = newNameKey;
                            m_TreeView.RefreshDirContents();
                        }
                        else
                        {
                            m_DataImage.GetFileLayout().RenameFolder(GetDataImagePath(node.FullVirtualPath), newName);
                            treeNode.Name = newNameKey;
                            node.Name = newName;
                        }
                    }

                    OnCompilationChanged(false, true);
                }
            }

            if (node is FileNode)
            {
                if (newNameKey == node.Name.ToLower())
                {
                    // If the same name, only different casing, less work
                    node.Name = newName;
                    return true;
                }

                if (node.Parent != null)
                {
                    if (node.Parent.Directories.ContainsKey(newNameKey) || node.Parent.Files.ContainsKey(newNameKey))
                    {
                        ShowEditNodeErrorMessage(Translation[UIText.CouldNotRenameAlreadyExists], null, node);
                        return false;
                    }
                }

                if (m_DataImage != null)
                {
                    // Bug in NMS: Renaming does not work in UDF images
                    // Do a move-operation instead.
                    if (m_DataImage.VolumeType == eVolumeTypeX.vtISO9660_Joliet_UDF || m_DataImage.VolumeType == eVolumeTypeX.vtISO9660_UDF || m_DataImage.VolumeType == eVolumeTypeX.vtUDF)
                    {
                        m_DataImage.GetFileLayout().RemoveItem(node.FullVirtualPath);
                        m_DataImage.GetFileLayout().AddItem(node.FileSystemPath, PathCombine(node.Parent.FullVirtualPath, newName));
                    }
                    else
                    {
                        m_DataImage.GetFileLayout().RenameFile(GetDataImagePath(node.FullVirtualPath), newName);
                    }
                }

                OnCompilationChanged(false, true);
                node.Name = newName;
            }

            return true;
        }

        private void ShowEditNodeErrorMessage(string message, string content, IFileSystemNode node)
        {
            Icon icon = null;
            if (node is FileNode)
            {
                icon = IconReader.GetFileIcon(node.FileSystemPath, IconReader.IconSize.Large, false);
            }
            else
            {
                icon = IconReader.GetFolderIcon(IconReader.IconSize.Large, IconReader.FolderType.Closed);
            }

            using (icon)
            {
                TaskDialog.ShowTaskDialogBox(m_ListView, Application.ProductName, message, content, "", "", "", "", "", eTaskDialogButtons.OK, eSysIcons.Information, eSysIcons.Information, icon);
            }
        }

        /// <summary>
        /// Determines the corresponding data item for
        /// the given file node.
        /// </summary>
        internal IDataItemX GetDataImageItem(IFileSystemNode node)
        {
            return GetDataImageItem(node.FullVirtualPath);
        }

        /// <summary>
        /// Determines the corresponding data item for
        /// the given virtual path.
        /// </summary>
        public IDataItemX GetDataImageItem(string virtualPath)
        {
            return GetDataImageItem(m_DataImage, virtualPath);
        }

        /// <summary>
        /// Determines the corresponding data item for
        /// the given virtual path.
        /// </summary>
        private static IDataItemX GetDataImageItem(IDataImageX dataImage, string virtualPath)
        {
            if (dataImage == null) return null;

            string[] parts = SplitPath(virtualPath);
            DataItemX current = dataImage.GetFileLayout().GetRoot();

            foreach (string part in parts)
            {
                bool found = false;

                for (int i = 0; i < current.GetSubItemCount(); ++i)
                {
                    string imgName = current.GetSubItem(i).GetImageName();
                    if (string.IsNullOrEmpty(imgName))
                    {
                        imgName = current.GetSubItem(i).GetImagePath().TrimStart('\\');
                    }

                    if (string.Equals(imgName, part, StringComparison.OrdinalIgnoreCase))
                    {
                        current = current.GetSubItem(i);
                        found = true;
                        break;
                    }
                }

                if (!found) return null;
            }

            return current;
        }

        /// <summary>
        /// Determines the corresponding path in the data image
        /// for a given virtual path.
        /// </summary>
        public string GetDataImagePath(string virtualPath)
        {
            IDataItemX dataItem = GetDataImageItem(virtualPath);
            if (dataItem == null) return null;
            return dataItem.GetImagePath();
        }

        /// <summary>
        /// Determines whether or not usable clipboard data is available.
        /// </summary>
        public static bool HasClipboardData()
        {
            return (SafeClipboard.GetData(DataFormats.FileDrop) as string[] != null);
        }

        /// <summary>
        /// Returns the node at the given path. Null if it does not exist
        /// </summary>
        internal IFileSystemNode GetNode(string destVirtual)
        {
            CheckPath(destVirtual, "destVirtual");

            string[] parts = SplitPath(destVirtual);
            DirectoryNode currentNode = m_Root;

            for (int i = 0; i < parts.Length; i++)
            {
                if (!currentNode.Directories.ContainsKey(parts[i]))
                {
                    // Last part could be a file
                    if (i == parts.Length - 1 && currentNode.Files.ContainsKey(parts[i]))
                    {
                        return currentNode.Files[parts[i]];
                    }
                    return null;
                }
                currentNode = currentNode.Directories[parts[i]];
            }

            return currentNode;
        }

        /// <summary>
        /// Deletes the specified path.
        /// </summary>
        public void Delete(string path)
        {
            Delete(new string[] { path });
        }

        /// <summary>
        /// Deletes the specified paths.
        /// </summary>
        public void Delete(string[] paths)
        {
            if (paths == null)
            {
                throw new ArgumentNullException("paths");
            }

            if (paths.Length == 0) return; // nothing to do

            // Bring to the right thread for TreeView updates
            if (SynchronizedObject != null && SynchronizedObject.InvokeRequired)
            {
                SynchronizedObject.Invoke(new MethodInvoker(delegate() { Delete(paths); }), null);
                return;
            }

            foreach (string path in paths)
            {
                string[] parts = SplitPath(path);
                // Update directory structure
                IFileSystemNode node = GetNode(path);
                if (node == null) continue;

                if (m_DataImage != null)
                {
                    m_DataImage.GetFileLayout().RemoveItem(GetDataImagePath(node.FullVirtualPath));
                }

                if (node is DirectoryNode && node.Remove())
                {
                    // Update treeview
                    if (m_TreeView != null)
                    {
                        TreeNode delNode = m_TreeView.GetNode(path);
                        if (delNode != null)
                        {
                            delNode.Remove();
                        }
                        // Update ListView
                        m_TreeView.RefreshDirContents();
                    }
                }

                if (node is FileNode)
                {
                    node.Remove();
                }
            }

            m_TreeView.RefreshDirContents();
            OnCompilationChanged();
        }

        /// <summary>
        /// Shows the files which need to be renamed when burning
        /// the disc due to file system limitations.
        /// </summary>
        /// <param name="owner">Parent window</param>
        /// <returns>false, if the user cancels the burning process upon review of the list</returns>
        public FileRenameRequirement[] CheckForFileNameChanges(IWin32Window owner)
        {
            if (ApplicationSettings.IgnoreFileRenameRequirements)
            {
                return null;
            }

            bool usesUdf = (DataImage.VolumeType == eVolumeTypeX.vtUDF || DataImage.VolumeType == eVolumeTypeX.vtISO9660_UDF || DataImage.VolumeType == eVolumeTypeX.vtISO9660_Joliet_UDF);
            bool usesJoliet = (DataImage.VolumeType == eVolumeTypeX.vtISO9660_Joliet || DataImage.VolumeType == eVolumeTypeX.vtJoliet);
            m_CancelFileLimitationSearch = false;

            List<FileRenameRequirement> renamedItems = new List<FileRenameRequirement>();

            using (ProgressDialog dialog = new ProgressDialog("Checking file names", "Analysing compilation for file rename requirements..."))
            {
                dialog.OnDoWork = delegate()
                {
                    FindRenamedFiles(usesUdf, usesJoliet, renamedItems, DataImage.GetFileLayout().GetRoot());
                    return null;
                };
                dialog.OnCancel = delegate()
                {
                    m_CancelFileLimitationSearch = true;
                };
                dialog.ShowDialog(owner);
            }

            // User does not want to know
            if (m_CancelFileLimitationSearch)
            {
                return null;
            }

            return renamedItems.ToArray();
        }

        /// <summary>
        /// Shows the files which need to be renamed when burning
        /// the disc due to file system limitations.
        /// </summary>
        /// <param name="owner">Parent window</param>
        /// <returns>false, if the user cancels the burning process upon review of the list</returns>
        public bool ShowFileNameChanges(IWin32Window owner)
        {
            FileRenameRequirement[] requirements = CheckForFileNameChanges(owner);

            if (requirements == null) return true;

            return FileNameLimitationsDialog.Show(owner, requirements, this);
        }

        /// <summary>
        /// Finds all files and folders within rootItem which will be renamed on the final disc.
        /// </summary>
        /// <param name="renamedItems">Reference to a list which contains the results</param>
        private void FindRenamedFiles(bool usesUdf, bool usesJoliet, List<FileRenameRequirement> renamedItems, IDataItemX rootItem)
        {
            for (int i = 0; i < rootItem.GetSubItemCount(); i++)
            {
                if (m_CancelFileLimitationSearch) return;

                IDataItemX item = rootItem.GetSubItem(i);

                if (usesUdf)
                {
                    if (item.GetUDFName() != item.GetImageName())
                    {
                        renamedItems.Add(new FileRenameRequirement(item, item.GetUDFName()));
                    }
                }
                else if (usesJoliet)
                {
                    if (item.GetJolietName() != item.GetImageName())
                    {
                        renamedItems.Add(new FileRenameRequirement(item, item.GetJolietName()));
                    }
                }
                else
                {
                    if (item.GetISOName() != item.GetImageName())
                    {
                        renamedItems.Add(new FileRenameRequirement(item, item.GetISOName()));
                    }
                }

                if (item.IsFolder())
                {
                    FindRenamedFiles(usesUdf, usesJoliet, renamedItems, item);
                }
            }
        }

        /// <summary>
        /// Shows the properties of the currently selected node.
        /// </summary>
        public void ShowProperties()
        {
            List<IFileSystemNode> selectedNodes = new List<IFileSystemNode>();

            if (TreeView != null && TreeView.Focused)
            {
                if (TreeView.SelectedNode != null) {
                    selectedNodes.Add(TreeView.SelectedNode.Tag as IFileSystemNode);
                }
            }
            else
            {
                foreach (IFileSystemNode node in ListView.SelectedObjects)
                {
                    selectedNodes.Add(node);
                }
            }

            // Nothing to show?
            if (selectedNodes.Count == 0) return;

            using (PropertyDialog dialog = new PropertyDialog())
            {
                dialog.FileNodes = selectedNodes.ToArray();
                if (dialog.ShowDialog(BaseForm.GetParentForm(TreeView, ListView)) == DialogResult.OK)
                {
                    TreeView.RefreshDirContents();
                }
            }
        }

        /// <summary>
        /// Resets the remembered file-replace choice.
        /// </summary>
        internal void StartReplace()
        {
            m_ReplaceAction = RememberReplaceAction.None;
            m_CancelAddingFiles = false;
        }

        private delegate ReplaceAction RequestReplaceDelegate(FileNode oldNode, FileNode newNode);

        /// <summary>
        /// If a file already exists, the result of this function
        /// determines if the file is replaced or not.
        /// </summary>
        /// <remarks>Fires an event, which can be used to display a dialog
        /// and ask the user whether or not to replace.</remarks>
        internal ReplaceAction RequestReplace(FileNode oldNode, FileNode newNode)
        {
            // Determine file differences
            bool hasDifference = false;
            string changeInfoText = "";
            if (oldNode.Size != newNode.Size)
            {
                string formatText = (newNode.Size > oldNode.Size) ? "Size: {0} (larger)" : "Size: {0} (smaller)";
                changeInfoText += "\n" + BaseForm.FormatTranslation(formatText, FormatFileSize.Format(newNode.Size));
                hasDifference = true;
            }
            // Consider dates equal if they either match the actual time or if the UTC time matches
            if (Math.Abs((oldNode.ModificationDate - newNode.ModificationDate).TotalSeconds) >= 1 && Math.Abs((oldNode.ModificationDate.ToUniversalTime() - newNode.ModificationDate).TotalSeconds) >= 1)
            {
                string formatText = (newNode.Size > oldNode.Size) ? "Date: {0} (newer)" : "Date: {0} (older)";
                changeInfoText += "\n" + BaseForm.FormatTranslation(formatText, newNode.ModificationDate.ToString());
                hasDifference = true;
            }
            if (string.IsNullOrEmpty(changeInfoText))
            {
                changeInfoText = "\n" + BaseForm.RM.GetString("No differences, file appears to be the same.");
            }
            else
            {
                changeInfoText = "\n" + BaseForm.RM.GetString("Differences:") + changeInfoText;
            }

            // Remember choice?
            switch (m_ReplaceAction)
            {
                case RememberReplaceAction.AllChanged:
                    return (hasDifference ? ReplaceAction.Replace : ReplaceAction.Keep);
                case RememberReplaceAction.NoAll:
                    return ReplaceAction.Keep;
                case RememberReplaceAction.Cancel:
                    return ReplaceAction.Cancel;
                case RememberReplaceAction.YesAll:
                    return ReplaceAction.Replace;
            }

            // Try to find a suitable parent form for the dialog
            Form parentForm = BaseForm.GetParentForm(TreeView, ListView);
            if (parentForm != null && parentForm.InvokeRequired)
            {
                return (ReplaceAction)parentForm.Invoke(new RequestReplaceDelegate(RequestReplace), oldNode, newNode);
            }

            // Get file icon
            Icon fileIcon = null;
            try
            {
                fileIcon = IconReader.GetFileIcon(newNode.FileSystemPath, IconReader.IconSize.Large, false);
            }
            catch { }

            TaskDialogResult dialogRes = TaskDialog.ShowTaskDialogBox(parentForm, Application.ProductName,
                BaseForm.FormatTranslation("The compilation already contains a file named '{0}'.", oldNode.Name), "", "",
              "", BaseForm.RM.GetString("Do this for &all following conflicts"),
              "", BaseForm.RM.GetString("&Replace file") + "|" + BaseForm.RM.GetString("Replace file if &changed") + changeInfoText + "|" + BaseForm.RM.GetString("Do&n't replace file"), eTaskDialogButtons.Cancel,
              eSysIcons.Question,
              eSysIcons.Information, fileIcon);

            if (dialogRes.DialogResult != DialogResult.Cancel)
            {
                if (dialogRes.SelectedCommand == 0)
                {
                    if (dialogRes.VerificationChecked)
                    {
                        m_ReplaceAction = RememberReplaceAction.YesAll;
                    }
                    return ReplaceAction.Replace;
                }
                else if (dialogRes.SelectedCommand == 1)
                {
                    if (dialogRes.VerificationChecked)
                    {
                        m_ReplaceAction = RememberReplaceAction.AllChanged;
                    }
                    return (hasDifference ? ReplaceAction.Replace : ReplaceAction.Keep);
                }
                else if (dialogRes.SelectedCommand == 2)
                {
                    if (dialogRes.VerificationChecked)
                    {
                        m_ReplaceAction = RememberReplaceAction.NoAll;
                    }
                    return ReplaceAction.Keep;
                }
            }

            m_ReplaceAction = RememberReplaceAction.Cancel;
            return ReplaceAction.Cancel;
        }

        /// <summary>
        /// Returns the specified directory node, or creates a new
        /// one if it doesn't exist
        /// </summary>
        /// <param name="destVirtual">Directory to create</param>
        private DirectoryNode GetDirectoryNodeOrCreate(string destVirtual)
        {
            return GetDirectoryNodeOrCreate(destVirtual, false);
        }

        /// <summary>
        /// Returns the specified directory node, or creates a new
        /// one if it doesn't exist
        /// </summary>
        /// <param name="destVirtual">Directory to create</param>
        private DirectoryNode GetDirectoryNodeOrCreate(string destVirtual, bool isImported)
        {
            CheckPath(destVirtual, "destVirtual");

            string[] partKeys = SplitPath(destVirtual);
            string[] parts = SplitPath(destVirtual, false);
            DirectoryNode currentNode = m_Root;
            bool added = false;
            string updatePath = string.Empty;

            for (int i = 0; i < partKeys.Length; i++)
            {
                if (!currentNode.Directories.ContainsKey(partKeys[i]))
                {
                    DirectoryNode newDirNode = new DirectoryNode(this, currentNode, parts[i]);
                    newDirNode.IsImported = isImported;
                    currentNode.Directories.Add(partKeys[i], newDirNode);
                    added = true;
                }
                else
                {
                    updatePath += partKeys[i] + Path.DirectorySeparatorChar;
                }
                currentNode = currentNode.Directories[partKeys[i]];
            }

            if (added)
            {
                if (m_DataImage != null && !isImported)
                {
                    m_DataImage.GetFileLayout().AddItem("", destVirtual);
                }
                if (m_TreeView != null)
                {
                    m_TreeView.UpdateNode(updatePath);
                }
                OnCompilationChanged(true, true);
            }

            return currentNode;
        }

        /// <summary>
        /// Creates a new virtual directory (recursively)
        /// at the specified location.
        /// </summary>
        /// <param name="destVirtual">Directory to create</param>
        public void CreateDirectory(string destVirtual)
        {
            GetDirectoryNodeOrCreate(destVirtual);
        }

        /// <summary>
        /// Splits a path into its folder parts (all lowercase).
        /// </summary>
        public static string[] SplitPath(string path)
        {
            return SplitPath(path, true);
        }

        /// <summary>
        /// Splits a path into its folder parts.
        /// </summary>
        public static string[] SplitPath(string path, bool lowercase)
        {
            path = path.Trim(Path.DirectorySeparatorChar);
            if (path == string.Empty) return new string[0];
            if (lowercase) path = path.ToLower();
            return path.Split(Path.DirectorySeparatorChar);
        }

        /// <summary>
        /// Checks whether or not a path is valid, and
        /// throws an exception if it is not.
        /// </summary>
        /// <param name="path"></param>
        protected void CheckPath(string path, string param) {
            if (path == null)
            {
                throw new ArgumentNullException(param);
            }

            char[] chars = Path.GetInvalidPathChars();
            foreach (char c in chars)
            {
                if (path.IndexOf(c) >= 0)
                {
                    throw new ArgumentException("The path is not valid.", param);
                }
            }
        }

        #region Compilation methods

        public DataCompilation GetDataCompilation()
        {
            DataCompilation comp = new DataCompilation(m_Root);
            comp.Name = LayoutName;
            comp.DataImage = m_DataImage;
            return comp;
        }

        public void LoadDataCompilation(string filename)
        {
            DataCompilation comp = new DataCompilation();

            using (ProgressDialog dialog = new ProgressDialog("Loading Compilation", "Loading your compilation, please wait..."))
            {
                comp.LoadProgressChanged += delegate(object sender, CompilationProgressEventArgs e)
                {
                    dialog.ReportProgress(e.Value);
                };
                dialog.OnDoWork = delegate()
                {
                    comp.Load(filename);
                    if (!dialog.Cancelled)
                    {
                        dialog.OnCancel = null;
                        Clear();
                        m_DataImage = comp.DataImage;
                        LayoutName = comp.Name;
                        m_Root.AddDirectory(comp.Layout);

                        if (m_TreeView != null && m_TreeView.InvokeRequired)
                        {
                            m_TreeView.Invoke(new MethodInvoker(delegate()
                            {
                                m_TreeView.Nodes[0].Tag = m_Root;
                                m_TreeView.UpdateNode(m_TreeView.Nodes[0]);
                                m_TreeView.Nodes[0].Expand();
                                m_TreeView.RefreshDirContents();
                                UpdateRootIcon();
                            }));
                        }

                        m_UnsavedChanges = false;
                        m_SavePath = null;
                    }
                    return null;
                };
                dialog.OnCancel = delegate()
                {
                    comp.Cancel();
                };
                dialog.ShowDialog(BaseForm.GetParentForm(TreeView, ListView));

                if (dialog.Error != null) throw dialog.Error;

                if (!dialog.Cancelled)
                {
                    FileAddErrorsDialog.Show(BaseForm.GetParentForm(TreeView, ListView), comp.Errors);
                    m_SavePath = filename;
                    OnCompilationChanged(true, false);
                }
            }
        }

        /// <summary>
        /// Saves the current compilation as ISO file to the given path.
        /// </summary>
        /// <param name="filename">Destination of the ISO file</param>
        /// <returns>true if the operation suceeded, false otherwise</returns>
        public bool SaveAsIso(string filename)
        {
            return SaveAsIso(BaseForm.GetParentForm(TreeView, ListView), m_DataImage, filename, "Creating ISO image, please wait...");
        }

        /// <summary>
        /// Saves the current compilation as ISO file to the given path.
        /// </summary>
        /// <param name="filename">Destination of the ISO file</param>
        /// <returns>true if the operation suceeded, false otherwise</returns>
        public static bool SaveAsIso(IWin32Window owner, IDataImageX m_DataImage, string filename, string workDescription)
        {
            bool cancelIsoCreation = false;
            bool errorOccured = false;
            bool doneSavingIso = false;

            long requiredSize = Convert.ToInt64(m_DataImage.GetImageSizeBytes());
            if (!DiskSpaceRequiredDialog.Show(owner, requiredSize, filename))
            {
                // not an error, user must have clicked "cancel"
                return true;
            }

            using (ProgressDialog dialog = new ProgressDialog("Creating ISO Image", workDescription))
            {
                // Attach necessary watching events
                _IDVDEngineXEvents_OnWriteStatusEventHandler statusDelegate = delegate(int lImageID, short nImageProgress, int lImageBlocksWritten, int lImageTotalBlocks, short nCurrentTrack, short nTrackCount, short nTrackProgress, int lTrackBlocksWritten, int lTrackTotalBlocks, eWriteStatusCodeX status, ref bool bCancel)
                {
                    if (cancelIsoCreation) bCancel = true;
                    if (status == eWriteStatusCodeX.wsWriteSuccess || status == eWriteStatusCodeX.wsWriteCancelled) doneSavingIso = true;
                    if (status == eWriteStatusCodeX.wsWriteError)
                    {
                        doneSavingIso = true;
                        errorOccured = true;
                    }
                    dialog.ReportProgress(nImageProgress);
                };
                _IDVDEngineXEvents_OnTrackFileStatusEventHandler trackStatusDelegate = new _IDVDEngineXEvents_OnTrackFileStatusEventHandler(delegate(int lImageID, int nTrackNumber, int lFileCount, int lFileIndex, string sFileName)
                {
                    dialog.StatusText = sFileName;
                });

                BurningEngine.NmsEngine.OnWriteStatus += statusDelegate;
                BurningEngine.NmsEngine.OnTrackFileStatus += trackStatusDelegate;

                // Actual operation
                try
                {
                    dialog.OnCancel = delegate() { cancelIsoCreation = true; };
                    dialog.OnDoWork = delegate()
                    {
                        bool res = m_DataImage.SaveImageFile(filename);
                        while (!doneSavingIso) System.Threading.Thread.Sleep(500);
                        return res && !errorOccured;
                    };
                    dialog.ShowDialog(owner);
                }
                finally
                {
                    BurningEngine.NmsEngine.OnWriteStatus -= statusDelegate;
                    BurningEngine.NmsEngine.OnTrackFileStatus -= new _IDVDEngineXEvents_OnTrackFileStatusEventHandler(trackStatusDelegate);
                }
                return (bool)dialog.Result;
            }
        }

        public void SaveDataCompilation(string filename)
        {
            using (ProgressDialog dialog = new ProgressDialog("Saving Compilation", "Saving your compilation, please wait..."))
            {
                DataCompilation comp = GetDataCompilation();
                comp.SaveProgressChanged += delegate(object sender, CompilationProgressEventArgs e)
                {
                    dialog.ReportProgress(e.Value);
                };
                dialog.OnDoWork = delegate()
                {
                    comp.Save(filename); return null;
                };
                dialog.ShowDialog(BaseForm.GetParentForm(TreeView, ListView));

                if (dialog.Error != null) throw dialog.Error;
            }

            // "Changed" only in the sense of "we are no longer unsaved"!
            OnCompilationChanged(false, false);
            m_SavePath = filename;
        }

        #endregion

        /// <summary>
        /// Goes through all directories of the current
        /// compilation and removes missing + adds new files.
        /// </summary>
        public void UpdateDataLayout()
        {
            m_CancelUpdatingCompilation = false;
            DataCompilation comp = GetDataCompilation();

            ProgressDialog dialog = new ProgressDialog("Updating Compilation", "Looking for changes, please wait...");
            dialog.OnDoWork = delegate()
            {
                return UpdateDataLayoutInternal(m_Root);
            };
            dialog.OnCancel = delegate()
            {
                this.CancelUpdatingCompilation();
            };

            dialog.ShowDialog(BaseForm.GetParentForm(m_TreeView, m_ListView));

            UpdateInstruction[] instructions = dialog.Result as UpdateInstruction[];

            // Already up-to-date?
            if (dialog.Cancelled || instructions == null || instructions.Length == 0) return;

            m_CancelUpdatingCompilation = false;

            // Done with finding the changes, now we need to apply them
            dialog = new ProgressDialog("Updating Compilation", "Updating your data compilation, please wait...");
            dialog.OnDoWork = delegate()
            {
                foreach (UpdateInstruction instruction in instructions)
                {
                    if (m_CancelUpdatingCompilation) return null;

                    if (instruction.Action == UpdateInstruction.ActionType.Add)
                    {
                        AddPaths(instruction.TargetPath, instruction.SourcePaths);
                    }
                    else if (instruction.Action == UpdateInstruction.ActionType.Remove)
                    {
                        Delete(instruction.TargetPath);
                    }
                    else if (instruction.Action == UpdateInstruction.ActionType.UpdateDate)
                    {
                        FileNode fileNode = GetNode(instruction.TargetPath) as FileNode;
                        if (fileNode == null) continue;

                        try
                        {
                            DateTime currentDate = File.GetLastWriteTime(instruction.SourcePaths[0]);
                            fileNode.ModificationDate = currentDate;
                        }
                        catch (Exception) { }
                    }
                }
                return null;
            };
            dialog.OnCancel = delegate()
            {
                this.CancelUpdatingCompilation();
                this.CancelAddingFiles();
            };
            dialog.OnUpdateStatus = delegate()
            {
                return CurrentProgressString;
            };

            dialog.ShowDialog(BaseForm.GetParentForm(m_TreeView, m_ListView));
            dialog.Dispose();

            m_TreeView.RefreshDirContents();
        }

        private UpdateInstruction[] UpdateDataLayoutInternal(DirectoryNode current)
        {
            List<UpdateInstruction> instructions = new List<UpdateInstruction>();

            if (m_CancelUpdatingCompilation) return instructions.ToArray();

            // Do we know the corresponding file system path at all? (root does not, manipulated data compilations might not)
            if (!string.IsNullOrEmpty(current.FileSystemPath))
            {
                try
                {
                    // Does the directory still exist?
                    if (!Directory.Exists(current.FileSystemPath))
                    {
                        instructions.Add(new UpdateInstruction(null, current.FullVirtualPath, UpdateInstruction.ActionType.Remove));
                        return instructions.ToArray();
                    }

                    List<string> files = new List<string>(Directory.GetFiles(current.FileSystemPath));
                    // Check new files
                    List<string> newFiles = new List<string>();
                    foreach (string file in files)
                    {
                        if (m_CancelUpdatingCompilation) return instructions.ToArray();

                        string filename = Path.GetFileName(file).ToLower();
                        if (!current.Files.ContainsKey(filename))
                        {
                            newFiles.Add(file);
                        }
                    }
                    if (newFiles.Count > 0)
                    {
                        instructions.Add(new UpdateInstruction(newFiles.ToArray(), current.FullVirtualPath, UpdateInstruction.ActionType.Add));
                    }

                    // Check removed files
                    foreach (FileNode file in current.Files.Values)
                    {
                        if (m_CancelUpdatingCompilation) return instructions.ToArray();

                        if (!files.Contains(file.FileSystemPath))
                        {
                            instructions.Add(new UpdateInstruction(null, file.FullVirtualPath, UpdateInstruction.ActionType.Remove));
                        }
                    }

                    // Check new directories
                    string[] directories = Directory.GetDirectories(current.FileSystemPath);
                    List<string> newDirs = new List<string>();
                    foreach (string dir in directories)
                    {
                        if (m_CancelUpdatingCompilation) return instructions.ToArray();

                        string dirName = Path.GetFileName(dir).ToLower();
                        if (!current.Directories.ContainsKey(dirName))
                        {
                            newDirs.Add(dir);
                        }
                    }
                    if (newDirs.Count > 0)
                    {
                        instructions.Add(new UpdateInstruction(newDirs.ToArray(), current.FullVirtualPath, UpdateInstruction.ActionType.Add));
                    }
                }
                catch (Exception)
                {
                    // If check fails for whatever reason, ignore and proceed.
                }
            }

            // Update file date if necessary
            foreach (FileNode file in current.Files.Values)
            {
                if (m_CancelUpdatingCompilation) return instructions.ToArray();

                if (string.IsNullOrEmpty(file.FileSystemPath)) continue;

                try
                {
                    DateTime currentDate = File.GetLastWriteTime(file.FileSystemPath);
                    if (currentDate != file.ModificationDate)
                    {
                        instructions.Add(new UpdateInstruction(new string[] { file.FileSystemPath }, file.FullVirtualPath, UpdateInstruction.ActionType.UpdateDate));
                    }
                }
                catch (Exception) { }
            }

            foreach (DirectoryNode dirNode in current.Directories.Values)
            {
                instructions.AddRange(UpdateDataLayoutInternal(dirNode));
            }

            return instructions.ToArray();
        }

        private string DebugImageContents(IDataItemX item)
        {
            StringBuilder sb = new StringBuilder();
            DebugImageContents(sb, item);
            return sb.ToString();
        }

        private void DebugImageContents(StringBuilder b, IDataItemX item)
        {
            for (int i = 0; i < item.GetSubItemCount(); i++)
            {
                b.Append(item.GetSubItem(i).GetImagePath());
                b.Append(Environment.NewLine);
                DebugImageContents(b, item.GetSubItem(i));
            }
        }

        public FileLayoutManager()
        {
            m_Root = new DirectoryNode(this, null, string.Empty);
            DataCompilation comp = new DataCompilation();
            m_DataImage = comp.DataImage;
        }
    }
}
