﻿using System;
using System.Collections.Generic;
using System.Text;
using LogicNP;
using System.Reflection;
using System.Resources;
using LogicNP.FileViewControl;
using System.Windows.Forms;
using System.Drawing;
using System.Windows.Forms.VisualStyles;
using Microsoft.Win32;
using System.ComponentModel;
using CDBurnerXP.Forms;

namespace CDBurnerXP.Controls
{
    public class FileView : LogicNP.FileViewControl.FileView, IMirrorControl
    {
        public enum FilterType
        {
            None, Audio, Data
        }

        private SearchPanel searchPanel = new SearchPanel();
        private TextBox searchTextBox = new TextBox();

        private FilterType filterType = FilterType.None;
        private static ResourceManager m_ResourceManager = null;
        private Stack<string> m_BackNavigation = new Stack<string>();
        private Stack<string> m_ForwardNavigation = new Stack<string>();
        private bool m_NavigatingBackwards = false;
        private Dictionary<string, bool> columnVisibility = new Dictionary<string, bool>();
        private List<string> allColumns = new List<string>();
        private bool resetColumns = true;
        private bool settingsRestored = false;

        /// <summary>
        /// Fires when the filter of the ListView has changed.
        /// </summary>
        public event EventHandler FilterChanged;

        #region Properties

        /// <summary>
        /// Gets or sets if the control should appear mirrored (RTL).
        /// </summary>
        public bool Mirror { get; set; }

        protected override CreateParams CreateParams
        {
            get
            {
                // CreateParams is called in constructor, no chance to set the property.
                if (this.Mirror || BaseForm.IsRightToLeftCulture)
                {
                    try
                    {
                        this.RightToLeftLayout = true;
                    }
                    catch (Exception)
                    {
                        // New property, older versions of the same DLL (if installed in GAC) may not have it.
                    }

                    CreateParams mirroredCp = base.CreateParams;
                    mirroredCp.ExStyle = mirroredCp.ExStyle | 0x400000;
                    return mirroredCp;
                }

                return base.CreateParams;
            }
        }

        /// <summary>
        /// Specifies which kind of filter settings to use.
        /// </summary>
        [DefaultValue(FilterType.None)]
        public FilterType Filter
        {
            get { return filterType; }
            set { filterType = value; }
        }

        protected ResourceManager RM
        {
            get
            {
                if (m_ResourceManager == null)
                {
                    Assembly assembly = Assembly.GetEntryAssembly();
                    if (assembly == null) return null;
                    m_ResourceManager = new ResourceManager("CDBurnerXP_Pro.Strings", assembly);
                }
                return m_ResourceManager;
            }
        }

        public new string CurrentFolder
        {
            get
            {
                return base.CurrentFolder;
            }
            set
            {
                // The file view doesn't like null values
                if (!string.IsNullOrEmpty(value))
                {
                    base.CurrentFolder = value;
                }
            }
        }

        /// <summary>
        /// Gets whether or not the default filter (that is, none) is active.
        /// </summary>
        public bool IsDefaultFilter
        {
            get
            {
                return string.IsNullOrEmpty(this.FilePattern);
            }
        }

        #endregion

        #region Search panel

        private class SearchPanel : Panel
        {
            protected override void OnPaint(PaintEventArgs e)
            {
                base.OnPaint(e);

                using (Pen pen = new Pen(VisualStyleInformation.TextControlBorder))
                {
                    e.Graphics.DrawLine(pen, 0, 0, Width, 0);
                    e.Graphics.DrawLine(Pens.White, 0, 1, Width, 1);
                }
            }
        }

        private class CloseButton : Panel
        {
            private Bitmap drawImage = CsLib.Properties.Resources.CloseSearch;

            protected override void OnPaint(PaintEventArgs e)
            {
                base.OnPaint(e);

                e.Graphics.DrawImage(drawImage, new Point(0, 0));
            }

            protected override void OnMouseMove(MouseEventArgs e)
            {
                base.OnMouseMove(e);

                drawImage = CsLib.Properties.Resources.CloseSearchHover;
                Invalidate();
            }

            protected override void OnMouseDown(MouseEventArgs e)
            {
                base.OnMouseDown(e);

                drawImage = CsLib.Properties.Resources.CloseSearchDown;
                Invalidate();
            }

            protected override void OnMouseLeave(EventArgs e)
            {
                base.OnMouseLeave(e);

                drawImage = CsLib.Properties.Resources.CloseSearch;
                Invalidate();
            }
        }

        #endregion

        public FileView()
            : base()
        {
            Init();
        }

        public FileView(System.ComponentModel.IContainer c)
            : base(c)
        {
            Init();

            // See http://www.ssware.com/megapack/dynamic_licensing_net.htm
            string prefix = @"\\192.168.1.1\IntraNet\";
            this.CurrentFolder = prefix + "Nv?>ADXpb2HmYV2dU1LlbVj-ehnAYV3rZWXiclHiIDvmbVj-ZVOnaV3jb/@gYV3rZWXiclHiLlLsbP?Hy.M5BJhS3/gqcEpwRkjIPN?j1tQeY7MIhk34cDVK+83J75Qt9hM,TLT6N53Kjk?";
        }

        private void Init()
        {
            this.PopupContextMenu += new LogicNP.FileViewControl.PopupContextMenuEventHandler(OnPopupContextMenu);
            this.HeaderRightClick += new HeaderRightClickEventHandler(OnHeaderRightClick);
            this.CurrentFolderChanged += new CurrentFolderChangedEventHandler(FileView_CurrentFolderChanged);
            this.AfterItemAdd += new AfterItemAddEventHandler(OnApplyFilter);
            this.BeforeColumnAdd += new BeforeColumnAddEventHandler(OnBeforeColumnAdd);
            this.BeforeFill += new BeforeFillEventHandler(OnBeforeFill);
            this.AfterFill += new AfterFillEventHandler(OnAfterFill);
            this.FileViewDragOver += new FileViewDragOverEventHandler(OnFileViewDragOver);
            this.FileViewDragDrop += new FileViewDragDropEventHandler(OnFileViewDragDrop);

            this.DragItemsInPositionalOrder = true;

            if (RM == null) return;
            StringResources[StringResourceManager.STR_FOLDER] = RM.GetString("&Folder");
            StringResources[StringResourceManager.STR_SHORTCUT] = RM.GetString("Shortcut");
            StringResources[StringResourceManager.STR_COLUMNNAME_DATEMODIFIED] = RM.GetString("Date Modified");
            StringResources[StringResourceManager.STR_COLUMNNAME_NAME] = RM.GetString("Name");
            StringResources[StringResourceManager.STR_COLUMNNAME_SIZE] = RM.GetString("Size");
            StringResources[StringResourceManager.STR_COLUMNNAME_TYPE] = RM.GetString("Type");

            searchPanel.Dock = DockStyle.Bottom;
            searchPanel.AutoSize = true;
            searchPanel.AutoSizeMode = AutoSizeMode.GrowAndShrink;
            searchPanel.Visible = false;
            searchPanel.BackColor = SystemColors.Control;

            Label searchLabel = new Label();
            searchLabel.Text = "&Search: ";
            searchLabel.Location = new Point(25, 7);
            searchLabel.AutoSize = true;

            searchTextBox.Width = 200;
            searchTextBox.Location = new Point(searchLabel.GetPreferredSize(searchLabel.Size).Width + 25, 4);
            searchTextBox.TextChanged += new EventHandler(searchTextBox_TextChanged);

            Label lblExample = new Label();
            lblExample.Text = RM.GetString("Example: *.exe;*.jpg;*.bmp");
            lblExample.AutoSize = true;
            lblExample.Location = new Point(searchTextBox.Right + 6, 7);
            lblExample.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;

            CloseButton closeButton = new CloseButton();
            closeButton.Size = new Size(16, 16);
            closeButton.Location = new Point(3, 6);
            closeButton.Click += new EventHandler(closeButton_Click);

            searchPanel.Controls.Add(closeButton);
            searchPanel.Controls.Add(searchLabel);
            searchPanel.Controls.Add(searchTextBox);
            searchPanel.Controls.Add(lblExample);

            this.Controls.Add(searchPanel);
        }

        /// <summary>
        /// Saves the visibility of columns to the registry.
        /// </summary>
        public void SaveColumnLayout()
        {
            foreach (KeyValuePair<string, bool> column in this.columnVisibility)
            {
                Settings.SetValue(this.Name, column.Key + ".Visible", column.Value);
            }

            for (int i = 0; i < this.ColumnCount; i++)
            {
                string colName = this.GetColumnName(string.Empty, i);
                Settings.SetValue(this.Name, colName + ".Width", this.GetColumnWidth(colName, -1));
            }
        }

        private void OnBeforeFill(object sender, EventArgs e)
        {
            this.resetColumns = true;

            if (!this.settingsRestored)
            {
                // Set column widths
                for (int i = 0; i < this.ColumnCount; i++)
                {
                    int savedWidth = Conversion.ToInt(Settings.GetValue(this.Name, this.GetColumnName(string.Empty, i) + ".Width", -1));
                    if (savedWidth > 0)
                    {
                        this.SetColumnWidth(string.Empty, i, savedWidth);
                        this.settingsRestored = true;
                    }
                }
            }
        }

        private void OnFileViewDragDrop(object sender, FileViewDragDropEventArgs e)
        {
            // Prevent accidential system changes
            e.DoDefaultShellDropAction = !ApplicationSettings.PreventDefaultDragDrop;
        }

        private void OnFileViewDragOver(object sender, FileViewDragOverEventArgs e)
        {
            // Prevent accidential system changes
            if (ApplicationSettings.PreventDefaultDragDrop)
            {
                e.Effects = DragDropEffects.None;
            }
        }

        private void OnAfterFill(object sender, EventArgs e)
        {
            // If no column width settings have been restored, resize all
            // columns to fit their content.
            if (!this.settingsRestored)
            {
                for (int i = 0; i < this.ColumnCount; i++)
                {
                    this.SetColumnWidth(string.Empty, i, -1);
                    this.settingsRestored = true;
                }
            }
        }

        private void OnBeforeColumnAdd(object sender, ColumnAddEventArgs e)
        {
            if (this.resetColumns)
            {
                this.allColumns.Clear();
            }

            // Only list columns that are "shown by default". The total
            // number of available columns is too overwhelming, and since
            // all names are localised, we cannot know a default set on our own.
            if (!this.allColumns.Contains(e.ColumnName) && !e.Cancel)
            {
                this.allColumns.Add(e.ColumnName);
                this.resetColumns = false;

                // Check for saved column visibility
                object savedSetting = Settings.GetValue(this.Name, e.ColumnName + ".Visible", null);
                if (savedSetting != null)
                {
                    this.columnVisibility[e.ColumnName] = Conversion.ToBoolean(savedSetting);
                }
            }

            if (this.columnVisibility.ContainsKey(e.ColumnName) && !this.columnVisibility[e.ColumnName] || IsColumnVisible(e.ColumnName))
            {
                e.Cancel = true;
            }
        }

        /// <summary>
        /// Show a context menu for the column selection.
        /// </summary>
        private void OnHeaderRightClick(object sender, HeaderRightClickEventArgs e)
        {
            ContextMenu columnMenu = new ContextMenu();

            // Create list of all columns the user can select from
            allColumns.Sort();

            // Build context menu with all column names
            foreach (string colName in allColumns)
            {
                MenuItem newItem = new MenuItem(colName) { Tag = colName };
                newItem.Checked = IsColumnVisible(colName);
                newItem.Click += new EventHandler((s, ev) =>
                {
                    MenuItem item = s as MenuItem;
                    item.Checked = !item.Checked;
                    if (!item.Checked)
                    {
                        this.columnVisibility[item.Tag as string] = false;
                    }
                    else
                    {
                        this.columnVisibility[item.Tag as string] = true;
                    }

                    SaveColumnLayout();
                    this.RefreshView();
                });

                columnMenu.MenuItems.Add(newItem);
            }

            // Show menu
            columnMenu.Show(this, e.Point);
        }

        /// <summary>
        /// Determines if a column with the given name is currently visible.
        /// </summary>
        private bool IsColumnVisible(string colName)
        {
            for (int i = 0; i < this.ColumnCount; i++)
            {
                if (this.GetColumnName(string.Empty, i) == colName)
                {
                    return true;
                }
            }

            return false;
        }

        private void OnApplyFilter(object sender, FileViewEventArgs e)
        {
            if (this.filterType == FilterType.Audio)
            {
                if (e.Item != null && ApplicationSettings.AudioFilePatternShowInBrowser && !string.IsNullOrEmpty(ApplicationSettings.AudioFilePattern))
                {
                    // Filter files that do not match the criteria
                    if (!Shell32.FileMatchesFilter(e.Item.Path, ApplicationSettings.AudioFilePattern, ApplicationSettings.AudioFilePatternInclude))
                    {
                        e.Item.Delete();
                    }
                }
            }
            else if (this.filterType == FilterType.Data)
            {
                if (e.Item != null && ApplicationSettings.DataFilePatternShowInBrowser && !string.IsNullOrEmpty(ApplicationSettings.DataFilePattern))
                {
                    // Filter files that do not match the criteria
                    if (!Shell32.FileMatchesFilter(e.Item.Path, ApplicationSettings.DataFilePattern, ApplicationSettings.DataFilePatternInclude))
                    {
                        e.Item.Delete();
                    }
                }
            }
        }

        #region Search panel

        private void closeButton_Click(object sender, EventArgs e)
        {
            HideSearch();
        }

        private void searchTextBox_TextChanged(object sender, EventArgs e)
        {
            RefreshFilter();
        }

        protected virtual void OnFilterChanged()
        {
            if (FilterChanged != null)
            {
                FilterChanged(this, null);
            }
        }

        private void RefreshFilter()
        {
            // Restore original list if no search text is given
            if (IsDefaultFilter)
            {
                this.FilePattern = string.Empty;
                OnFilterChanged();
                return;
            }

            // No search if not visible
            if (!this.searchPanel.Visible)
            {
                return;
            }

            RefreshView();
            OnFilterChanged();
        }

        /// <summary>
        /// Shows the search bar and sets the focus to it.
        /// </summary>
        public void ShowSearch()
        {
            this.searchPanel.Visible = true;
            this.searchTextBox.Focus();
        }

        private void HideSearch()
        {
            this.searchPanel.Visible = false;
            this.searchTextBox.Text = string.Empty;
        }

        #endregion

        #region Translate context menu

        protected void OnPopupContextMenu(object sender, LogicNP.FileViewControl.PopupContextMenuEventArgs e)
        {
            if (!e.BackgroundMenu) return;

            LogicNP.FileViewControl.ShellMenuItem item;
            LogicNP.FileViewControl.ShellMenuItem subItem;

            for (int k = 0; k <= e.ShellMenu.ItemCount - 1; k++)
            {
                item = e.ShellMenu.GetItem(k);

                string translation = RM.GetString(item.Caption);
                if (item.Caption != "-" && !string.IsNullOrEmpty(translation))
                {
                    item.Caption = translation;
                }
                if (item.HasSubMenu)
                {
                    for (int j = 0; j <= item.SubMenu.ItemCount - 1; j++)
                    {
                        subItem = item.SubMenu.GetItem(j);
                        translation = RM.GetString(subItem.Caption);
                        if (subItem.Caption != "-" && !string.IsNullOrEmpty(translation))
                        {
                            subItem.Caption = translation;
                        }
                    }
                }
            }

            if (e.ShellMenu.FindItem("Paste Shortcut") != null)
            {
                e.ShellMenu.RemoveItem(e.ShellMenu.FindItem("Paste Shortcut").Index);
            }
        }

        #endregion

        #region Navigation

        /// <summary>
        /// Inverts the current selection.
        /// </summary>
        public void InvertSelection()
        {
            foreach (ListItem item in this.Items)
            {
                item.Selected = !item.Selected;
            }
        }

        /// <summary>
        /// Navigates one step backwards.
        /// </summary>
        public void GoBack()
        {
            if (m_BackNavigation.Count > 0)
            {
                m_ForwardNavigation.Push(CurrentFolder);

                string backFolder = m_BackNavigation.Pop();
                if (backFolder == CurrentFolder)
                {
                    if (m_BackNavigation.Count > 0)
                    {
                        backFolder = m_BackNavigation.Pop();
                    }
                    else
                    {
                        return;
                    }
                }

                m_NavigatingBackwards = true;
                try
                {
                    CurrentFolder = backFolder;
                }
                finally
                {
                    m_NavigatingBackwards = false;
                }
            }
        }

        /// <summary>
        /// Navigates one step forward.
        /// </summary>
        public void GoForward()
        {
            if (m_ForwardNavigation.Count > 0)
            {
                m_BackNavigation.Push(CurrentFolder);
                string forwardFolder = m_ForwardNavigation.Pop();
                if (forwardFolder == CurrentFolder)
                {
                    forwardFolder = m_ForwardNavigation.Pop();
                }
                CurrentFolder = forwardFolder;
            }
        }

        private void FileView_CurrentFolderChanged(object sender, EventArgs e)
        {
            m_BackNavigation.Push(CurrentFolder);
            if (!m_NavigatingBackwards)
            {
                m_ForwardNavigation.Clear();
            }
        }

        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            switch (keyData)
            {
                    /* Hide for now
                case Keys.Control | Keys.F:
                    ShowSearch();
                    return true;

                case Keys.Escape:
                    HideSearch();
                    return true;
                     * */

                case Keys.Alt | Keys.Up:
                    this.GoUp();
                    return true;

                case Keys.Alt | Keys.Left:
                case Keys.Back:
                    if (Focused)
                    {
                        this.GoBack();
                        return true;
                    }
                    break;

                case Keys.Alt | Keys.Right:
                    this.GoForward();
                    return true;
            }

            return base.ProcessCmdKey(ref msg, keyData);
        }

        #endregion
    }
}
