﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Drawing;
using CDBurnerXP.IO;
using Microsoft.Win32;
using System.ComponentModel;
using CDBurnerXP.Drawing;
using CDBurnerXP.Forms;

namespace CDBurnerXP.Controls
{
    public class ManagedAudioListView : FastObjectListView
    {
        private ImageList m_ImageList = new ImageList();
        private Panel m_DragDropLine = null;
        private AudioLayoutManager m_Manager = null;
        private int m_DropIndex = -1;

        private OLVColumn m_ColumnTrackNo;
        private OLVColumn m_ColumnTitle;
        private OLVColumn m_ColumnArtist;
        private OLVColumn m_ColumnAlbum;
        private OLVColumn m_ColumnLength;
        private OLVColumn m_ColumnBitrate;
        private OLVColumn m_ColumnFileType;
        private OLVColumn m_ColumnTrackNoTags;
        private OLVColumn m_ColumnFullPath;
        private OLVColumn m_ColumnFileName;

        #region Properties

        internal AudioLayoutManager LayoutManager
        {
            set
            {
                m_Manager = value;
            }
        }

        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public new string EmptyListMsg
        {
            get
            {
                return base.EmptyListMsg;
            }
            set
            {
                base.EmptyListMsg = value;
            }
        }

        protected Panel DragDropLine
        {
            get
            {
                if (m_DragDropLine == null)
                {
                    Panel line = new Panel();
                    line.Visible = true;
                    line.Width = 100;
                    line.Height = 2;
                    line.BackColor = Color.Black;
                    // Whenever we move the line to a new location,
                    // we expect it to be visible (afterwards)
                    line.LocationChanged += delegate(object sender, EventArgs e)
                    {
                        line.Visible = true;
                    };

                    Controls.Add(line);
                    m_DragDropLine = line;
                }
                return m_DragDropLine;
            }
        }

        #endregion

        public ManagedAudioListView()
        {
            AllowDrop = true;
            HideSelection = false;
            ShowGroups = false;
            SmallImageList = m_ImageList;
            FullRowSelect = true;
            GridLines = true;
            AllColumns.Clear();
            ShowItemToolTips = true;

            // Name
            m_ColumnTrackNo = new OLVColumn();
            m_ColumnTrackNo.Name = "cTrackNo";
            m_ColumnTrackNo.Text = AudioLayoutManager.Translation[AudioLayoutManager.UIText.TrackNo];
            m_ColumnTrackNo.AspectGetter = ShowTrackNumber;
            m_ColumnTrackNo.ImageGetter = ShowFileIcon;
            m_ColumnTrackNo.IsEditable = false;
            AllColumns.Add(m_ColumnTrackNo);

            m_ColumnTitle = new OLVColumn();
            m_ColumnTitle.Name = "cTitle";
            m_ColumnTitle.Text = AudioLayoutManager.Translation[AudioLayoutManager.UIText.Title];
            m_ColumnTitle.AspectGetter = delegate(object x)
            {
                AudioTrack track = x as AudioTrack;
                return (string.IsNullOrEmpty(track.CDTextTitle) ? track.Title : track.CDTextTitle);
            };
            m_ColumnTitle.FreeSpaceProportion = 100;
            m_ColumnTitle.AspectPutter = delegate(object rowObject, object newValue)
            {
                AudioTrack track = rowObject as AudioTrack;
                track.CDTextTitle = newValue as string;
            };
            m_ColumnTitle.Width = 150;
            AllColumns.Add(m_ColumnTitle);

            m_ColumnArtist = new OLVColumn();
            m_ColumnArtist.Name = "cArtist";
            m_ColumnArtist.Text = AudioLayoutManager.Translation[AudioLayoutManager.UIText.Artist];
            m_ColumnArtist.FreeSpaceProportion = 100;
            m_ColumnArtist.Width = 150;
            m_ColumnArtist.AspectPutter = delegate(object rowObject, object newValue)
            {
                AudioTrack track = rowObject as AudioTrack;
                track.CDTextArtist = newValue as string;
            };
            m_ColumnArtist.AspectGetter = delegate(object x)
            {
                AudioTrack track = x as AudioTrack;
                return (string.IsNullOrEmpty(track.CDTextArtist) ? track.Artist : track.CDTextArtist);
            };
            AllColumns.Add(m_ColumnArtist);

            m_ColumnAlbum = new OLVColumn();
            m_ColumnAlbum.Name = "cAlbum";
            m_ColumnAlbum.Text = AudioLayoutManager.Translation[AudioLayoutManager.UIText.Album];
            m_ColumnAlbum.AspectName = "Album";
            m_ColumnAlbum.FreeSpaceProportion = 100;
            m_ColumnAlbum.Width = 150;
            m_ColumnAlbum.IsEditable = false;
            AllColumns.Add(m_ColumnAlbum);

            m_ColumnLength = new OLVColumn();
            m_ColumnLength.Name = "cLength";
            m_ColumnLength.TextAlign = HorizontalAlignment.Right;
            m_ColumnLength.Text = AudioLayoutManager.Translation[AudioLayoutManager.UIText.Length];
            m_ColumnLength.AspectGetter = ShowLength;
            m_ColumnLength.IsEditable = false;
            AllColumns.Add(m_ColumnLength);

            m_ColumnBitrate = new OLVColumn();
            m_ColumnBitrate.Name = "cBitrate";
            m_ColumnBitrate.TextAlign = HorizontalAlignment.Right;
            m_ColumnBitrate.Text = AudioLayoutManager.Translation[AudioLayoutManager.UIText.Bitrate];
            m_ColumnBitrate.AspectName = "Bitrate";
            m_ColumnBitrate.Width = 65;
            m_ColumnBitrate.IsEditable = false;
            m_ColumnBitrate.AspectToStringFormat = "{0:0} " + AudioLayoutManager.Translation[AudioLayoutManager.UIText.KBits];
            AllColumns.Add(m_ColumnBitrate);

            m_ColumnFileType = new OLVColumn();
            m_ColumnFileType.Name = "cFileType";
            m_ColumnFileType.Width = 50;
            m_ColumnFileType.TextAlign = HorizontalAlignment.Left;
            m_ColumnFileType.Text = AudioLayoutManager.Translation[AudioLayoutManager.UIText.Type];
            m_ColumnFileType.AspectName = "FileType";
            m_ColumnFileType.IsEditable = false;
            AllColumns.Add(m_ColumnFileType);

            m_ColumnTrackNoTags = new OLVColumn();
            m_ColumnTrackNoTags.Name = "cTrackNoByTags";
            m_ColumnTrackNoTags.Width = 60;
            m_ColumnTrackNoTags.IsVisible = false;
            m_ColumnTrackNoTags.TextAlign = HorizontalAlignment.Left;
            m_ColumnTrackNoTags.Text = AudioLayoutManager.Translation[AudioLayoutManager.UIText.TrackNoByTag];
            m_ColumnTrackNoTags.AspectName = "TrackNo";
            m_ColumnTrackNoTags.IsEditable = false;
            AllColumns.Add(m_ColumnTrackNoTags);

            m_ColumnFileName = new OLVColumn();
            m_ColumnFileName.Name = "cFileName";
            m_ColumnFileName.Width = 120;
            m_ColumnFileName.IsVisible = false;
            m_ColumnFileName.Text = AudioLayoutManager.Translation[AudioLayoutManager.UIText.FileName];
            m_ColumnFileName.AspectGetter = delegate(object x)
            {
                return Path.GetFileNameWithoutExtension(((AudioTrack)x).FileName);
            };
            m_ColumnFileName.IsEditable = false;
            AllColumns.Add(m_ColumnFileName);

            m_ColumnFullPath = new OLVColumn();
            m_ColumnFullPath.Name = "cFullPath";
            m_ColumnFullPath.Width = 150;
            m_ColumnFullPath.IsVisible = false;
            m_ColumnFullPath.Text = AudioLayoutManager.Translation[AudioLayoutManager.UIText.FullPath];
            m_ColumnFullPath.AspectName = "FileName";
            m_ColumnFullPath.IsEditable = false;
            AllColumns.Add(m_ColumnFullPath);

            RebuildColumns();

            EmptyListMsg = FileLayoutManager.Translation[FileLayoutManager.UIText.AddFilesHere];
        }

        #region Column Aspects

        public object ShowFileIcon(object x)
        {
            AudioTrack track = x as AudioTrack;
            ImageList.ImageCollection images = m_ImageList.Images;

            // Need to determine file icon?
            string ext = Path.GetExtension(track.FileName);
            if (!string.IsNullOrEmpty(ext))
            {
                try
                {
                    ext = ext.ToLower();
                    if (!images.ContainsKey(ext))
                    {
                        Icon myIcon = IconReader.GetFileIcon(track.FileName, IconReader.IconSize.Small, false);
                        images.Add(ext, RtlHelper.FlipIfNecessary(myIcon, BaseForm.IsRightToLeftCulture));
                    }
                    return images.IndexOfKey(ext);
                }
                catch { /* return defaults */ }
            }

            return 0;
        }

        private object ShowLength(object x)
        {
            AudioTrack track = x as AudioTrack;
            return new TimeSpan(track.Length.Hours, track.Length.Minutes, track.Length.Seconds);
        }

        private object ShowTrackNumber(object x)
        {
            return IndexOf(x) + 1; // track No. is not zero-based
        }

        #endregion

        #region Drag&Drop

        /// <summary>
        /// If multiple tracks are moved at the same time, the
        /// target position may not be within the selection.
        /// </summary>
        /// <param name="dataObject">Current drag'n'drop data object</param>
        /// <returns>true, if the move is not possible, false otherwise</returns>
        private bool IsForbiddenAudioTrackMove(IDataObject dataObject)
        {
            AudioTrack[] tracks = dataObject.GetData("AudioTrack[]") as AudioTrack[];
            if (tracks != null)
            {
                int firstItem = IndexOf(tracks[0]);
                int lastItem = IndexOf(tracks[tracks.Length - 1]);
                return (m_DropIndex >= firstItem && m_DropIndex <= lastItem + 1);
            }
            return false;
        }

        protected override void OnDragEnter(DragEventArgs drgevent)
        {
            base.OnDragEnter(drgevent);

            SetDragDropEffect(drgevent);
        }

        protected override void OnDragOver(DragEventArgs drgevent)
        {
            base.OnDragOver(drgevent);

            // Find out where to show the drop line
            Point mp = PointToClient(new Point(drgevent.X, drgevent.Y));

            if (GetItemAt(mp.X, mp.Y) == null)
            {
                if (Items.Count > 0)
                {
                    Rectangle posRect = GetItemRect(Items.Count - 1);
                    DragDropLine.Top = posRect.Top + posRect.Height;
                    m_DropIndex = -1;
                }
            }
            else
            {
                m_DropIndex = GetItemAt(mp.X, mp.Y).Index;
                Rectangle posRect = GetItemRect(m_DropIndex);
                DragDropLine.Top = posRect.Top;
                GetItemAt(mp.X, mp.Y).EnsureVisible();
            }

            SetDragDropEffect(drgevent);
        }

        private void SetDragDropEffect(DragEventArgs drgevent)
        {
            if (drgevent.Data.GetDataPresent(DataFormats.FileDrop))
            {
                drgevent.Effect = DragDropEffects.Copy;
            }
            else if (drgevent.Data.GetDataPresent("AudioTrack[]"))
            {
                if (IsForbiddenAudioTrackMove(drgevent.Data))
                {
                    drgevent.Effect = DragDropEffects.None;
                }
                else
                {
                    drgevent.Effect = DragDropEffects.Move;
                }
            }
        }

        protected override void OnItemDrag(ItemDragEventArgs e)
        {
            base.OnItemDrag(e);

            if (e.Button == MouseButtons.Left) {
                OLVListItem item = e.Item as OLVListItem;
                List<AudioTrack> audioTracks = new List<AudioTrack>();
                foreach (AudioTrack track in SelectedObjects) {
                    audioTracks.Add(track);
                }
                DoDragDrop(new DataObject("AudioTrack[]",  audioTracks.ToArray()), DragDropEffects.Move);
            }
        }

        protected override void OnDragLeave(EventArgs e)
        {
            base.OnDragLeave(e);

            DragDropLine.Visible = false;
        }

        protected override void OnDragDrop(DragEventArgs drgevent)
        {
            base.OnDragDrop(drgevent);

            // As soon as we force a custom order, sorting is gone.
            LastSortOrder = SortOrder.None;

            DragDropLine.Visible = false;

            string[] paths = drgevent.Data.GetData(DataFormats.FileDrop) as string[];
            if (paths != null)
            {
                m_Manager.BeginAddPaths(paths, m_DropIndex);
            }

            AudioTrack[] tracks = drgevent.Data.GetData("AudioTrack[]") as AudioTrack[];
            if (tracks != null)
            {
                m_Manager.MoveTracks(tracks, m_DropIndex);
            }
        }

        #endregion

        protected override void OnColumnClick(ColumnClickEventArgs e)
        {
            base.OnColumnClick(e);

            // We want it to be a "one time sort", otherwise it
            // is not possible to move items after sorting the list
            this.Sorting = SortOrder.None;

            OLVColumn col = GetColumn(e.Column);
            // This cannot be sorted actually. Restore the default order
            if (col.Name == "cTrackNo")
            {
                m_Manager.RefreshList();
                this.LastSortOrder = SortOrder.None;
            }
        }

        public void RemoveSelectedTracks()
        {
            m_Manager.RemoveTracks(GetSelectedTracks());
            DeselectAll();
        }

        public string[] GetSelectedTrackPaths()
        {
            List<string> tracks = new List<string>();
            foreach (AudioTrack track in SelectedObjects)
            {
                tracks.Add(track.FileName);
            }
            return tracks.ToArray();
        }

        protected AudioTrack[] GetSelectedTracks()
        {
            List<AudioTrack> tracks = new List<AudioTrack>();
            foreach (AudioTrack track in SelectedObjects)
            {
                tracks.Add(track);
            }
            return tracks.ToArray();
        }

        public void MoveSelectedTracks(bool up)
        {
            LastSortOrder = SortOrder.None;
            Sorting = SortOrder.None;

            // Calculate move-to index
            int upperIndex = -1;
            int lowerIndex = -1;
            if (SelectedIndices.Count > 0)
            {
                upperIndex = SelectedIndices[0];
                lowerIndex = SelectedIndices[SelectedIndices.Count - 1];
            }

            if (up && upperIndex > 0) m_Manager.MoveTracks(GetSelectedTracks(), upperIndex - 1);
            if (!up && lowerIndex > -1) m_Manager.MoveTracks(GetSelectedTracks(), lowerIndex + 2);
        }

        /// <summary>
        /// Determines whether or not clipboard data for
        /// pasting is available.
        /// </summary>
        /// <returns>Returns true if data is available, false otherwise.</returns>
        public bool CanPaste()
        {
            string[] files = SafeClipboard.GetData(DataFormats.FileDrop) as string[];
            return (files != null);
        }

        /// <summary>
        /// Pastes files copied to the clipboard into the list.
        /// </summary>
        /// <returns>Returns true if pasting was possible, false otherwise.</returns>
        public bool Paste()
        {
            string[] files = SafeClipboard.GetData(DataFormats.FileDrop) as string[];
            if (files != null)
            {
                m_Manager.BeginAddPaths(files, -1);
                return true;
            }
            return false;
        }

        /// <summary>
        /// Inverts the current selection.
        /// </summary>
        public void InvertSelection()
        {
            // Build a hashset of the currently selected indicies
            int[] selectedArray = new int[SelectedIndices.Count];
            SelectedIndices.CopyTo(selectedArray, 0);
            List<int> selected = new List<int>();
            selected.AddRange(selectedArray);

            // Reselect everything that wasn't selected before
            SelectedIndices.Clear();
            for (int i = 0; i < VirtualListSize; i++)
            {
                if (!selected.Contains(i))
                {
                    SelectedIndices.Add(i);
                }
            }
        }

        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            switch (keyData)
            {
                case Keys.Delete:
                    if (!this.IsCellEditing)
                    {
                        RemoveSelectedTracks();
                    }
                    break;
                case Keys.Control | Keys.Up:
                    MoveSelectedTracks(true);
                    return true;
                case Keys.Control | Keys.Down:
                    MoveSelectedTracks(false);
                    return true;
                case Keys.Control | Keys.V:
                    return Paste();
            }

            return base.ProcessCmdKey(ref msg, keyData);
        }
    }
}
