using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Globalization;
using System.Resources;
using System.Drawing;
using System.Diagnostics;
using CDBurnerXP.Burning;
using CDBurnerXP.Controls;
using CDBurnerXP.IO;
using NMSDVDXLib;
using Microsoft.Win32;
using Microsoft.Win32.Windows7;
using StarBurnX = RocketDivision.StarBurnX;
using System.Runtime.InteropServices;

namespace CDBurnerXP.Forms
{
    public /*abstract for our poor designer */ class BaseForm : PersistentForm
    {
        private static CultureInfo m_CultureInfo = null;
        private static ResourceManager m_ResourceManager = null;
        private delegate object MethodReturnInvoker();
        
        /// <summary>
        /// Defines whether or not CDBurnerXP forms should
        /// currently provide some kind of disc checking or
        /// auto importing of sessions.
        /// </summary>
        protected static bool m_AutoImportEnabled = true;

        private RemoveAutoPlay m_RemoveAutoPlay = null;

        #region Properties

        /// <summary>
        /// Gets or sets if the form should automatically apply RTL layout if suggested by the culture.
        /// </summary>
        protected virtual bool AutoMirrorEnabled
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Specifies, whether or not the form should
        /// automatially be translated.
        /// </summary>
        protected virtual bool AutoTranslate
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Gets whether or not the application should be exited
        /// when the dialog has been closed.
        /// </summary>
        public virtual bool ExitApplicationAfterClose
        {
            get
            {
                return false;
            }
        }

        public static bool IsRightToLeftCulture
        {
            get
            {
                return BaseForm.Culture == null ? false : BaseForm.Culture.TextInfo.IsRightToLeft;
            }
        }

        public static CultureInfo Culture
        {
            get {
                return m_CultureInfo;
            }
            set {
                m_CultureInfo = value;
            }
        }

        public static ResourceManager RM
        {
            get
            {
                if (m_ResourceManager == null)
                {
                    m_ResourceManager = GetResourceManager();
                }
                return m_ResourceManager;
            }
        }

        #endregion

        protected override void OnClosed(EventArgs e)
        {
            BurningEngine.NmsEngine.OnTrackFileWarning -= new _IDVDEngineXEvents_OnTrackFileWarningEventHandler(EngineOnTrackFileWarning);
            BurningEngine.NmsEngine.OnWriteError -= new _IDVDEngineXEvents_OnWriteErrorEventHandler(EngineOnWriteError);
            BurningEngine.NmsEngine.OnWriteStatus -= new _IDVDEngineXEvents_OnWriteStatusEventHandler(EngineOnWriteStatus);
            BurningEngine.NmsEngine.OnTrackFileStatus -= new _IDVDEngineXEvents_OnTrackFileStatusEventHandler(EngineOnTrackFileStatus);
            BurningEngine.NmsEngine.OnFileVerify -= new _IDVDEngineXEvents_OnFileVerifyEventHandler(EngineOnFileVerify);
            BurningEngine.NmsEngine.OnFileVerifyComplete -= new _IDVDEngineXEvents_OnFileVerifyCompleteEventHandler(EngineOnFileVerifyComplete);
            BurningEngine.NmsEngine.OnFileVerifyStart -= new _IDVDEngineXEvents_OnFileVerifyStartEventHandler(EngineOnFileVerifyStart);
            MediaNotifier.MediaInserted -= new MediaNotifier.MediaInsertedEventHandler(Notifier_MediaInserted);
            MediaNotifier.MediaRemoved -= new MediaNotifier.MediaRemovedEventHandler(Notifier_MediaRemoved);
            BurningEngine.DisableLog(false);

            base.OnClosed(e);
        }

        [System.Diagnostics.DebuggerStepThrough()]
        protected override void WndProc(ref Message m)
        {
            if (m.Msg == User32.WM_DEVICECHANGE)
            {
                MediaNotifier.ProcessDeviceChangedEvent(m.WParam, m.LParam);
            }

            base.WndProc(ref m);
        }

        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            switch (keyData)
            {
                case (Keys.Control | Keys.Shift | Keys.L):
                    if (!string.IsNullOrEmpty(BurningEngine.LastLogPath))
                    {
                        try
                        {
                            Process.Start("notepad", BurningEngine.LastLogPath);
                        }
                        catch { }
                        return true;
                    }
                    break;
            }

            return base.ProcessCmdKey(ref msg, keyData);
        }

        protected override void OnLoad(EventArgs e)
        {
            if (DesignMode) return;

            try
            {
                if (Culture != null)
                {
                    Thread.CurrentThread.CurrentUICulture = Culture;
                }
            }
            catch (Exception)
            {
                Thread.CurrentThread.CurrentUICulture = new CultureInfo("");
            }

            m_RemoveAutoPlay = new RemoveAutoPlay(this.Handle);

            // Fix CenterParent start position
            if (Parent != null && StartPosition == FormStartPosition.CenterParent) {
                StartPosition = FormStartPosition.Manual;
                Location = new Point(Parent.Location.X + Parent.Width - 3 * Width / 2, Parent.Location.Y + Parent.Height - 3 * Height / 2);
            }

            if (AutoTranslate)
            {
                if (!string.IsNullOrEmpty(RM.GetString(this.Text)))
                {
                    this.Text = RM.GetString(this.Text);
                }
                AutoTranslateControls(this.Controls);
            }

            if (this.RightToLeft == RightToLeft.Yes)
            {
                this.AdjustMirroredControls(this.Controls);
            }

            base.OnLoad(e);

            this.BeginInvoke(new EventHandler(OnLoaded));
        }

        #region Medium change

        private void Notifier_MediaRemoved(string VolumeLabel)
        {
            if (IsDisposed || !Created) return;
            this.BeginInvoke(new MethodInvoker(delegate()
            {
                OnMediaChanged(VolumeLabel);
                OnMediaRemoved(VolumeLabel);
            }));
        }

        private void Notifier_MediaInserted(string VolumeLabel)
        {
            if (IsDisposed || !Created) return;
            this.BeginInvoke(new MethodInvoker(delegate()
            {
                OnMediaChanged(VolumeLabel);
                OnMediaInserted(VolumeLabel);
            }));
        }

        protected virtual void OnMediaChanged(string VolumeLabel)
        {
        }

        protected virtual void OnMediaInserted(string VolumeLabel)
        {
        }

        protected virtual void OnMediaRemoved(string VolumeLabel)
        {
        }

        #endregion

        /// <summary>
        /// Occurs after the form has completly loaded.
        /// </summary>
        protected virtual void OnLoaded(object sender, EventArgs e)
        {
        }

        /// <summary>
        /// Automatically applies RM.GetString() to the Text-
        /// Property of all Controls.
        /// </summary>
        protected virtual void AutoTranslateControls(Control.ControlCollection controls)
        {
            foreach (Control c in controls)
            {
                AutoTranslateControls(c.Controls);

                if (!(c is DateTimePicker) && !(c is TextBox))
                {
                    c.Text = TranslateText(c.Text);
                }

                if (c.ContextMenu != null)
                {
                    foreach (MenuItem item in c.ContextMenu.MenuItems)
                    {
                        item.Text = TranslateText(item.Text);
                    }
                }
            }
        }

        /// <summary>
        /// Sets Mirror = true if RTL is active.
        /// </summary>
        protected virtual void AdjustMirroredControls(Control.ControlCollection controls)
        {
            foreach (Control c in controls)
            {
                AdjustMirroredControls(c.Controls);

                IMirrorControl mirrorControl = c as IMirrorControl;
                if (mirrorControl != null)
                {
                    mirrorControl.Mirror = true;
                }

                TabControl tabControl = c as TabControl;
                if (tabControl != null)
                {
                    tabControl.RightToLeftLayout = true;
                }
            }
        }

        private static string TranslateText(string text)
        {
            if (string.IsNullOrEmpty(text)) return text;

            for (int i = 0; i < text.Length; i++)
            {
                // Only translate strings that contain alpha-numeric text
                if (Char.IsLetterOrDigit(text, i))
                {
                    string result = RM.GetString(text);

                    if (string.IsNullOrEmpty(result))
                    {
                        result = RM.GetString(text.Replace("&", ""));
                    }

                    return string.IsNullOrEmpty(result) ? text : result;
                }
            }

            return text;
        }

        /// <summary>
        /// Sizes a button based on its content so that it is just big enough.
        /// </summary>
        protected void OptimizeSize(Button button)
        {
            Size size = TextRenderer.MeasureText(button.Text.Replace("&", ""), button.Font);
            button.Width = Math.Max(size.Width + 12, 75);
        }

        #region Burning engine events

        protected virtual void EngineOnFileVerifyStart(int lImageID, int lFileCount)
        {
        }

        protected virtual void EngineOnFileVerifyComplete(int lImageID, int lFileCount, int lMatchedFiles)
        {
        }

        protected virtual void EngineOnFileVerify(int lImageID, ref DataItemX objDataItem, double dFileBytesCompared, double dFileBytesTotal, eFileVerifyStatusCodeX fvStatus, int lFileIndex, int lFileCount, ref bool bCancel)
        {
        }

        protected virtual void EngineOnTrackFileStatus(int lImageID, int nTrackNumber, int lFileCount, int lFileIndex, string sFileName)
        {
        }

        protected virtual void EngineOnWriteStatus(int lImageID, short nImageProgress, int lImageBlocksWritten, int lImageTotalBlocks, short nCurrentTrack, short nTrackCount, short nTrackProgress, int lTrackBlocksWritten, int lTrackTotalBlocks, eWriteStatusCodeX status, ref bool bCancel)
        {
        }

        protected virtual void EngineOnWriteError(int lImageID, ref WriteErrorX objError)
        {
        }

        protected virtual void EngineOnTrackFileWarning(int lImageID, ref FileErrorX objError, int nTrackNumber, ref bool bIgnoreWarning)
        {
        }

        #endregion

        public BaseForm()
        {
            // We provide some overrides for NMS events, and can easily make sure 
            // that the events are removed again at a central location
            BurningEngine.NmsEngine.OnTrackFileWarning += new _IDVDEngineXEvents_OnTrackFileWarningEventHandler(EngineOnTrackFileWarning);
            BurningEngine.NmsEngine.OnWriteError += new _IDVDEngineXEvents_OnWriteErrorEventHandler(EngineOnWriteError);
            BurningEngine.NmsEngine.OnWriteStatus += new _IDVDEngineXEvents_OnWriteStatusEventHandler(EngineOnWriteStatus);
            BurningEngine.NmsEngine.OnTrackFileStatus += new _IDVDEngineXEvents_OnTrackFileStatusEventHandler(EngineOnTrackFileStatus);
            BurningEngine.NmsEngine.OnFileVerify += new _IDVDEngineXEvents_OnFileVerifyEventHandler(EngineOnFileVerify);
            BurningEngine.NmsEngine.OnFileVerifyComplete += new _IDVDEngineXEvents_OnFileVerifyCompleteEventHandler(EngineOnFileVerifyComplete);
            BurningEngine.NmsEngine.OnFileVerifyStart += new _IDVDEngineXEvents_OnFileVerifyStartEventHandler(EngineOnFileVerifyStart);
            MediaNotifier.MediaInserted += new MediaNotifier.MediaInsertedEventHandler(Notifier_MediaInserted);
            MediaNotifier.MediaRemoved += new MediaNotifier.MediaRemovedEventHandler(Notifier_MediaRemoved);

            InitializeComponent();

            if (this.AutoMirrorEnabled && BaseForm.IsRightToLeftCulture)
            {
                this.RightToLeft = RightToLeft.Yes;
                this.RightToLeftLayout = true;
            }
        }

        /// <summary>
        /// Tries to determine the parent form of the given controls.
        /// As soon as a parent form for one of the controls is found,
        /// the result is returned.
        /// </summary>
        /// <returns>null if no parent form could be determined</returns>
        public static Form GetParentForm(params Control[] controls)
        {
            foreach (Control c in controls)
            {
                Control parent = c.Parent;
                while (parent != null)
                {
                    Form result = parent as Form;
                    if (result != null) return result;

                    parent = parent.Parent;
                }
            }

            return null;
        }

        /// <summary>
        /// Provides thread-safe access to the Checked property of a CheckBox.
        /// </summary>
        public bool IsChecked(CheckBox checkbox)
        {
            if (checkbox.InvokeRequired)
            {
                return (bool)checkbox.Invoke((MethodReturnInvoker)delegate() { return checkbox.Checked; });
            }
            return checkbox.Checked;
        }

        /// <summary>
        /// Translated version of string.Format, with fallback if format string is invalid.
        /// </summary>
        public static string FormatTranslation(string formatString, params object[] values)
        {
            try
            {
                return string.Format(RM.GetString(formatString), values);
            }
            catch (FormatException)
            {
                return string.Format(RM.GetString(formatString, CultureInfo.GetCultureInfo("en-US")), values);
            }
            catch (ArgumentException)
            {
                return string.Format(formatString, values);
            }
        }

        /// <summary>
        /// Returns a new instance of a resource manager, which contains all language strings.
        /// </summary>
        public static ResourceManager GetResourceManager()
        {
            return new ResourceManager("CDBurnerXP_Pro.Strings", System.Reflection.Assembly.GetEntryAssembly());
        }

        public void SwitchToWindow(Form form)
        {
            TaskBarExtensions.SetMainWindowHandle(form.Handle);
            // Hide this window as soon as the new window is visible
            EventHandler visibleHandler = new EventHandler(delegate(object sender, EventArgs e)
            {
                this.Visible = false;
            });
            form.VisibleChanged += visibleHandler;
            form.ShowDialog(this);
            form.VisibleChanged -= visibleHandler;

            // Do not make this window re-appear if application should exit
            BaseForm baseForm = form as BaseForm;
            if (baseForm == null || !baseForm.ExitApplicationAfterClose)
            {
                TaskBarExtensions.SetMainWindowHandle(this.Handle);
                this.Visible = true;
            }
        }

        #region Utility burning features

        protected void LoadDisc(StarBurnX.Drive drive)
        {
            if (drive == null) return;

            try
            {
                drive.Load();
            }
            catch (COMException)
            {
                string msg = string.Format(RM.GetString("The tray of drive '{0}' cannot be closed. Most likely, it does not support closing the tray by software. Close the tray manually."), BurningEngine.FormatDevice(drive));
                MessageBox.Show(this, msg, RM.GetString("Close Tray"), MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /// <summary>
        /// Handles a StarBurn error by showing an appropriate message box.
        /// </summary>
        /// <param name="owner">Parent window for the message box</param>
        /// <param name="ex">Exception which occured</param>
        /// <param name="currentDevice">Currently used device</param>
        /// <returns>true, if the error could be handled, false otherwise</returns>
        protected static bool HandleStarBurnError(IWin32Window owner, COMException ex, object currentDevice)
        {
            if (ex.IsError(StarBurnX.STARBURN_RESULT_CODES.STARBURN_ERR_DEVICE_DISCONNECTED))
            {
                string msg = string.Format(RM.GetString("The drive '{0}' has changed or been disconnected and can no longer be used. If possible, reconnect the drive to your computer."), BurningEngine.FormatDevice(currentDevice));
                MessageBox.Show(owner, msg, Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
                BurningEngine.StarBurnEngine.Refresh();
                return true;
            }

            if (ex.IsError(StarBurnX.STARBURN_RESULT_CODES.STARBURN_ERR_DISC_READ_ERROR))
            {
                string msg = string.Format(RM.GetString("Cannot read the disc in drive '{0}'. Possibly, the drive is currently in use by another application. Close other applications and try again."), BurningEngine.FormatDevice(currentDevice));
                MessageBox.Show(owner, msg, Application.ProductName, MessageBoxButtons.OK, MessageBoxIcon.Error);
                return true;
            }

            return false;
        }

        #endregion

        private void InitializeComponent()
        {
            System.ComponentModel.ComponentResourceManager resources = new System.ComponentModel.ComponentResourceManager(typeof(BaseForm));
            this.SuspendLayout();
            // 
            // BaseForm
            // 
            this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
            this.Name = "BaseForm";
            this.ResumeLayout(false);
        }
    }
}
