﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using CDBurnerXP.Burning;
using CDBurnerXP.Controls;

namespace CDBurnerXP.Forms
{
    /// <summary>
    /// Represents a dialog which allows to edit CD-Text information.
    /// </summary>
    public partial class CDTextDialog : BaseForm
    {
        private AudioLayoutManager layoutManager = null;
        private int trackNumber = 0;
        private string trackInfoText = string.Empty;
        private Dictionary<int, CDTextInfo> cdTextInfo = new Dictionary<int, CDTextInfo>();

        #region Properties

        protected override bool AutoTranslate
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Gets or sets for which track the CD-Text information is displayed.
        /// </summary>
        public int TrackNumber
        {
            get
            {
                return this.trackNumber;
            }
            set
            {
                this.trackNumber = value;
                LoadTrackInformation();
            }
        }

        /// <summary>
        /// Gets or sets the audio image which contains the audio tracks.
        /// </summary>
        public AudioLayoutManager LayoutManager
        {
            get
            {
                return this.layoutManager;
            }
            set
            {
                this.layoutManager = value;
                if (this.layoutManager != null)
                {
                    for (int i = 0; i < this.layoutManager.AudioTracks.Length; i++)
                    {
                        this.cdTextInfo[i + 1] = new CDTextInfo(this.layoutManager.AudioTracks[i].CDTextArtist, this.layoutManager.AudioTracks[i].CDTextTitle);
                    }                    
                }
            }
        }

        #endregion

        /// <summary>
        /// Temporarily stores CD-Text information.
        /// </summary>
        private class CDTextInfo
        {
            private string artist;
            private string title;

            public string Artist
            {
                get { return artist; }
                set { artist = value; }
            }

            public string Title
            {
                get { return title; }
                set { title = value; }
            }

            public CDTextInfo(string artist, string title)
            {
                this.title = title;
                this.artist = artist;
            }
        }

        public CDTextDialog()
        {
            InitializeComponent();
            AcceptButton = bOK;
            CancelButton = bCancel;

            this.trackInfoText = sepTrack.Text;
        }

        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);

            if (this.layoutManager != null)
            {
                txtDiscArtist.Text = this.layoutManager.CDTextArtist;
                txtDiscTitle.Text = this.layoutManager.CDTextTitle;
            }

            if (TrackNumber == 0)
            {
                TrackNumber = 1;
            }
        }

        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            switch (keyData)
            {
                case Keys.Alt | Keys.Left:
                    bPreviousTrack.PerformClick();
                    return true;

                case Keys.Alt | Keys.Right:
                    bNextTrack.PerformClick();
                    return true;
            }

            return base.ProcessCmdKey(ref msg, keyData);
        }

        /// <summary>
        /// Loads the CD-Text information for the current track.
        /// </summary>
        private void LoadTrackInformation()
        {
            if (layoutManager == null) return;

            this.bPreviousTrack.Enabled = (this.trackNumber > 1);
            this.bNextTrack.Enabled = (this.trackNumber < this.layoutManager.AudioTracks.Length);

            bool validTrack = (this.trackNumber > 0 && this.trackNumber <= this.layoutManager.AudioTracks.Length);

            this.txtTrackArtist.Enabled = validTrack;
            this.txtTrackTitle.Enabled = validTrack;

            if (validTrack)
            {
                sepTrack.Text = string.Format(RM.GetString(this.trackInfoText), TrackNumber);
                
                this.txtTrackArtist.Text = this.cdTextInfo[this.trackNumber].Artist;
                this.txtTrackTitle.Text = this.cdTextInfo[this.trackNumber].Title;

                // Try to load track information from tags
                if (string.IsNullOrEmpty(this.txtTrackArtist.Text))
                {
                    this.txtTrackArtist.Text = LayoutManager.AudioTracks[this.trackNumber-1].Artist;
                }
                if (string.IsNullOrEmpty(this.txtTrackTitle.Text))
                {
                    this.txtTrackTitle.Text = LayoutManager.AudioTracks[this.trackNumber-1].Title;
                }
            }
            else
            {
                sepTrack.Text = RM.GetString("Track information");
            }
        }

        private void bOK_Click(object sender, EventArgs e)
        {
            if (this.layoutManager != null)
            {
                this.layoutManager.CDTextArtist = txtDiscArtist.Text;
                this.layoutManager.CDTextTitle = txtDiscTitle.Text;

                for (int i = 0; i < this.layoutManager.AudioTracks.Length; i++)
                {
                    this.layoutManager.AudioTracks[i].CDTextArtist = this.cdTextInfo[i + 1].Artist;
                    this.layoutManager.AudioTracks[i].CDTextTitle = this.cdTextInfo[i + 1].Title;
                }
            }
        }

        private void bNextTrack_Click(object sender, EventArgs e)
        {
            this.TrackNumber++;
        }

        private void bPreviousTrack_Click(object sender, EventArgs e)
        {
            this.TrackNumber--;
        }

        private void txtTrackTitle_TextChanged(object sender, EventArgs e)
        {
            this.cdTextInfo[TrackNumber].Title = txtTrackTitle.Text;
        }

        private void txtTrackArtist_TextChanged(object sender, EventArgs e)
        {
            this.cdTextInfo[TrackNumber].Artist = txtTrackArtist.Text;
        }
    }
}
