using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using CDBurnerXP.IO;
using CDBurnerXP.IO.Cue;
using CDBurnerXP.Controls;
using System.IO;

namespace CDBurnerXP.Forms
{
    public partial class SplitCueDialog
    {
        Dictionary<string, List<double>> m_Splits = new Dictionary<string, List<double>>();
        private List<string> m_ResultFiles = new List<string>();

        #region Properties

        public string[] ResultFiles
        {
            get
            {
                return m_ResultFiles.ToArray();
            }
        }

        #endregion

        public SplitCueDialog(CueSheet cue, string targetPath) : base()
        {
            InitializeComponent();
            AudioConverter.TempPath = targetPath;

            if (cue.Tracks.Length == 0)
            {
                m_Tracks = new AudioTrack[0];
                return;
            }

            // Go through all tracks and extract the first Index
            foreach (Track track in cue.Tracks)
            {
                if (!m_Splits.ContainsKey(track.DataFile.Filename))
                {
                    m_Splits.Add(track.DataFile.Filename, new List<double>());
                }
                foreach (Index index in track.Indices)
                {
                    if (index.Number == 1)
                    {
                        m_Splits[track.DataFile.Filename].Add(index.TotalSeconds);
                    }
                }
            }

            string[] tracks = new string[m_Splits.Count];
            m_Splits.Keys.CopyTo(tracks, 0);

            m_Tracks = new AudioTrack[m_Splits.Count];
            for (int i = 0; i < m_Tracks.Length; i++)
            {
                m_Tracks[i] = new AudioTrack(tracks[i], true);
            }
            
            AudioConverter.ProcessReplayGain = ReplayGainOption.None;
        }

        protected override void OnWorkerDoWork(object sender, DoWorkEventArgs e)
        {
            foreach (ProgressAddingTrack workItem in m_WorkItems)
            {
                if (m_Worker.CancellationPending)
                {
                    e.Cancel = true;
                    return;
                }

                m_CurrentTrack = workItem;
                workItem.Status = Status.Working;
                this.BeginInvoke(new MethodInvoker(delegate() { olvProgress.RefreshObject(workItem); }));
                string tempFileName = null;

                // Convert BIN to WAV if necessary (add WAVE header to raw data)
                if (PathEx.IsExtension(m_CurrentTrack.FileName, "bin"))
                {
                    if (!ConvertBinToWav(out tempFileName))
                    {
                        continue;
                    }
                }

                string actualFileName = tempFileName ?? m_CurrentTrack.FileName;

                double[] splitPos = m_Splits[m_CurrentTrack.FileName].ToArray();
                // No split necessary if track is added as a whole
                if (splitPos.Length == 1 && splitPos[0] == 0)
                {
                    m_ResultFiles.Add(actualFileName);
                    continue;
                }

                string[] resultFiles = AudioConverter.ToWav(actualFileName, -1, splitPos);
                if (resultFiles != null)
                {
                    PathEx.TryDeleteFiles(tempFileName);

                    m_ResultFiles.AddRange(resultFiles);
                }

                workItem.Progress = 100;
                workItem.Status = (resultFiles == null) ? Status.Failure : Status.Success;

                this.BeginInvoke(new MethodInvoker(delegate() { olvProgress.RefreshObject(workItem); }));
            }

            if (m_Worker.CancellationPending)
            {
                e.Cancel = true;
            }
        }
        
        /// <summary>
        /// Converts a BIN file to WAV.
        /// </summary>
        /// <remarks>A BIN file contains the raw audio data, so basically a WAV file without header. For conversion, we just create a 
        /// new WAV file and fill it with the raw data of the BIN file.</remarks>
        /// <param name="tempFileName">Temporary file name where the result WAV file is located</param>
        /// <returns>false if the conversion has been cancelled</returns>
        private bool ConvertBinToWav(out string tempFileName)
        {
            // Some disk space is necessary for storing the temp file
            if (!DiskSpaceRequiredDialog.Show(this, PathEx.TryGetFileSize(m_CurrentTrack.FileName)))
            {
                tempFileName = null;
                return false;
            }

            string tempFileNameLocal = PathEx.GetTempFileName("wav");

            using (ProgressDialog dialog = new ProgressDialog("Converting BIN to Audio", "Converting BIN/CUE data to audio data..."))
            {
                bool cancel = false;

                dialog.OnCancel = delegate
                {
                    cancel = true;
                };

                dialog.OnDoWork = delegate
                {
                    Un4seen.Bass.Misc.WaveWriter wavWriter = new Un4seen.Bass.Misc.WaveWriter(tempFileNameLocal, 2, 44100, 16, true);
                    using (FileStream fs = File.OpenRead(m_CurrentTrack.FileName))
                    {
                        BinaryReader binaryReader = new BinaryReader(fs);
                        long pos = 0;
                        // Cache stream length for performance
                        long length = binaryReader.BaseStream.Length;

                        while (pos < length)
                        {
                            if (cancel)
                            {
                                break;
                            }

                            // Read 1024 * 1024 bytes or less at stream end
                            short[] buffer = new short[Math.Min((length - pos) / sizeof(short), 1024 * 1024)];
                            for (int i = 0; i < buffer.Length; i++)
                            {
                                // WavWriter expects shorts (16 bit samples) instead of bytes (8 bit samples).
                                buffer[i] = binaryReader.ReadInt16();
                            }

                            dialog.ReportProgress(Conversion.ToInt((double)pos / length * 100.0));

                            wavWriter.Write(buffer, buffer.Length * 2);
                            pos += buffer.Length * 2;
                        }
                    }

                    // Dispose() actually deletes the file, so we only Close()
                    wavWriter.Close();

                    return true;
                };

                this.Invoke(new MethodInvoker(delegate { dialog.ShowDialog(this); }));

                if (cancel)
                {
                    tempFileName = null;
                    PathEx.TryDeleteFiles(tempFileNameLocal);
                    return false;
                }
                else if (dialog.Error != null)
                {
                    PathEx.TryDeleteFiles(tempFileNameLocal);
                    throw dialog.Error;
                }
            }

            tempFileName = tempFileNameLocal;
            return true;
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            if (m_Worker.IsBusy)
            {
                bCancel.Enabled = false;
                m_Worker.CancelAsync();
                e.Cancel = true;
            }
        }
    }
}