﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using NMSDVDXLib;
using CDBurnerXP.Controls;
using CDBurnerXP.Burning;
using CDBurnerXP.IO;

namespace CDBurnerXP.Forms
{
    /// <summary>
    /// Represents a dialog which allows the user to perform disc spanning.
    /// </summary>
    public partial class SplitDiscDialog : BaseForm
    {
        private DataCompilation dataCompilation = null;
        private DiscSplitter splitter = new DiscSplitter();
        private string resultingDiscsString = string.Empty;
        private string resultString = string.Empty;
        private List<IDataImageX> result = null;
        private DiscType previousDiscType = null;
        private DiscType customItem = null;

        #region Properties

        protected override bool AutoTranslate
        {
            get
            {
                return true;
            }
        }

        public IDataImageX[] ResultDataImages
        {
            get { return result.ToArray(); }
        }

        public DataCompilation DataCompilation
        {
            get { return this.dataCompilation; }
            set { this.dataCompilation = value; }
        }

        #endregion

        public SplitDiscDialog()
        {
            InitializeComponent();

            AcceptButton = bSplit;
            CancelButton = bCancelStart;

            this.splitter.TableOfContentsName = RM.GetString(splitter.TableOfContentsName);
            this.splitter.ProgressChanged += new EventHandler<DiscSplitter.SplitProgressChangedEventArgs>(SplitterProgressChanged);
            this.splitter.Completed += new EventHandler<DiscSplitter.SplitCompletedEventArgs>(SplitterCompleted);
        }

        protected override void OnLoad(EventArgs e)
        {
            if (DesignMode) return;

            base.OnLoad(e);

            this.resultingDiscsString = lblResultingDiscs.Text;
            this.resultString = lblResult.Text;
            lblInfo.Text = string.Format(lblInfo.Text, FormatFileSize.Format((float)this.dataCompilation.DataImage.GetImageSizeBytes()));

            cboDiscSize.Left = lblDiscSize.Right + 3;
            cboDiscSize.Width = bCancelStart.Right - cboDiscSize.Left;
            rbOptimiseDiscCount.Left = cboDiscSize.Left;
            rbOptimiseFolderStructure.Left = cboDiscSize.Left;

            // Fill disc combo box
            cboDiscSize.Items.Add(DiscType.CD21);
            cboDiscSize.Items.Add(DiscType.CD63);
            cboDiscSize.Items.Add(DiscType.CD74);
            cboDiscSize.Items.Add(DiscType.CD80);
            cboDiscSize.Items.Add(DiscType.CD90);
            cboDiscSize.Items.Add(DiscType.CD99);
            cboDiscSize.Items.Add(DiscType.DVD);
            cboDiscSize.Items.Add(DiscType.DVDPlus);
            cboDiscSize.Items.Add(DiscType.DVDDL);
            cboDiscSize.Items.Add(DiscType.HDDVD);
            cboDiscSize.Items.Add(DiscType.HDDVDDL);
            cboDiscSize.Items.Add(DiscType.BD);
            cboDiscSize.Items.Add(DiscType.BDDL);
            cboDiscSize.Items.Add(RM.GetString("<Choose custom size...>"));

            try
            {
                chkTableOfConents.Checked = Conversion.ToBoolean(Settings.GetValue(this, "TableOfContents", false));
                rbOptimiseDiscCount.Checked = Conversion.ToBoolean(Settings.GetValue(this, "OptimiseDiscCount", true));
                rbOptimiseFolderStructure.Checked = !rbOptimiseDiscCount.Checked;

                long customSize = Convert.ToInt64(Settings.GetValue(this, "CustomDiscSize", 0));
                if (customSize > 0)
                {
                    AddCustomItem(new DiscType(FormatTranslation("(Custom: {0})", FormatFileSize.Format(customSize)), customSize));
                }
                else
                {
                    cboDiscSize.SelectedIndex = Conversion.ToInt(Settings.GetValue(this, "LastUsedDiscSize", 3));
                }
            }
            catch (Exception)
            {
                cboDiscSize.SelectedIndex = 3;
            }
        }

        private void cboDiscSize_SelectedIndexChanged(object sender, EventArgs e)
        {
            // Choose a custom size
            if (cboDiscSize.SelectedIndex == cboDiscSize.Items.Count - 1)
            {
                using (CustomDiscSizeDialog dialog = new CustomDiscSizeDialog())
                {
                    dialog.CustomDiscType = this.previousDiscType;

                    if (dialog.ShowDialog(this) == DialogResult.OK)
                    {
                        AddCustomItem(dialog.CustomDiscType);
                    }
                    else
                    {
                        cboDiscSize.SelectedItem = previousDiscType;
                    }
                }
            }
            else
            {
                previousDiscType = cboDiscSize.SelectedItem as DiscType;
            }
        }

        private void AddCustomItem(DiscType discType)
        {
            if (this.customItem != null)
            {
                cboDiscSize.Items.Remove(this.customItem);
            }
            this.customItem = discType;
            cboDiscSize.Items.Insert(0, customItem);
            cboDiscSize.SelectedItem = this.customItem;
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            // Do not close while working
            if (this.splitter.IsBusy) e.Cancel = true;

            Settings.SetValue(this, "LastUsedDiscSize", (this.customItem != null) ? cboDiscSize.SelectedIndex - 1 : cboDiscSize.SelectedIndex);
            Settings.SetValue(this, "TableOfContents", chkTableOfConents.Checked);
            Settings.SetValue(this, "OptimiseDiscCount", rbOptimiseDiscCount.Checked);
            Settings.SetValue(this, "CustomDiscSize", (cboDiscSize.SelectedItem == this.customItem) ? this.customItem.SizeInBytes : 0);

            base.OnClosing(e);
        }

        public new DialogResult ShowDialog(IWin32Window owner)
        {
            // Check for imported data. Splitting not possible.
            if (this.dataCompilation.DataImage.IsImported())
            {
                MessageBox.Show(owner, RM.GetString("Disc spanning is not possible, because the data compilation contains files imported from a previous session of a disc.\r\n\r\nStart a new compilation in order to perform disc spanning."), "CDBurnerXP", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return DialogResult.Cancel;
            }

            return base.ShowDialog(owner);
        }

        private void bSplit_Click(object sender, EventArgs e)
        {
            // In order to more or less fill the discs as good as possible,
            // sort all files by size.
            splitter.MaxDiscSize = ((DiscType)cboDiscSize.SelectedItem).SizeInBytes;

            // Check: Is disc size greater than total size? -> Splitting not necessary.
            if (splitter.MaxDiscSize >= this.dataCompilation.DataImage.GetImageSizeBytes())
            {
                MessageBox.Show(this, RM.GetString("The selected disc size is greater than the amount of data to divide.\r\n\r\nSplitting is not necessary."), "CDBurnerXP", MessageBoxButtons.OK, MessageBoxIcon.None);
                return;
            }

            // Check: We need at least one file to operate properly
            if (this.dataCompilation.DataImage.GetFileLayout().GetFileCount() == 0)
            {
                MessageBox.Show(this, RM.GetString("Disc spanning is not possible, because the compilation does not contain any files."), "CDBurnerXP", MessageBoxButtons.OK, MessageBoxIcon.None);
                return;
            }

            prgProgress.Value = 0;
            pnlProgress.Location = pnlStart.Location;
            lblResultingDiscs.Text = string.Format(this.resultingDiscsString, 0);
            pnlProgress.Visible = true;
            pnlStart.Visible = false;
            bCancelProgress.Enabled = true;

            this.splitter.TableOfContents = chkTableOfConents.Checked;
            this.splitter.SplitAlgorithm = rbOptimiseDiscCount.Checked ? DiscSplitter.Algorithm.OptimiseDiscCount : DiscSplitter.Algorithm.OptimiseFolderStructure;
            this.splitter.BeginSplitLayout(this.dataCompilation);            
        }

        private void bCancelProgress_Click(object sender, EventArgs e)
        {
            bCancelProgress.Enabled = false;
            this.splitter.Cancel();
        }

        private void SplitterCompleted(object sender, DiscSplitter.SplitCompletedEventArgs e)
        {
            this.result = e.DataImages;

            this.BeginInvoke((MethodInvoker)delegate()
            {
                if (this.splitter.Cancelled)
                {
                    if (e.FileTooBig != null)
                    {
                        MessageBox.Show(this, string.Format(RM.GetString("Disc spanning is not possible, because the file size of \"{0}\" is greater than a single disc.\r\n\r\nRemove the file or use a disc type with greater capacity."), e.FileTooBig.FullVirtualPath.TrimStart('\\')), "CDBurnerXP", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                    pnlStart.Visible = true;
                    pnlProgress.Visible = false;
                }
                else
                {
                    pnlFinish.Location = pnlStart.Location;
                    pnlFinish.Visible = true;
                    pnlFinish.BringToFront();
                }
            });
        }

        private void SplitterProgressChanged(object sender, DiscSplitter.SplitProgressChangedEventArgs e)
        {
            this.BeginInvoke((MethodInvoker)delegate()
            {
                this.prgProgress.Value = e.Progress;
                this.lblResultingDiscs.Text = string.Format(this.resultingDiscsString, e.CurrentNumberOfDiscs);
                this.lblResult.Text = string.Format(this.resultString, e.CurrentNumberOfDiscs, FormatFileSize.Format(this.splitter.MaxDiscSize));
            });
        }

        private void bSaveAsIso_Click(object sender, EventArgs e)
        {
            using (SaveFileDialog dialog = new SaveFileDialog())
            {
                dialog.Title = RM.GetString("Save As ISO File");
                dialog.InitialDirectory = ApplicationSettings.DataSaveISOPath;
                dialog.FileName = this.dataCompilation.Name + ".iso";
                dialog.Filter = RM.GetString("ISO File") + " (*.iso)|*.iso";

                if (dialog.ShowDialog(this) == DialogResult.OK)
                {
                    for (int i = 0; i < this.result.Count; i++)
                    {
                        string currentFileName = (i > 0) ? PathEx.AppendToFileName(" (" + (i+1) + ")", dialog.FileName) : dialog.FileName;
                        if (!FileLayoutManager.SaveAsIso(this, this.result[i], currentFileName, FormatTranslation("Saving ISO image {0} of {1}", i+1, this.result.Count)))
                        {
                            return;
                        }
                    }

                    ApplicationSettings.DataSaveISOPath = dialog.FileName;
                    this.Close();
                }
            }
        }
    }
}
