﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using System.IO;
using System.Drawing;

namespace CDBurnerXP.IO
{
    /// <summary>
    /// Provides methods to handle an INI-File.
    /// </summary>
    public class IniFile
    {
        private string path;

        [DllImport("kernel32")]
        private static extern long WritePrivateProfileString(string section, string key, string val, string filePath);
        [DllImport("kernel32")] 
        private static extern int GetPrivateProfileString(string section, string key, string def, StringBuilder retVal, int size, string filePath);

        /// <summary>
        /// INI-file Constructor.
        /// </summary>
        public IniFile(string iniPath)
        {
            this.path = iniPath;
        }

        /// <summary>
        /// INI-file Constructor.
        /// </summary>
        public IniFile(string iniName, bool userSpecifc)
        {
            if (userSpecifc)
            {
                try
                {
                    // Try 1: UserAppDataPath
                    this.path = Path.Combine(Path.GetDirectoryName(Application.UserAppDataPath), iniName + ".ini");
                }
                catch (Exception)
                {
                    try
                    {
                        // Try 2: UserAppDataPath, emulated
                        this.path = Path.Combine(PathEx.TryGetFolderPath(Environment.SpecialFolder.ApplicationData), Application.CompanyName);
                        this.path = Path.Combine(this.path, Application.ProductName);
                        Directory.CreateDirectory(this.path);
                    }
                    catch (Exception)
                    {
                        // Try 3: LocalApplicationData
                        this.path = Path.Combine(PathEx.TryGetFolderPath(Environment.SpecialFolder.LocalApplicationData), Application.CompanyName);
                        this.path = Path.Combine(this.path, Application.ProductName);
                        Directory.CreateDirectory(this.path);
                    }
                    this.path = Path.Combine(this.path, iniName + ".ini");
                }
            }
            else
            {
                this.path = Path.Combine(Path.GetDirectoryName(Application.CommonAppDataPath), iniName + ".ini");
            }
        }

        /// <summary>
        /// Write data to the INI file
        /// </summary>
        /// <param name="section">Name of the section</param>
        /// <param name="key">Name of the key</param>
        /// <param name="value">value to write</param
        public void SetValue(string Section, string Key, string Value)
        {
            try
            {
                WritePrivateProfileString(Section, Key, Value, this.path);
            }
            catch (Exception)
            {
                // Ignore errors
            }
        }

        /// <summary>
        /// Read value from the INI-file
        /// </summary>
        /// <param name="section">Name of the section</param>
        /// <param name="key">Name of the key</param>
        /// <returns>String.Empty if no value is found</returns>
        public string GetValue(string Section, string Key)
        {
            return GetValue(Section, Key, string.Empty);
        }

        /// <summary>
        /// Read value from the INI-file
        /// </summary>
        /// <param name="section">Name of the section</param>
        /// <param name="key">Name of the key</param>
        /// <returns>defaultValue if no value is found</returns>
        public string GetValue(string Section, string Key, string defaultValue)
        {
            try
            {
                StringBuilder temp = new StringBuilder(255);
                if (GetPrivateProfileString(Section, Key, "", temp, 255, this.path) == 0)
                {
                    return defaultValue;
                }

                return temp.ToString();
            }
            catch (Exception)
            {
                return defaultValue;
            }
        }

        /// <summary>
        /// Read a boolean value from the INI-file
        /// </summary>
        /// <param name="section">Name of the section</param>
        /// <param name="key">Name of the key</param>
        public bool GetBoolValue(string section, string key, bool defaultValue)
        {
            try
            {
                string stringValue = GetValue(section, key, defaultValue ? "1" : "0");
                return Conversion.ToBoolean(stringValue);
            }
            catch (FormatException)
            {
                return false;
            }
        }

        /// <summary>
        /// Writes a boolean value to an INI-file.
        /// </summary>
        /// <param name="section">Name of the section</param>
        /// <param name="key">Name of the key</param>
        /// <param name="value">Value to save</param>
        public void SetBoolValue(string section, string key, bool value)
        {
            SetValue(section, key, Conversion.ToInt(value).ToString());
        }

        /// <summary>
        /// Writes a Point value to an INI-file.
        /// </summary>
        /// <param name="section">Name of the section</param>
        /// <param name="key">Name of the key</param>
        /// <param name="value">Value to save</param>
        public void SetPoint(string section, string key, Point value)
        {
            SetValue(section, key, value.X + ";" + value.Y);
        }

        /// <summary>
        /// Read a Point value from the INI-file
        /// </summary>
        /// <param name="section">Name of the section</param>
        /// <param name="key">Name of the key</param>
        public Point GetPoint(string section, string key)
        {
            try
            {
                string[] stringValues = GetValue(section, key).Split(';');
                int x = Conversion.ToInt(stringValues[0]);
                int y = Conversion.ToInt(stringValues[1]);

                return new Point(x, y);
            }
            catch (IndexOutOfRangeException)
            {
                return Point.Empty;
            }
        }

        /// <summary>
        /// Writes a Size value to an INI-file.
        /// </summary>
        /// <param name="section">Name of the section</param>
        /// <param name="key">Name of the key</param>
        /// <param name="value">Value to save</param>
        public void SetSize(string section, string key, Size value)
        {
            SetValue(section, key, value.Width + "*" + value.Height);
        }

        /// <summary>
        /// Read a Size value from the INI-file
        /// </summary>
        /// <param name="section">Name of the section</param>
        /// <param name="key">Name of the key</param>
        public Size GetSize(string section, string key)
        {
            try
            {
                string[] stringValues = GetValue(section, key).Split('*');
                int x = Conversion.ToInt(stringValues[0]);
                int y = Conversion.ToInt(stringValues[1]);

                return new Size(x, y);
            }
            catch (IndexOutOfRangeException)
            {
                return Size.Empty;
            }
        }
    }
}
