using System.IO;
using System.Collections.Generic;
using System;
using System.Xml;
using System.Text;

namespace CDBurnerXP.IO
{
    public class Playlists
    {
        /// <summary>
        /// Returns a list of all supported audio file extensions (without dot).
        /// </summary>
        public static List<string> GetSupportedExtensions(bool formatAsFileFilter)
        {
            List<string> extensions = new List<string>();

            string prepend = string.Empty;
            if (formatAsFileFilter)
            {
                prepend = "*.";
            }

            extensions.Add(prepend + "flac");
            extensions.Add(prepend + "mpc");
            extensions.Add(prepend + "wv");
            extensions.Add(prepend + "ape");
            extensions.Add(prepend + "mp3");
            extensions.Add(prepend + "ogg");
            extensions.Add(prepend + "wma");
            extensions.Add(prepend + "wav");
            extensions.Add(prepend + "aif");
            extensions.Add(prepend + "aiff"); // Audio Interchange File Format, PCM, supported by BASS
            extensions.Add(prepend + "mp2");  // Supported by BASS
            extensions.Add(prepend + "bwf"); // Broadcast Wave Format, compatible to WAV

            if (Burning.BurningEngine.IsAacSupported)
            {
                extensions.Add(prepend + "m4a");
                extensions.Add(prepend + "m4b");
                extensions.Add(prepend + "m4p");
                extensions.Add(prepend + "m4v");
                extensions.Add(prepend + "m4r");
                extensions.Add(prepend + "3gp");
                extensions.Add(prepend + "mp4");
                extensions.Add(prepend + "aac");
            }

            return extensions;
        }

        /// <summary>
        /// Determines whether or not the given file is a supported playlist format.
        /// </summary>
        public static bool IsSupportedPlaylist(string filePath)
        {
            return IsM3U(filePath) || IsNRA(filePath) || IsWPL(filePath);
        }

        /// <summary>
        /// Extracts all audio files contained in the given playlist (if supported).
        /// </summary>
        public static string[] GetFilesFromPlaylist(string filePath)
        {
            if (IsM3U(filePath))
            {
                return GetFilesFromM3u(filePath);
            }
            else if (IsWPL(filePath))
            {
                return GetFilesFromWpl(filePath);
            }
            else if (IsNRA(filePath))
            {
                return GetFilesFromNra(filePath);
            }

            return new string[0];
        }

        public static bool IsCUE(string filePath)
        {
            return Path.GetExtension(filePath).ToUpper() == ".CUE";
        }

        public static bool IsM3U(string filePath)
        {
            return Path.GetExtension(filePath).ToUpper() == ".M3U" || Path.GetExtension(filePath).ToUpper() == ".M3U8";
        }

        public static bool IsWPL(string filePath)
        {
            return Path.GetExtension(filePath).ToUpper() == ".WPL";
        }

        public static bool IsNRA(string filePath)
        {
            return Path.GetExtension(filePath).ToUpper() == ".NRA";
        }

        public static bool IsAudioFile(string filePath)
        {
            string ext = Path.GetExtension(filePath).ToLower().TrimStart('.');
            return GetSupportedExtensions(false).Contains(ext);            
        }

        public static string[] GetFilesFromNra(string filePath)
        {
            List<string> result = new List<string>();
            byte[] fileContent = File.ReadAllBytes(filePath);

            for (int i = 0; i < fileContent.Length; i++)
            {
                if (i + 4 >= fileContent.Length) break;

                // Search for x:\ as indication for a file path (consider \0 chars for indices)
                if (fileContent[i] >= 65 && fileContent[i] <= 122 && fileContent[i + 2] == 0x3a && fileContent[i + 4] == 0x5c)
                {
                    // Read all bytes until \0\0, then convert these as utf-16 string
                    int end = i;
                    for (end = i; end < fileContent.Length - 1; end++)
                    {
                        if (fileContent[end] == 0 && fileContent[end + 1] == 0) break;
                    }

                    // Create byte array from start/end indices
                    List<byte> fileNameBytes = new List<byte>();
                    for (int j = i; j <= end; j++)
                    {
                        fileNameBytes.Add(fileContent[j]);
                    }

                    // Convert bytes to string
                    string resultFileName = Encoding.GetEncoding("utf-16").GetString(fileNameBytes.ToArray());
                    // Do not add files twice in a row, file names are duplicated in the NRA file
                    if (result.Count == 0 || result[result.Count - 1] != resultFileName)
                    {
                        result.Add(resultFileName);
                    }
                }
            }

            return result.ToArray();
        }

        public static string[] GetFilesFromM3u(string filePath)
        {
            FileStream fs = null;
            StreamReader sr = null;
            List<string> files = new List<string>();

            try
            {
                fs = new FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read);
            }
            catch (Exception)
            {
                return files.ToArray();
            }

            // Use UTF-8 encoding for M3U8  Playlists (UTF-8 encoded playlists, Winamp)
            Encoding fileEncoding = (Path.GetExtension(filePath).ToLower() == ".m3u8") ? Encoding.UTF8 : Encoding.Default;

            try
            {
                string line;
                sr = new StreamReader(fs, fileEncoding, true);
                do
                {
                    line = sr.ReadLine();
                    if (line.IndexOf("#EXT") < 0)
                    {
                        string fullPath = PathEx.GetFullPath(filePath, line);
                        if (Playlists.IsAudioFile(fullPath) && File.Exists(fullPath))
                        {
                            files.Add(fullPath);
                        }
                    }
                } while (!sr.EndOfStream);
            }
            finally
            {
                if ((sr != null))
                    sr.Close();
                fs.Close();
            }

            return files.ToArray();
        }

        public static string[] GetFilesFromWpl(string filePath)
        {
            List<string> files = new List<string>();
            System.Xml.XmlDocument xml = new System.Xml.XmlDocument();
            System.Xml.XmlElement element;
            System.Xml.XmlNodeList media_list;
            int count = 0;

            // Read file
            try
            {
                xml.Load(filePath);
            }
            catch
            {
                return files.ToArray();
            }

            // Go to /smil/head/body
            try
            {
                element = (XmlElement)xml.DocumentElement.GetElementsByTagName("body").Item(0);
                element = (XmlElement)element.GetElementsByTagName("seq").Item(0);
                media_list = element.GetElementsByTagName("media");
            }
            catch
            {
                return files.ToArray();
            }

            // Add media files to mediafiles()
            for (count = 0; count <= media_list.Count - 1; count++)
            {
                string line = ((System.Xml.XmlElement)media_list.Item(count)).GetAttribute("src");

                string fullPath = PathEx.GetFullPath(filePath, line);
                if (Playlists.IsAudioFile(fullPath) && File.Exists(fullPath))
                {
                    files.Add(fullPath);
                }
            }

            return files.ToArray();
        }

        public static bool IsCda(string strpath)
        {
            if (Path.GetExtension(strpath).ToLower() != ".cda")
            {
                return false;
            }

            // Check for mixed mode CDAs, need special handling though.
            if (strpath.StartsWith("MixedMode:"))
            {
                return true;
            }

            
            string idCDDA = "";
            string idFMT = "";
            short idVer;
            byte[] buf = new byte[26];

            try
            {
                using (FileStream fs = new FileStream(strpath, FileMode.Open, FileAccess.Read, FileShare.Read))
                {
                    fs.Read(buf, 0, 24);
                }
            }
            catch
            {
                return false;
            }

            int i;
            for (i = 8; i <= 11; i++)
            {
                idCDDA += Convert.ToChar(buf[i]);
            }
            for (i = 12; i <= 14; i++)
            {
                idFMT += Convert.ToChar(buf[i]);
            }
            idVer = buf[20];
            if (idVer == 1 & idCDDA == "CDDA" & idFMT == "fmt")
            {
                return true;
            }

            return false;
        }
    }

}